
/*
 * The information contained in this file is confidential and proprietary to
 * Broadcom Corporation.  No part of this file may be reproduced or
 * distributed, in any form or by any means for any purpose, without the
 * express written permission of Broadcom Corporation.
 *
 * (c) COPYRIGHT 2008-2012 Broadcom Corporation, ALL RIGHTS RESERVED.
 */

/*
 * Copyright (c) 2002, 2011, Oracle and/or its affiliates. All rights reserved.
 */

#include "bnxe.h"
#ifdef __sparc
#include <v9/sys/mutex_impl.h>
#else
#include <sys/mutex_impl.h>
#endif
#include <sys/mdb_modapi.h>

#define OPT_NONE (uint64_t)0x0deadead

#define HW_NO_INIT "   Hardware not initialized...\n"

char _depends_on[]= "misc/kmdbmod";


typedef struct
{
    u16_t txq_idx[MAX_ETH_CONS];
    u16_t rxq_idx[MAX_ETH_CONS];
    u16_t rcq_idx[MAX_ETH_CONS];
} bnxe_chain_idx_t;


/* from mdb_ks.c */
extern int mdb_name_to_major(const char *name, major_t *major);


static int major_to_addr(major_t     major,
                         uintptr_t * vaddr)
{
    uint_t    devcnt;
    uintptr_t devnamesp;

    if (mdb_readvar(&devcnt, "devcnt") == -1)
    {
        mdb_warn("failed to read 'devcnt'");
        return -1;
    }

    if (mdb_readvar(&devnamesp, "devnamesp") == -1)
    {
        mdb_warn("failed to read 'devnamesp'");
        return -1;
    }

    if (major >= devcnt)
    {
        mdb_warn("%x is out of range [0x0-0x%x]\n", major, devcnt - 1);
        return -1;
    }

    *vaddr = devnamesp + (major * sizeof (struct devnames));
    return 0;
}


static void BnxeMdbBasic(um_device_t * pUM)
{
    lm_device_t * pLM = (lm_device_t *)&pUM->lm_dev;

    mdb_printf("Version:            %-10s Firmware:              %s\n",
               pUM->version, pUM->versionFW);
    mdb_printf("ChipName:           %-10s ChipID:                %s\n",
               pUM->chipName, pUM->chipID);
    mdb_printf("SF/MF:              %s (%d)\n",
               IS_MULTI_VNIC(pLM) ? "MF" : "SF",
               IS_MULTI_VNIC(pLM) ? pLM->params.vnics_per_port : 1);
    mdb_printf("\n");
    mdb_printf("L2 MacAddr:         %02x:%02x:%02x:%02x:%02x:%02x\n",
               pLM->params.mac_addr[0], pLM->params.mac_addr[1],
               pLM->params.mac_addr[2], pLM->params.mac_addr[3],
               pLM->params.mac_addr[4], pLM->params.mac_addr[5]);
    mdb_printf("\n");
    mdb_printf("HwInitDone:         %-10d ClientsHw:             %s\n",
               pUM->hwInitDone, BnxeClientsHw(pUM));
    mdb_printf("ClientsDevi:        %-10s ClientsBound:          %s\n",
               BnxeClientsDevi(pUM), BnxeClientsBound(pUM));
    mdb_printf("\n");
    mdb_printf("txMsgPullUp:        %-10d\n",
               pUM->txMsgPullUp);
    mdb_printf("\n");
    mdb_printf("memAllocBlocks:     %-10d memAllocDMAs:          %d\n",
               d_list_entry_cnt(&pUM->memBlockList),
               d_list_entry_cnt(&pUM->memDmaList));
    mdb_printf("memAllocBARs:       %-10d\n",
               d_list_entry_cnt(&pUM->memRegionList));
    mdb_printf("\n");
    mdb_printf("timerFired:         %-10d timerReply:            %d\n",
               pLM->vars.stats.stats_collect.timer_wakeup,
               pLM->vars.stats.stats_collect.stats_fw.drv_counter);
    mdb_printf("timerNoReplyTotal:  %-10d timerNoReplyCurrent:   %d\n",
               pLM->vars.stats.stats_collect.stats_fw.timer_wakeup_no_completion_total,
               pLM->vars.stats.stats_collect.stats_fw.timer_wakeup_no_completion_current);
    mdb_printf("timerDone:          %-10d\n",
               pLM->vars.stats.stats_collect.stats_fw.b_completion_done);
    mdb_printf("\n");
    mdb_printf("workQueueInstCnt:   %-10d workItemInstQueued:    %d\n",
               s_list_entry_cnt(&pUM->workqs.instq.workQueue),
               pUM->workqs.instq.workItemQueued);
    mdb_printf("workItemInstError:  %-10d workItemInstComplete:  %d\n",
               pUM->workqs.instq.workItemError,
               pUM->workqs.instq.workItemComplete);
    mdb_printf("\n");
    mdb_printf("workQueueDelayCnt:  %-10d workItemDelayQueued:   %d\n",
               s_list_entry_cnt(&pUM->workqs.delayq.workQueue),
               pUM->workqs.delayq.workItemQueued);
    mdb_printf("workItemDelayError: %-10d workItemDelayComplete: %d\n",
               pUM->workqs.delayq.workItemError,
               pUM->workqs.delayq.workItemComplete);
    mdb_printf("\n");
    mdb_printf("pcie_lane_width:    %-10d pcie_lane_speed:       %s\n",
               pLM->hw_info.pcie_lane_width,
               (pLM->hw_info.pcie_lane_speed == PCIE_LANE_SPEED_2_5G) ?
                   "2.5Gb"                                            :
               (pLM->hw_info.pcie_lane_speed == PCIE_LANE_SPEED_5G)   ?
                   "5Gb"                                              :
               (pLM->hw_info.pcie_lane_speed == PCIE_LANE_SPEED_8G)   ?
                   "8Gb"                                              :
                   "???");
    mdb_printf("intrAlloc:          %-10s intrFired:             %d\n",
               pUM->intrAlloc, pUM->intrFired);
}


static void BnxeLock(char *    pName,
                     uintptr_t addr)
{
    mutex_impl_t lock;

    if (mdb_vread(&lock, sizeof(lock), addr) == -1)
    {
        mdb_printf("ERROR: Unable to read mutex data for %s (%p)\n",
                   pName, addr);
        return;
    }

    if (MUTEX_TYPE_SPIN(&lock))
    {
        mdb_printf("WARNING: spin lock %s (%p) ???\n", pName, addr);
    }
    else
    {
        kthread_t * owner   = MUTEX_OWNER(&lock);
        char *      waiters = MUTEX_HAS_WAITERS(&lock) ? "yes" : "no";

#if 0
        if (!MUTEX_TYPE_ADAPTIVE(&lock) ||
            ((owner == NULL) && MUTEX_HAS_WAITERS(&lock)))
        {
            mdb_printf("ERROR: Invalid adaptive mutex %s (%p)\n" pName, addr);
            return
        }
#endif

        if (owner != NULL)
        {
            mdb_printf(" %-12s %0?p %7s %?p\n",
                       pName, addr, waiters, owner);
        }
        else
        {
            mdb_printf(" %-12s %0?p %7s %?s\n",
                       pName, addr, waiters, "no");
        }
    }
}


static void BnxeMdbLocks(um_device_t * pUM,
                         um_device_t * pReal)
{
    char buf[64];
    int idx;

    mdb_printf("-- MUTEXES --\n");

    mdb_printf(" %-12s %-?s %7s %?s\n",
               "LOCK", "ADDR", "WAITERS", "HELD");

    for (idx = 0; idx < (MAX_RSS_CHAINS + 1); idx++)
    {
        snprintf(buf, sizeof(buf), "%s%02d", "intr", idx);
        BnxeLock(buf, (uintptr_t)&pReal->intrMutex[idx]);
    }

    for (idx = 0; idx < (MAX_RSS_CHAINS + 1); idx++)
    {
        snprintf(buf, sizeof(buf), "%s%02d", "intrFlip", idx);
        BnxeLock(buf, (uintptr_t)&pReal->intrFlipMutex[idx]);
    }

    for (idx = 0; idx < MAX_ETH_CONS; idx++)
    {
        snprintf(buf, sizeof(buf), "%s%02d", "tx", idx);
        BnxeLock(buf, (uintptr_t)&pReal->txq[idx].txMutex);
    }

    for (idx = 0; idx < MAX_ETH_CONS; idx++)
    {
        snprintf(buf, sizeof(buf), "%s%02d", "freeTxDesc", idx);
        BnxeLock(buf, (uintptr_t)&pReal->txq[idx].freeTxDescMutex);
    }

    for (idx = 0; idx < MAX_ETH_CONS; idx++)
    {
        snprintf(buf, sizeof(buf), "%s%02d", "rx", idx);
        BnxeLock(buf, (uintptr_t)&pReal->rxq[idx].rxMutex);
    }

    for (idx = 0; idx < MAX_ETH_CONS; idx++)
    {
        snprintf(buf, sizeof(buf), "%s%02d", "doneRx", idx);
        BnxeLock(buf, (uintptr_t)&pReal->rxq[idx].doneRxMutex);
    }

    if (pUM->hwInitDone)
    {
        BnxeLock("instq", (uintptr_t)&pReal->workqs.instq.workQueueMutex);
        BnxeLock("delayq", (uintptr_t)&pReal->workqs.delayq.workQueueMutex);
    }

    BnxeLock("phy", (uintptr_t)&pReal->phyMutex);
    BnxeLock("ind", (uintptr_t)&pReal->indMutex);
    BnxeLock("cid", (uintptr_t)&pReal->cidMutex);
    BnxeLock("spq", (uintptr_t)&pReal->spqMutex);
    BnxeLock("spReq", (uintptr_t)&pReal->spReqMutex);
    BnxeLock("rrReq", (uintptr_t)&pReal->rrReqMutex);
    BnxeLock("islesCtrl", (uintptr_t)&pReal->islesCtrlMutex);
    BnxeLock("mem", (uintptr_t)&pReal->memMutex);
    BnxeLock("offload", (uintptr_t)&pReal->offloadMutex);
    BnxeLock("hwInit", (uintptr_t)&pReal->hwInitMutex);
    BnxeLock("timer", (uintptr_t)&pReal->timerMutex);
    BnxeLock("kstat", (uintptr_t)&pReal->kstatMutex);
    BnxeLock("gld", (uintptr_t)&pReal->gldMutex);

    mdb_printf("\n");

    mdb_printf("-- RWLOCKS --\n");

    mdb_printf(" gldTxMutex\n");
    if (mdb_call_dcmd("rwlock", (uintptr_t)&pReal->gldTxMutex,
                      DCMD_ADDRSPEC, 0, NULL) == -1)
    {
        mdb_printf("ERROR: Unable to read rwlock data for %s (%p)\n",
                   "gldTxMutex", (uintptr_t)&pReal->gldTxMutex);
        return;
    }
}


static void BnxeMdbIntr(um_device_t * pUM)
{
    lm_device_t * pLM = (lm_device_t *)&pUM->lm_dev;

    mdb_printf("-- INTERRUPTS --\n");
    mdb_printf("   intrAlloc:      %-10s intrFired:      %d\n",
               pUM->intrAlloc, pUM->intrFired);
    mdb_printf("   intrWrongState: %-10d intrInDisabled: %d\n",
               pLM->vars.dbg_intr_in_wrong_state,
               pLM->vars.dbg_intr_in_disabled);
    mdb_printf("   intrZeroStatus: %-10d\n",
               pLM->vars.dbg_intr_zero_status);
    mdb_printf("  < IGU Fired >\n");
    mdb_printf("   sb_00:          %-10d sb_09:          %d\n",
               pUM->intrSbCnt[0], pUM->intrSbCnt[9]);
    mdb_printf("   sb_01:          %-10d sb_10:          %d\n",
               pUM->intrSbCnt[1], pUM->intrSbCnt[10]);
    mdb_printf("   sb_02:          %-10d sb_11:          %d\n",
               pUM->intrSbCnt[2], pUM->intrSbCnt[11]);
    mdb_printf("   sb_03:          %-10d sb_12:          %d\n",
               pUM->intrSbCnt[3], pUM->intrSbCnt[12]);
    mdb_printf("   sb_04:          %-10d sb_13:          %d\n",
               pUM->intrSbCnt[4], pUM->intrSbCnt[13]);
    mdb_printf("   sb_05:          %-10d sb_14:          %d\n",
               pUM->intrSbCnt[5], pUM->intrSbCnt[14]);
    mdb_printf("   sb_06:          %-10d sb_15:          %d\n",
               pUM->intrSbCnt[6], pUM->intrSbCnt[15]);
    mdb_printf("   sb_07:          %-10d sb_16:          %d\n",
               pUM->intrSbCnt[7], pUM->intrSbCnt[16]);
    mdb_printf("   sb_08:          %-10d\n",
               pUM->intrSbCnt[8]);
    mdb_printf("  < IGU Fired (no change) >\n");
    mdb_printf("   sb_nc_00:       %-10d sb_nc_09:       %d\n",
               pUM->intrSbNoChangeCnt[0], pUM->intrSbNoChangeCnt[9]);
    mdb_printf("   sb_nc_01:       %-10d sb_nc_10:       %d\n",
               pUM->intrSbNoChangeCnt[1], pUM->intrSbNoChangeCnt[10]);
    mdb_printf("   sb_nc_02:       %-10d sb_nc_11:       %d\n",
               pUM->intrSbNoChangeCnt[2], pUM->intrSbNoChangeCnt[11]);
    mdb_printf("   sb_nc_03:       %-10d sb_nc_12:       %d\n",
               pUM->intrSbNoChangeCnt[3], pUM->intrSbNoChangeCnt[12]);
    mdb_printf("   sb_nc_04:       %-10d sb_nc_13:       %d\n",
               pUM->intrSbNoChangeCnt[4], pUM->intrSbNoChangeCnt[13]);
    mdb_printf("   sb_nc_05:       %-10d sb_nc_14:       %d\n",
               pUM->intrSbNoChangeCnt[5], pUM->intrSbNoChangeCnt[14]);
    mdb_printf("   sb_nc_06:       %-10d sb_nc_15:       %d\n",
               pUM->intrSbNoChangeCnt[6], pUM->intrSbNoChangeCnt[15]);
    mdb_printf("   sb_nc_07:       %-10d sb_nc_16:       %d\n",
               pUM->intrSbNoChangeCnt[7], pUM->intrSbNoChangeCnt[16]);
    mdb_printf("   sb_nc_08:       %-10d\n",
               pUM->intrSbNoChangeCnt[8]);
    mdb_printf("  < IGU Polled >\n");
    mdb_printf("   sb_poll_00:     %-10d sb_poll_09:     %d\n",
               pUM->intrSbPollCnt[0], pUM->intrSbPollCnt[9]);
    mdb_printf("   sb_poll_01:     %-10d sb_poll_10:     %d\n",
               pUM->intrSbPollCnt[1], pUM->intrSbPollCnt[10]);
    mdb_printf("   sb_poll_02:     %-10d sb_poll_11:     %d\n",
               pUM->intrSbPollCnt[2], pUM->intrSbPollCnt[11]);
    mdb_printf("   sb_poll_03:     %-10d sb_poll_12:     %d\n",
               pUM->intrSbPollCnt[3], pUM->intrSbPollCnt[12]);
    mdb_printf("   sb_poll_04:     %-10d sb_poll_13:     %d\n",
               pUM->intrSbPollCnt[4], pUM->intrSbPollCnt[13]);
    mdb_printf("   sb_poll_05:     %-10d sb_poll_14:     %d\n",
               pUM->intrSbPollCnt[5], pUM->intrSbPollCnt[14]);
    mdb_printf("   sb_poll_06:     %-10d sb_poll_15:     %d\n",
               pUM->intrSbPollCnt[6], pUM->intrSbPollCnt[15]);
    mdb_printf("   sb_poll_07:     %-10d sb_poll_16:     %d\n",
               pUM->intrSbPollCnt[7], pUM->intrSbPollCnt[16]);
    mdb_printf("   sb_poll_08:     %-10d\n",
               pUM->intrSbPollCnt[8]);
    mdb_printf("  < IGU Polled (no change) >\n");
    mdb_printf("   sb_poll_nc_00:  %-10d sb_poll_nc_09:  %d\n",
               pUM->intrSbPollNoChangeCnt[0], pUM->intrSbPollNoChangeCnt[9]);
    mdb_printf("   sb_poll_nc_01:  %-10d sb_poll_nc_10:  %d\n",
               pUM->intrSbPollNoChangeCnt[1], pUM->intrSbPollNoChangeCnt[10]);
    mdb_printf("   sb_poll_nc_02:  %-10d sb_poll_nc_11:  %d\n",
               pUM->intrSbPollNoChangeCnt[2], pUM->intrSbPollNoChangeCnt[11]);
    mdb_printf("   sb_poll_nc_03:  %-10d sb_poll_nc_12:  %d\n",
               pUM->intrSbPollNoChangeCnt[3], pUM->intrSbPollNoChangeCnt[12]);
    mdb_printf("   sb_poll_nc_04:  %-10d sb_poll_nc_13:  %d\n",
               pUM->intrSbPollNoChangeCnt[4], pUM->intrSbPollNoChangeCnt[13]);
    mdb_printf("   sb_poll_nc_05:  %-10d sb_poll_nc_14:  %d\n",
               pUM->intrSbPollNoChangeCnt[5], pUM->intrSbPollNoChangeCnt[14]);
    mdb_printf("   sb_poll_nc_06:  %-10d sb_poll_nc_15:  %d\n",
               pUM->intrSbPollNoChangeCnt[6], pUM->intrSbPollNoChangeCnt[15]);
    mdb_printf("   sb_poll_nc_07:  %-10d sb_poll_nc_16:  %d\n",
               pUM->intrSbPollNoChangeCnt[7], pUM->intrSbPollNoChangeCnt[16]);
    mdb_printf("   sb_poll_nc_08:  %-10d\n",
               pUM->intrSbPollNoChangeCnt[8]);
}


static void BnxeMdbSb(um_device_t * pUM,
                      int           idx)
{
    lm_device_t *                   pLM = (lm_device_t *)&pUM->lm_dev;
    struct host_sp_status_block     host_sp_sb;
    struct host_hc_status_block_e2  e2_sb;
    struct host_hc_status_block_e1x e1x_sb;
    u16_t hwEqConIdx  = 0;
    u16_t hwU, hwC;
    int i;

    if (idx == DEF_STATUS_BLOCK_INDEX)
    {
        mdb_printf("-- SB: Default --\n");

        if (pLM->vars.gen_sp_status_block.hc_sp_status_blk == NULL)
        {
            mdb_printf(HW_NO_INIT);
            return;
        }

        if (mdb_vread(&host_sp_sb, sizeof(struct host_sp_status_block),
                      (uint64_t)pLM->vars.gen_sp_status_block.hc_sp_status_blk) == -1)
        {
            mdb_printf("   ERROR: Unable to read host_sp_status_block\n");
            return;
        }

        mdb_printf("   hwSpIdx:        %-12d lcSpIdx:        %d\n",
                   mm_le16_to_cpu(host_sp_sb.sp_sb.running_index),
                   pLM->vars.hc_def_ack);
        mdb_printf("   hwAttnIdx:      %-12d lcAttnIdx:      %d\n",
                   mm_le16_to_cpu(host_sp_sb.atten_status_block.attn_bits_index),
                   pLM->vars.attn_def_ack);

        mdb_printf("\n");

        mdb_printf("-- Send Queue --\n");

        mdb_printf("   state:          %-12d sqCompSched:    %d\n",
                   pLM->sq_info.sq_state, pLM->sq_info.sq_comp_scheduled);
        mdb_printf("   sqProdIdx:      %-12d sqConsIdx:      %d\n",
                   pLM->sq_info.sq_chain.prod_idx,
                   pLM->sq_info.sq_chain.con_idx);
        mdb_printf("   sqBdLeft:       %-12d sqPendComplete: %d\n",
                   pLM->sq_info.sq_chain.bd_left,
                   d_list_entry_cnt(&pLM->sq_info.pending_complete));
        mdb_printf("   sqPendNormal:   %-12d sqPendHigh:     %d\n",
                   d_list_entry_cnt(&pLM->sq_info.pending_normal),
                   d_list_entry_cnt(&pLM->sq_info.pending_high));
        for (i = 0; i < MAX_CONNECTION_TYPE; i++)
        {
            if (!pLM->sq_info.sq_comp_cb[i]) continue;

            mdb_printf("   sqCompCb %02d:    %-12p\n",
                       i, pLM->sq_info.sq_comp_cb[i]);
        }

        mdb_printf("-- Event Queue --\n");

        if (pLM->eq_info.eq_chain.hw_con_idx_ptr != NULL)
        {
            mdb_vread(&hwEqConIdx, sizeof(u16_t),
                      (uint64_t)pLM->eq_info.eq_chain.hw_con_idx_ptr);
        }

        mdb_printf("   hwEqConIdx:     %-12d eqBdLeft:       %d\n",
                   mm_le16_to_cpu(hwEqConIdx),
                   pLM->eq_info.eq_chain.bd_chain.bd_left);
        mdb_printf("   eqBdProdIdx:    %-12d eqBdConsIdx:    %d\n",
                   pLM->eq_info.eq_chain.bd_chain.prod_idx,
                   pLM->eq_info.eq_chain.bd_chain.cons_idx);
    }
    else
    {
        mdb_printf(" < SB >\n");

        if (CHIP_IS_E1x(pLM))
        {
            if (pLM->vars.status_blocks_arr[idx].host_hc_status_block.e1x_sb == NULL)
            {
                mdb_printf(HW_NO_INIT);
                return;
            }

            if (mdb_vread(&e1x_sb, sizeof(struct host_hc_status_block_e1x),
                          (uint64_t)pLM->vars.status_blocks_arr[idx].host_hc_status_block.e1x_sb) == -1)
            {
                mdb_printf("   ERROR: Unable to read host_hc_status_block_e2\n");
                return;
            }

            hwU = mm_le16_to_cpu(e1x_sb.sb.running_index[SM_RX_ID]);
            hwC = mm_le16_to_cpu(e1x_sb.sb.running_index[SM_TX_ID]);
        }
        else /* E2 and E3 */
        {
            if (pLM->vars.status_blocks_arr[idx].host_hc_status_block.e2_sb == NULL)
            {
                mdb_printf(HW_NO_INIT);
                return;
            }

            if (mdb_vread(&e2_sb, sizeof(struct host_hc_status_block_e2),
                          (uint64_t)pLM->vars.status_blocks_arr[idx].host_hc_status_block.e2_sb) == -1)
            {
                mdb_printf("   ERROR: Unable to read host_hc_status_block_e2\n");
                return;
            }

            hwU = mm_le16_to_cpu(e2_sb.sb.running_index[SM_RX_ID]);
            hwC = mm_le16_to_cpu(e2_sb.sb.running_index[SM_TX_ID]);
        }

        mdb_printf("   hwSpUstorm:     %-12d lcSpUstorm:     %d\n",
                   hwU, pLM->vars.u_hc_ack[idx]);
        mdb_printf("   hwSpCstorm:     %-12d lcSpCstorm:     %d\n",
                   hwC, pLM->vars.c_hc_ack[idx]);
    }
}


static inline void BnxeMdbGetChainIdxs(um_device_t *      pUM,
                                       bnxe_chain_idx_t * pIdxs)
{
    lm_device_t *    pLM = (lm_device_t *)&pUM->lm_dev;
    lm_tx_chain_t *  pTxq;
    lm_rx_chain_t *  pRxq;
    lm_rcq_chain_t * pRcq;
    int idx;

    LM_FOREACH_RSS_IDX(&pUM->lm_dev, idx)
    {
        pTxq = &LM_TXQ(pLM, idx);
        pRxq = &LM_RXQ(pLM, idx);
        pRcq = &LM_RCQ(pLM, idx);

        if ((pTxq->hw_con_idx_ptr != NULL) ||
            (pRxq->hw_con_idx_ptr != NULL) ||
            (pRcq->hw_con_idx_ptr != NULL))
        {
            /* some might fail... */
            mdb_vread(&pIdxs->txq_idx[idx], sizeof(u16_t), (uint64_t)pTxq->hw_con_idx_ptr);
            mdb_vread(&pIdxs->rxq_idx[idx], sizeof(u16_t), (uint64_t)pRxq->hw_con_idx_ptr);
            mdb_vread(&pIdxs->rcq_idx[idx], sizeof(u16_t), (uint64_t)pRcq->hw_con_idx_ptr);
        }
    }

    if (BNXE_FCOE(pUM))
    {
        idx = FCOE_CID(&pUM->lm_dev);

        pTxq = &LM_TXQ(pLM, idx);
        pRxq = &LM_RXQ(pLM, idx);
        pRcq = &LM_RCQ(pLM, idx);

        if ((pTxq->hw_con_idx_ptr != NULL) ||
            (pRxq->hw_con_idx_ptr != NULL) ||
            (pRcq->hw_con_idx_ptr != NULL))
        {
            /* some might fail... */
            mdb_vread(&pIdxs->txq_idx[idx], sizeof(u16_t), (uint64_t)pTxq->hw_con_idx_ptr);
            mdb_vread(&pIdxs->rxq_idx[idx], sizeof(u16_t), (uint64_t)pRxq->hw_con_idx_ptr);
            mdb_vread(&pIdxs->rcq_idx[idx], sizeof(u16_t), (uint64_t)pRcq->hw_con_idx_ptr);
        }
    }
}


static void BnxeMdbChain(um_device_t *      pUM,
                         bnxe_chain_idx_t * pIdxs,
                         int                idx)
{
    lm_device_t *    pLM  = (lm_device_t *)&pUM->lm_dev;
    lm_tx_chain_t *  pTxq = &LM_TXQ(pLM, idx);
    lm_rx_chain_t *  pRxq = &LM_RXQ(pLM, idx);
    lm_rcq_chain_t * pRcq = &LM_RCQ(pLM, idx);
    lm_eq_chain_t *  pEq  = NULL;
    u16_t hwEqConIdx  = 0;

    mdb_printf("-- Chain %d %s --\n", idx,
               (idx == FCOE_CID(pLM)) ? "(FCoE) " : "");

    mdb_printf(" < Conn >\n");
#if 0
    mdb_printf("   state:          %-12s con_cid: %d\n",
               (pLM->vars.connections[idx].con_state == LM_CON_STATE_CLOSE)     ? "CLOSED"     :
               (pLM->vars.connections[idx].con_state == LM_CON_STATE_OPEN)      ? "OPENED"     :
               (pLM->vars.connections[idx].con_state == LM_CON_STATE_HALT)      ? "HALTED"     :
               (pLM->vars.connections[idx].con_state == LM_CON_STATE_TERMINATE) ? "TERMINATED" :
                                                                                  "UNKNOWN",
               pLM->vars.connections[idx].con_cid);
#endif

    BnxeMdbSb(pUM, (idx == FCOE_CID(pLM)) ? LM_NON_RSS_SB(pLM) : idx);

    if (idx == FCOE_CID(pLM))
    {
        pEq = &LM_FC_EQ(pLM, LM_NON_RSS_SB(pLM));

        if (pEq->hw_con_idx_ptr != NULL)
        {
            mdb_vread(&hwEqConIdx, sizeof(u16_t), (uint64_t)pEq->hw_con_idx_ptr);
        }

        mdb_printf(" < EQ >\n");
        mdb_printf("   hwEqConIdx:     %-12d eqBdLeft:       %d\n",
                   mm_le16_to_cpu(hwEqConIdx), pEq->bd_chain.bd_left);
        mdb_printf("   eqBdProdIdx:    %-12d eqBdConsIdx:    %d\n",
                   pEq->bd_chain.prod_idx, pEq->bd_chain.cons_idx);
    }

    mdb_printf(" < RxQ >\n");
    mdb_printf("   rxqBdTotal:     %-12d rxqBdLeft:      %d\n",
               pRxq->chain_arr[LM_RXQ_CHAIN_IDX_BD].capacity,
               pRxq->chain_arr[LM_RXQ_CHAIN_IDX_BD].bd_left);
    mdb_printf("   rxqBdPageCnt:   %-12d rxqBdsPerPage:  %d\n",
               pRxq->chain_arr[LM_RXQ_CHAIN_IDX_BD].page_cnt,
               pRxq->chain_arr[LM_RXQ_CHAIN_IDX_BD].bds_per_page);
    mdb_printf("   rxqBdSize:      %-12d rxqBdsSkipEop:  %d\n",
               pRxq->chain_arr[LM_RXQ_CHAIN_IDX_BD].bd_size,
               pRxq->chain_arr[LM_RXQ_CHAIN_IDX_BD].bds_skip_eop);
    mdb_printf("   rxqBdProdIdx:   %-12d rxqBdConsIdx:   %d\n",
               pRxq->chain_arr[LM_RXQ_CHAIN_IDX_BD].prod_idx,
               pRxq->chain_arr[LM_RXQ_CHAIN_IDX_BD].cons_idx);
    mdb_printf("   hwRxqConIdx:    %-12d\n",
               mm_le16_to_cpu(pIdxs->rxq_idx[idx]));
    mdb_printf(" < SgeQ >\n");
    mdb_printf("   sgeBdTotal:     %-12d sgeBdLeft:      %d\n",
               pRxq->chain_arr[LM_RXQ_CHAIN_IDX_SGE].capacity,
               pRxq->chain_arr[LM_RXQ_CHAIN_IDX_SGE].bd_left);
    mdb_printf("   sgeBdPageCnt:   %-12d sgeBdsPerPage:  %d\n",
               pRxq->chain_arr[LM_RXQ_CHAIN_IDX_SGE].page_cnt,
               pRxq->chain_arr[LM_RXQ_CHAIN_IDX_SGE].bds_per_page);
    mdb_printf("   sgeBdSize:      %-12d sgeBdsSkipEop:  %d\n",
               pRxq->chain_arr[LM_RXQ_CHAIN_IDX_SGE].bd_size,
               pRxq->chain_arr[LM_RXQ_CHAIN_IDX_SGE].bds_skip_eop);
    mdb_printf("   sgeBdProdIdx:   %-12d sgeBdConsIdx:   %d\n",
               pRxq->chain_arr[LM_RXQ_CHAIN_IDX_SGE].prod_idx,
               pRxq->chain_arr[LM_RXQ_CHAIN_IDX_SGE].cons_idx);
    mdb_printf(" < RcQ >\n");
    mdb_printf("   rcqBdTotal:     %-12d rcqBdLeft:      %d\n",
               pRcq->bd_chain.capacity,
               pRcq->bd_chain.bd_left);
    mdb_printf("   rcqBdPageCnt:   %-12d rcqBdsPerPage:  %d\n",
               pRcq->bd_chain.page_cnt,
               pRcq->bd_chain.bds_per_page);
    mdb_printf("   rcqBdSize:      %-12d rcqBdsSkipEop:  %d\n",
               pRcq->bd_chain.bd_size,
               pRcq->bd_chain.bds_skip_eop);
    mdb_printf("   rcqBdProdIdx:   %-12d rcqBdConsIdx:   %d\n",
               pRcq->bd_chain.prod_idx,
               pRcq->bd_chain.cons_idx);
    mdb_printf("   hwRcqConIdx:    %-12d\n",
               mm_le16_to_cpu(pIdxs->rcq_idx[idx]));
    mdb_printf(" < Rx >\n");
    mdb_printf("   rxFreeDescs:    %-12d rxActiveDescs:  %d\n",
               s_list_entry_cnt(&LM_RXQ(pLM, idx).common.free_descq),
               s_list_entry_cnt(&LM_RXQ(pLM, idx).active_descq));
    mdb_printf("   rxDoneDescs:    %-12d rxWaitingDescs: %d\n",
               s_list_entry_cnt(&pUM->rxq[idx].doneRxQ),
               s_list_entry_cnt(&pUM->rxq[idx].waitRxQ));
    mdb_printf("   rxCopied:       %-12d rxDiscards:     %d\n",
               pUM->rxq[idx].rxCopied,
               pUM->rxq[idx].rxDiscards);
    mdb_printf("   rxBufUpInStack: %-12d rxLowWater:     %d\n",
               pUM->rxq[idx].rxBufUpInStack,
               pUM->rxq[idx].rxLowWater);
    mdb_printf("   inPollMode:     %-12d pollCnt:        %d\n",
               pUM->rxq[idx].inPollMode,
               pUM->rxq[idx].pollCnt);
    mdb_printf("   intrEnableCnt:  %-12d intrDisableCnt: %d\n",
               pUM->rxq[idx].intrEnableCnt,
               pUM->rxq[idx].intrDisableCnt);
    mdb_printf("   genNumber:      %d\n",
               pUM->rxq[idx].genNumber);

    mdb_printf(" < TxQ >\n");
    mdb_printf("   txBdTotal:      %-12d txBdLeft:       %d\n",
               pTxq->bd_chain.capacity,
               pTxq->bd_chain.bd_left);
    mdb_printf("   txBdPageCnt:    %-12d txBdsPerPage:   %d\n",
               pTxq->bd_chain.page_cnt,
               pTxq->bd_chain.bds_per_page);
    mdb_printf("   txBdSize:       %-12d txBdsSkipEop:   %d\n",
               pTxq->bd_chain.bd_size,
               pTxq->bd_chain.bds_skip_eop);
    mdb_printf("   txBdProdIdx:    %-12d txBdConsIdx:    %d\n",
               pTxq->bd_chain.prod_idx,
               pTxq->bd_chain.cons_idx);
    mdb_printf("   hwTxqConIdx:    %-12d txPktIdx:       %d\n",
               mm_le16_to_cpu(pIdxs->txq_idx[idx]),
               pTxq->pkt_idx);
    mdb_printf("   txSentDescs:    %-12d txFreeDesc:     %d\n",
               s_list_entry_cnt(&pUM->txq[idx].sentTxQ),
               s_list_entry_cnt(&pUM->txq[idx].freeTxDescQ));
    mdb_printf("   txWaitingDescs: %-12d txLowWater:     %d\n",
               s_list_entry_cnt(&pUM->txq[idx].waitTxDescQ),
               pUM->txq[idx].txLowWater);
    mdb_printf("   txFailed:       %-12d txDiscards:     %d\n",
               pUM->txq[idx].txFailed,
               pUM->txq[idx].txDiscards);
    mdb_printf("   txRecycle:      %-12d txCopied:       %d\n",
               pUM->txq[idx].txRecycle,
               pUM->txq[idx].txCopied);
    mdb_printf("   txBlocked:      %-12d txWait:         %d\n",
               pUM->txq[idx].txBlocked,
               pUM->txq[idx].txWait);

    {
        u32_t pktCnt;
        u32_t pktFree;
        u32_t pktWait;
        u32_t pktSent;
        u32_t pktPend;

        pktCnt = (idx == FCOE_CID(pLM)) ?
                     pUM->devParams.numRxDesc[LM_CLI_IDX_FCOE] :
                     pUM->devParams.numRxDesc[LM_CLI_IDX_NDIS];

        pktFree = s_list_entry_cnt(&pUM->txq[idx].freeTxDescQ);
        pktWait = s_list_entry_cnt(&pUM->txq[idx].waitTxDescQ);
        pktSent = s_list_entry_cnt(&pUM->txq[idx].sentTxQ);
        pktPend = (pktCnt - (pktWait + pktSent + pktFree));

        mdb_printf("   - Descs: %d free, %d wait, %d sent, %d pend (%d)\n",
                   pktFree, pktWait, pktSent, pktPend,
                   (pktFree + pktWait + pktSent + pktPend));
        mdb_printf("   - BDs:   %d pend (chain %d used)\n",
                   (pTxq->bd_chain.prod_idx - pTxq->bd_chain.cons_idx),
                   (pTxq->bd_chain.capacity - pTxq->bd_chain.bd_left));
    }
}


static void BnxeMdbFcoe(um_device_t *      pUM,
                        bnxe_chain_idx_t * pIdxs)
{
    lm_device_t *    pLM  = (lm_device_t *)&pUM->lm_dev;
    d_list_entry_t * pNode;
    BnxeFcoeState    fcoe_state;

    mdb_printf("-- FCoE --\n");

    if (!BNXE_FCOE(pUM))
    {
        mdb_printf("   Not supported\n");
        return;
    }

    mdb_printf("   Attached:         %p\n",
               pUM->fcoe.pDev);
    mdb_printf("   FCoE MacAddr:     %02x:%02x:%02x:%02x:%02x:%02x\n",
               pLM->hw_info.fcoe_mac_addr[0], pLM->hw_info.fcoe_mac_addr[1],
               pLM->hw_info.fcoe_mac_addr[2], pLM->hw_info.fcoe_mac_addr[3],
               pLM->hw_info.fcoe_mac_addr[4], pLM->hw_info.fcoe_mac_addr[5]);

    //if (!CLIENT_BOUND(pUM, LM_CLI_IDX_FCOE)) return;

    mdb_printf("   initWqeTx:        %-12d initWqeTxErr:        %d\n",
               pUM->fcoe.stats.initWqeTx,
               pUM->fcoe.stats.initWqeTxErr);
    mdb_printf("   initCqeRx:        %-12d initCqeRxErr:        %d\n",
               pUM->fcoe.stats.initCqeRx,
               pUM->fcoe.stats.initCqeRxErr);
    mdb_printf("   offloadConnWqeTx: %-12d offloadConnWqeTxErr: %d\n",
               pUM->fcoe.stats.offloadConnWqeTx,
               pUM->fcoe.stats.offloadConnWqeTxErr);
    mdb_printf("   offloadConnCqeRx: %-12d offloadConnCqeRxErr: %d\n",
               pUM->fcoe.stats.offloadConnCqeRx,
               pUM->fcoe.stats.offloadConnCqeRxErr);
    mdb_printf("   enableConnWqeTx:  %-12d enableConnWqeTxErr:  %d\n",
               pUM->fcoe.stats.enableConnWqeTx,
               pUM->fcoe.stats.enableConnWqeTxErr);
    mdb_printf("   enableConnCqeRx:  %-12d enableConnCqeRxErr:  %d\n",
               pUM->fcoe.stats.enableConnCqeRx,
               pUM->fcoe.stats.enableConnCqeRxErr);
    mdb_printf("   disableConnWqeTx: %-12d disableConnWqeTxErr: %d\n",
               pUM->fcoe.stats.disableConnWqeTx,
               pUM->fcoe.stats.disableConnWqeTxErr);
    mdb_printf("   disableConnCqeRx: %-12d disableConnCqeRxErr: %d\n",
               pUM->fcoe.stats.disableConnCqeRx,
               pUM->fcoe.stats.disableConnCqeRxErr);
    mdb_printf("   destroyConnWqeTx: %-12d destroyConnWqeTxErr: %d\n",
               pUM->fcoe.stats.destroyConnWqeTx,
               pUM->fcoe.stats.destroyConnWqeTxErr);
    mdb_printf("   destroyConnCqeRx: %-12d destroyConnCqeRxErr: %d\n",
               pUM->fcoe.stats.destroyConnCqeRx,
               pUM->fcoe.stats.destroyConnCqeRxErr);
    mdb_printf("   destroyWqeTx:     %-12d destroyWqeTxErr:     %d\n",
               pUM->fcoe.stats.destroyWqeTx,
               pUM->fcoe.stats.destroyWqeTxErr);
    mdb_printf("   destroyCqeRx:     %-12d destroyCqeRxErr:     %d\n",
               pUM->fcoe.stats.destroyCqeRx,
               pUM->fcoe.stats.destroyCqeRxErr);
    mdb_printf("   compRequestCqeRx: %-12d compRequestCqeRxErr: %d\n",
               pUM->fcoe.stats.compRequestCqeRx,
               pUM->fcoe.stats.compRequestCqeRxErr);
    mdb_printf("   statWqeTx:        %-12d statWqeTxErr:        %d\n",
               pUM->fcoe.stats.statWqeTx,
               pUM->fcoe.stats.statWqeTxErr);
    mdb_printf("   statCqeRx:        %-12d statCqeRxErr:        %d\n",
               pUM->fcoe.stats.statCqeRx,
               pUM->fcoe.stats.statCqeRxErr);

    mdb_printf("\n");

    BnxeMdbChain(pUM, pIdxs, FCOE_CID(&pUM->lm_dev));

    mdb_printf("\n");

    mdb_printf("-- Offloaded FCoE Connections (%d) --\n",
               d_list_entry_cnt(&pLM->fcoe_info.run_time.fcoe_list));

    pNode = d_list_peek_head(&pLM->fcoe_info.run_time.fcoe_list);

    while (pNode)
    {
        if (mdb_vread(&fcoe_state, sizeof(BnxeFcoeState), (uintptr_t)pNode) == -1)
        {
            mdb_printf("ERROR: Unable to read FCoE state\n");
            return;
        }

        mdb_printf("   * %p: cid: %d / fcoe_conn_id: %d\n",
                   pNode,
                   fcoe_state.lm_fcoe.cid,
                   fcoe_state.lm_fcoe.fcoe_conn_id);
        mdb_printf("       ctx_virt:    %016p ctx_phys:    %016p\n",
                   fcoe_state.lm_fcoe.ctx_virt,
                   fcoe_state.lm_fcoe.ctx_phys.as_ptr);

        pNode = d_list_next_entry((d_list_entry_t *)&fcoe_state);
    }
}


static void BnxeMdbHelp(void)
{
    mdb_printf("addr::  dev_info address ('::devbindings -q bnxe')\n");
    mdb_printf("-i #    driver instance # (instead of 'addr::')\n");
    mdb_printf("-a      show all (basic, interrupts, all chains, fcoe)\n");
    mdb_printf("-b      basic status (default)\n");
    mdb_printf("-r      interrupt status\n");
    mdb_printf("-l      mutex locks status\n");
    mdb_printf("-c #    status for chain # (-1 for all chains)\n");
    mdb_printf("-f      fcoe status\n");
}


static int BnxeMdb(uintptr_t         addr,
                   uint_t            flags,
                   int               argc,
                   const mdb_arg_t * argv)
{
    struct dev_info    dip;
    um_device_t *      pUM;
    bnxe_chain_idx_t * pIdxs;
    struct devnames    dn;
    uintptr_t          dn_addr;
    major_t            major;
    uint_t             opt_a = 0;
    uint_t             opt_b = 0;
    uint_t             opt_r = 0;
    uint_t             opt_l = 0;
    uint_t             opt_f = 0;
    uint64_t           opt_i = OPT_NONE;
    uint64_t           opt_c = OPT_NONE;
    int idx;

    if (!(flags & DCMD_ADDRSPEC))
    {
        if (mdb_name_to_major("bnxe", &major) != 0)
        {
            mdb_printf("ERROR: Failed to get major number for bnxe\n");
            return DCMD_ERR;
        }

        if (major_to_addr(major, &dn_addr) != 0)
        {
            mdb_printf("ERROR: Failed to get address for bnxe\n");
            return DCMD_ERR;
        }

        if (mdb_vread(&dn, sizeof(struct devnames), dn_addr) == -1)
        {
            mdb_printf("ERROR: Couldn't read devnames array at %p\n",
                       dn_addr);
            return DCMD_ERR;
        }

        if (mdb_pwalk_dcmd("devi_next", "bnxe", argc, argv,
                           (uintptr_t)dn.dn_head) != 0)
        {
            mdb_printf("ERROR: Couldn't walk the devinfo chain at %p\n",
                       dn.dn_head);
            return DCMD_ERR;
        }

        return DCMD_OK;
    }

    if (mdb_vread(&dip, sizeof(struct dev_info), addr) == -1)
    {
        mdb_printf("ERROR: Unable to read dev_info\n");
        return DCMD_ERR;
    }

    if ((pUM = (um_device_t *)mdb_alloc(sizeof(um_device_t),
                                        UM_NOSLEEP)) == NULL)
    {
        mdb_printf("ERROR: Unable to alloc cache for um_device_t");
        return DCMD_ERR;
    }

    if ((pIdxs = (bnxe_chain_idx_t *)mdb_alloc(sizeof(bnxe_chain_idx_t),
                                               UM_NOSLEEP)) == NULL)
    {
        mdb_printf("ERROR: Unable to alloc cache for bnxe_chain_idx_t");
        mdb_free(pUM, sizeof(um_device_t));
        return DCMD_ERR;
    }

    if (mdb_vread(pUM, sizeof(um_device_t),
                  (uint64_t)dip.devi_driver_data) == -1)
    {
        mdb_printf("ERROR: Unable to read um_device_t\n");
        mdb_free(pIdxs, sizeof(bnxe_chain_idx_t));
        mdb_free(pUM, sizeof(um_device_t));
        return DCMD_ERR;
    }

    BnxeMdbGetChainIdxs(pUM, pIdxs);

    if (pUM->magic != BNXE_MAGIC)
    {
        mdb_printf("ERROR: Invalid dip address for a bnxe instance\n");
        mdb_free(pIdxs, sizeof(bnxe_chain_idx_t));
        mdb_free(pUM, sizeof(um_device_t));
        return DCMD_ERR;
    }

    if (mdb_getopts(argc, argv,
                    'i', MDB_OPT_UINT64, &opt_i,
                    'a', MDB_OPT_SETBITS, 1, &opt_a,
                    'b', MDB_OPT_SETBITS, 1, &opt_b,
                    'r', MDB_OPT_SETBITS, 1, &opt_r,
                    'l', MDB_OPT_SETBITS, 1, &opt_l,
                    'f', MDB_OPT_SETBITS, 1, &opt_f,
                    'c', MDB_OPT_UINT64, &opt_c) != argc)
    {
        mdb_free(pIdxs, sizeof(bnxe_chain_idx_t));
        mdb_free(pUM, sizeof(um_device_t));
        return DCMD_USAGE;
    }

    if ((opt_i != OPT_NONE) && (opt_i != pUM->instance))
    {
        mdb_free(pIdxs, sizeof(bnxe_chain_idx_t));
        mdb_free(pUM, sizeof(um_device_t));
        return DCMD_OK;
    }

    if (opt_a && (opt_b ||
                  opt_r ||
                  opt_l ||
                  opt_f ||
                  (opt_c != OPT_NONE)))
    {
        mdb_free(pIdxs, sizeof(bnxe_chain_idx_t));
        mdb_free(pUM, sizeof(um_device_t));
        return DCMD_USAGE;
    }

    if (!(opt_a ||
          opt_b ||
          opt_r ||
          opt_l ||
          opt_f ||
          (opt_c != OPT_NONE)))
    {
        opt_b = 1;
    }

    mdb_printf("--------------------------------------------------\n");
    mdb_printf("\n");
    mdb_printf("%s -> %p (%p)\n", pUM->devName, addr, dip.devi_driver_data);
    mdb_printf("\n");

    if (opt_a || opt_b)
    {
        BnxeMdbBasic(pUM);
        mdb_printf("\n");

        BnxeMdbSb(pUM, DEF_STATUS_BLOCK_INDEX);
        mdb_printf("\n");
    }

    if (opt_a || opt_r)
    {
        BnxeMdbIntr(pUM);
        mdb_printf("\n");
    }

    if (opt_a || opt_l)
    {
        BnxeMdbLocks(pUM, (um_device_t *)dip.devi_driver_data);
        mdb_printf("\n");
    }

    if (opt_a || (opt_c != OPT_NONE))
    {
        if (opt_a || (opt_c == (uint64_t)-1))
        {
            LM_FOREACH_RSS_IDX(&pUM->lm_dev, idx)
            {
                BnxeMdbChain(pUM, pIdxs, idx);
                mdb_printf("\n");
            }

            if (BNXE_FCOE(pUM))
            {
                BnxeMdbChain(pUM, pIdxs, FCOE_CID(&pUM->lm_dev));
            }
        }
        else
        {
            if ((opt_c < pUM->lm_dev.params.rss_chain_cnt) ||
                (opt_c == FCOE_CID(&pUM->lm_dev)))
            {
                BnxeMdbChain(pUM, pIdxs, opt_c);
            }
            else
            {
                mdb_printf("ERROR: Invalid chain index (%d)\n", opt_c);
            }
        }

        mdb_printf("\n");
    }

    if (opt_a || opt_f)
    {
        BnxeMdbFcoe(pUM, pIdxs);
        mdb_printf("\n");
    }

    mdb_printf("--------------------------------------------------\n");

    mdb_free(pIdxs, sizeof(bnxe_chain_idx_t));
    mdb_free(pUM, sizeof(um_device_t));

    return DCMD_OK;
}


static void BnxeMdbMemGetAll(um_device_t * pUM,
                             int           verbose)
{
    d_list_entry_t * pNode;
    BnxeMemRegion    memRegion;
    BnxeMemDma       memDma;
    BnxeMemBlock     memBlock;

    mdb_printf("---> %d BAR Allocations (BnxeMemRegion) <---\n",
               d_list_entry_cnt(&pUM->memRegionList));

    pNode = d_list_peek_head(&pUM->memRegionList);
    while (pNode)
    {
        if (mdb_vread(&memRegion, sizeof(BnxeMemRegion),
                      (uintptr_t)pNode) == -1)
        {
            mdb_printf("ERROR: Unable to read region struct\n");
            return;
        }

        mdb_printf("(%016p) bar:%02d virt:%016p phys:%016p offset:%d size:%d\n",
                   pNode,
                   memRegion.regNumber,
                   memRegion.pRegAddr,
                   memRegion.baseAddr.as_ptr,
                   memRegion.offset,
                   memRegion.size);

        pNode = d_list_next_entry((d_list_entry_t *)&memRegion);
    }

    mdb_printf("---> %d DMA Allocations (BnxeMemDma) <---\n",
               d_list_entry_cnt(&pUM->memDmaList));

    pNode = d_list_peek_head(&pUM->memDmaList);
    while (pNode)
    {
        if (mdb_vread(&memDma, sizeof(BnxeMemDma), (uintptr_t)pNode) == -1)
        {
            mdb_printf("ERROR: Unable to read dma struct\n");
            return;
        }

        mdb_printf("(%016p) virt:%016p phys:%016p size:%d\n",
                   pNode,
                   memDma.pDmaVirt,
                   memDma.physAddr.as_ptr,
                   memDma.size);
        if (verbose)
        {
            mdb_printf(" - %s:%d\n", memDma.fileName, memDma.fileLine);
        }

        pNode = d_list_next_entry((d_list_entry_t *)&memDma);
    }

    mdb_printf("---> %d Heap Allocations (BnxeMemBlock) <---\n",
               d_list_entry_cnt(&pUM->memBlockList));

    pNode = d_list_peek_head(&pUM->memBlockList);
    while (pNode)
    {
        if (mdb_vread(&memBlock, sizeof(BnxeMemBlock), (uintptr_t)pNode) == -1)
        {
            mdb_printf("ERROR: Unable to read block struct\n");
            return;
        }

        mdb_printf("(%016p) virt:%016p size:%d\n",
                   pNode,
                   memBlock.pBuf,
                   memBlock.size);
        if (verbose)
        {
            mdb_printf(" - %s:%d\n", memBlock.fileName, memBlock.fileLine);
        }

        pNode = d_list_next_entry((d_list_entry_t *)&memBlock);
    }
}


static void BnxeMdbMemGet(um_device_t * pUM,
                          int           phys,
                          int           verbose,
                          u_longlong_t  addr)
{
    d_list_entry_t * pNode;
    BnxeMemRegion    memRegion;
    BnxeMemDma       memDma;
    BnxeMemBlock     memBlock;
    int found = 0;

    pNode = d_list_peek_head(&pUM->memRegionList);
    while (pNode)
    {
        if (mdb_vread(&memRegion, sizeof(BnxeMemRegion),
                      (uintptr_t)pNode) == -1)
        {
            mdb_printf("ERROR: Unable to read region struct\n");
            return;
        }

        if (((phys) &&
             (addr >= (u_longlong_t)memRegion.baseAddr.as_u64) &&
             (addr < ((u_longlong_t)memRegion.baseAddr.as_u64 + memRegion.size))) ||
            ((addr >= (u_longlong_t)memRegion.pRegAddr) &&
             (addr < ((u_longlong_t)memRegion.pRegAddr + memRegion.size))))
        {
            mdb_printf("(%016p) bar:%02d virt:%016p phys:%016p offset:%d size:%d\n",
                       pNode,
                       memRegion.regNumber,
                       memRegion.pRegAddr,
                       memRegion.baseAddr.as_ptr,
                       memRegion.offset,
                       memRegion.size);
            found = 1;
        }

        pNode = d_list_next_entry((d_list_entry_t *)&memRegion);
    }

    if (found)
    {
        return;
    }

    pNode = d_list_peek_head(&pUM->memDmaList);
    while (pNode)
    {
        if (mdb_vread(&memDma, sizeof(BnxeMemDma), (uintptr_t)pNode) == -1)
        {
            mdb_printf("ERROR: Unable to read dma struct\n");
            return;
        }

        if (((phys) &&
             (addr >= (u_longlong_t)memDma.physAddr.as_u64) &&
             (addr < ((u_longlong_t)memDma.physAddr.as_u64 + memDma.size))) ||
            ((addr >= (u_longlong_t)memDma.pDmaVirt) &&
             (addr < ((u_longlong_t)memDma.pDmaVirt + memDma.size))))
        {
            mdb_printf("(%016p) virt:%016p phys:%016p size:%d\n",
                       pNode,
                       memDma.pDmaVirt,
                       memDma.physAddr.as_ptr,
                       memDma.size);
            if (verbose)
            {
                mdb_printf(" - %s:%d\n", memDma.fileName, memDma.fileLine);
            }

            return;
        }

        pNode = d_list_next_entry((d_list_entry_t *)&memDma);
    }

    if (phys)
    {
        mdb_printf("Not found!\n");
        return;
    }

    pNode = d_list_peek_head(&pUM->memBlockList);
    while (pNode)
    {
        if (mdb_vread(&memBlock, sizeof(BnxeMemBlock), (uintptr_t)pNode) == -1)
        {
            mdb_printf("ERROR: Unable to read block struct\n");
            return;
        }

        if ((addr >= (u_longlong_t)memBlock.pBuf) &&
            (addr < ((u_longlong_t)memBlock.pBuf + memBlock.size)))
        {
            mdb_printf("(%016p) virt:%016p size:%d\n",
                       pNode,
                       memBlock.pBuf,
                       memBlock.size);
            if (verbose)
            {
                mdb_printf(" - %s:%d\n", memBlock.fileName, memBlock.fileLine);
            }

            return;
        }

        pNode = d_list_next_entry((d_list_entry_t *)&memBlock);
    }

    mdb_printf("Not found!\n");
}


static void BnxeMdbMemHelp(void)
{
    mdb_printf("addr::  dev_info address ('::devbindings -q bnxe')\n");
    mdb_printf("-i #    driver instance # (instead of 'addr::')\n");
    mdb_printf("-p      address is a physical address\n");
    mdb_printf("-v      show function and line number of allocation\n");
}


static int BnxeMdbMem(uintptr_t         addr,
                      uint_t            flags,
                      int               argc,
                      const mdb_arg_t * argv)
{
    struct dev_info dip;
    um_device_t *   pUM;
    struct devnames dn;
    uintptr_t       dn_addr;
    major_t         major;
    uint_t   opt_p = 0;
    uint_t   opt_v = 0;
    uint64_t opt_i = OPT_NONE;
    int idx;

    if (!(flags & DCMD_ADDRSPEC))
    {
        if (mdb_name_to_major("bnxe", &major) != 0)
        {
            mdb_printf("ERROR: Failed to get major number for bnxe\n");
            return DCMD_ERR;
        }

        if (major_to_addr(major, &dn_addr) != 0)
        {
            mdb_printf("ERROR: Failed to get address for bnxe\n");
            return DCMD_ERR;
        }

        if (mdb_vread(&dn, sizeof(struct devnames), dn_addr) == -1)
        {
            mdb_printf("ERROR: Couldn't read devnames array at %p\n",
                       dn_addr);
            return DCMD_ERR;
        }

        if (mdb_pwalk_dcmd("devi_next", "bnxe_mem", argc, argv,
                           (uintptr_t)dn.dn_head) != 0)
        {
            mdb_printf("ERROR: Couldn't walk the devinfo chain at %p\n",
                       dn.dn_head);
            return DCMD_ERR;
        }

        return DCMD_OK;
    }

    if (mdb_vread(&dip, sizeof(struct dev_info), addr) == -1)
    {
        mdb_printf("ERROR: Unable to read dev_info\n");
        return DCMD_ERR;
    }

    if ((pUM = (um_device_t *)mdb_alloc(sizeof(um_device_t),
                                        UM_NOSLEEP)) == NULL)
    {
        mdb_printf("ERROR: Unable to alloc cache for um_device_t");
        return DCMD_ERR;
    }

    if (mdb_vread(pUM, sizeof(um_device_t),
                  (uint64_t)dip.devi_driver_data) == -1)
    {
        mdb_printf("ERROR: Unable to read um_device_t\n");
        mdb_free(pUM, sizeof(um_device_t));
        return DCMD_ERR;
    }

    if (pUM->magic != BNXE_MAGIC)
    {
        mdb_printf("ERROR: Invalid dip address for a bnxe instance\n");
        mdb_free(pUM, sizeof(um_device_t));
        return DCMD_ERR;
    }

    idx = mdb_getopts(argc, argv,
                      'i', MDB_OPT_UINT64, &opt_i,
                      'p', MDB_OPT_SETBITS, 1, &opt_p,
                      'v', MDB_OPT_SETBITS, 1, &opt_v);

    if (idx < (argc - 1))
    {
        mdb_free(pUM, sizeof(um_device_t));
        return DCMD_USAGE;
    }

    if ((opt_i != OPT_NONE) && (opt_i != pUM->instance))
    {
        mdb_free(pUM, sizeof(um_device_t));
        return DCMD_OK;
    }

    mdb_printf("%s -> %p (%p)\n", pUM->devName, addr, dip.devi_driver_data);

    if (idx == argc)
    {
        BnxeMdbMemGetAll(pUM, opt_v);
    }
    else /* (idx == (argc - 1)) */
    {
        if (argv[idx].a_type != MDB_TYPE_STRING)
        {
            mdb_free(pUM, sizeof(um_device_t));
            return DCMD_OK;
        }

        BnxeMdbMemGet(pUM, opt_p, opt_v, mdb_strtoull(argv[idx].a_un.a_str));
    }

    mdb_free(pUM, sizeof(um_device_t));
    return DCMD_OK;
}


static const mdb_dcmd_t bnxe_dcmds[] =
{
    { "bnxe", "?[ -i # ] [ -a | -b | -r | -l | -c # | -f ]", "bnxe driver status", BnxeMdb, BnxeMdbHelp },
    { "bnxe_mem", "?[ -i # ] [ -p ] [ -v ] [ address ]", "bnxe memory allocation info", BnxeMdbMem, BnxeMdbMemHelp },
    { NULL }
};


static const mdb_walker_t bnxe_walkers[] =
{
    NULL
};


static mdb_modinfo_t bnxe_mdb =
{
    MDB_API_VERSION,
    bnxe_dcmds,
    bnxe_walkers
};


mdb_modinfo_t * _mdb_init(void)
{
    return &bnxe_mdb;
}


void _mdb_fini(void)
{
    return;
}



/*************************************************************************/
/*  The following aren't used but must be defined for bnxe_mdb to load.  */
/*************************************************************************/

void mm_memset(void * pBuf,
               u8_t   val,
               u32_t  memSize)
{
    memset(pBuf, val, memSize);
}


void mm_memcpy(void *       pDest,
               const void * pSrc,
               u32_t        memSize)
{
    memcpy(pDest, pSrc, memSize);
}


u8_t mm_memcmp(void * pBuf1,
               void * pBuf2,
               u32_t  count)
{
    return (memcmp(pBuf1, pBuf2, count) == 0) ? 1 : 0;
}

void * mm_rt_alloc_mem_imp(lm_device_t *       pDev,
                           u32_t               memSize,
                           const char *        sz_file,
                           const unsigned long line,
                           u8_t                cli_idx)
{
    return NULL;
}

u32_t lm_mp_get_reg_chain_from_chain(struct _lm_device_t * pdev,
                                     u32_t                 chain)
{
    return 0;
}

