
/*
 * The information contained in this file is confidential and proprietary to
 * Broadcom Corporation.  No part of this file may be reproduced or
 * distributed, in any form or by any means for any purpose, without the
 * express written permission of Broadcom Corporation.
 *
 * (c) COPYRIGHT 2008-2012 Broadcom Corporation, ALL RIGHTS RESERVED.
 */

#ifndef BNXEF_H
#define BNXEF_H

#include <sys/types.h>
#include <sys/stream.h>
#include <sys/stropts.h>
#include <sys/errno.h>
#include <sys/cred.h>
#include <sys/poll.h>
#include <sys/modctl.h>
#include <sys/stat.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>
#include <sys/ddifm.h>
#include <sys/fm/protocol.h>
#include <sys/fm/util.h>
#include <sys/fm/io/ddi.h>
#include <sys/pattr.h>
#include <sys/sysmacros.h>
#include <sys/strsun.h>
#include <sys/ksynch.h>
#include <sys/kstat.h>
#include <sys/strsubr.h>

#include "bcmtype.h"
#include "bnxe_binding.h"
#include "bnxef_debug.h"

#define USER_OPTION_GOOBAR_MIN      1
#define USER_OPTION_GOOBAR_MAX      2048
#define USER_OPTION_GOOBAR_DEFAULT  1024

#define BNXEF_MAGIC          0xfc0efc0e
#define BNXEF_STR_SIZE       32

#ifdef __sparc
#define BNXEF_DMA_ALIGNMENT  0x2000UL
#else
#define BNXEF_DMA_ALIGNMENT  0x1000UL
#endif


typedef struct _BnxefDevParams
{
    boolean_t logEnable;
    boolean_t foobar;
    u32_t     goobar;
} BnxefDevParams;


typedef struct _fcoe_device
{
    u32_t                 magic;
    dev_info_t *          pDev;

    int                   instance;
    char                  devName[BNXEF_STR_SIZE];
    char                  version[BNXEF_STR_SIZE];

    BnxefDevParams        devParams;

    BnxeBinding           bnxe;

    kstat_t *             kstats;
    kmutex_t              kstatMutex;

    int                   fmCapabilities; /* FMA capabilities */

} fcoe_device_t;


char * BnxefDevName(void * pDev);

void BnxefCfgInit(fcoe_device_t * pFCOE);

#if 0
int  BnxefHwStart(fcoe_device_t * pFCOE);
void BnxefHwStop(fcoe_device_t * pFCOE);
void BnxefUpdatePhy(fcoe_device_t * pFCOE,
                   boolean_t     cfgMap);
int BnxefHwResume(fcoe_device_t * pFCOE);
int BnxefHwSuspend(fcoe_device_t * pFCOE);
#if (DEVO_REV > 3)
int  BnxefHwQuiesce(fcoe_device_t * pFCOE);
#endif
#endif

boolean_t BnxefKstatInit(fcoe_device_t * pFCOE);
void      BnxefKstatFini(fcoe_device_t * pFCOE);

int  BnxefCheckAccHandle(ddi_acc_handle_t handle);
int  BnxefCheckDmaHandle(ddi_dma_handle_t handle);
void BnxefFmErrorReport(fcoe_device_t * pFCOE, char * detail);

int BnxefCliCtl(dev_info_t * pDev, int cmd, void * pData, int dataLen);
int BnxefCliIndicateTx(dev_info_t * pDev, mblk_t * pMblk);
int BnxefCliIndicateRx(dev_info_t * pDev, mblk_t * pMblk);
int BnxefCliIndicateCqes(dev_info_t * pDev, void * cqes[], int cqeCnt);

extern kmutex_t bnxefLoaderMutex;

#define BNXEF_LOCK_ENTER_STATS(pFCOE)  mutex_enter(&(pFCOE)->kstatMutex)
#define BNXEF_LOCK_EXIT_STATS(pFCOE)   mutex_exit(&(pFCOE)->kstatMutex)

#endif /* BNXEF_H */

