/*
 * The information contained in this file is confidential and proprietary to
 * Broadcom Corporation.  No part of this file may be reproduced or
 * distributed, in any form or by any means for any purpose, without the
 * express written permission of Broadcom Corporation.
 *
 * (c) COPYRIGHT 2010-2012 Broadcom Corporation, ALL RIGHTS RESERVED.
 */

#pragma ident "(C) 2010-2012 Broadcom Corporation"
#pragma ident "$File$$Revision$$Change$$DateTime$"

#ifndef __BNXEF_FCF_INFO_H__
#define	__BNXEF_FCF_INFO_H__

#include <sys/types.h>
#include <sys/conf.h>
#include <sys/stream.h>
#include <sys/stropts.h>
#include <sys/errno.h>
#include <sys/cred.h>
#include <sys/poll.h>
#include <sys/modctl.h>
#include <sys/stat.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>
#include <sys/ddifm.h>
#include <sys/fm/protocol.h>
#include <sys/fm/util.h>
#include <sys/fm/io/ddi.h>
#include <sys/pattr.h>
#include <sys/sysmacros.h>
#include <sys/strsun.h>
#include <sys/ksynch.h>
#include <sys/kstat.h>
#include <sys/strsubr.h>
#include <sys/ksynch.h>
#include <sys/ethernet.h>


#define	BNXEF_WWN_LEN		8
/*
 *  This describes the FCF as discovered.
 */
typedef struct __fcf_info__ {
	struct __fcf_info__  *next;	/* Next Ptr for next FCF  disc */
	kmutex_t  fcf_mutex;	/* To Guard updation of state in Flags */
	uint8_t	  fcf_state;	/* Indicate selected for Login, etc */
#define	FCF_INFO_DISC_ADV_RECEIVED	0x01
#define	FCF_INFO_VLAN_NOTIF_RECEIVED	0x02
#define	FCF_SELECTED_FOR_LOGIN	0x04	/* This FCF is selected for login */

	uint8_t	 mac_addr[ETHERADDRL]; /* FCF MAC addr described */
	uint8_t	 hp_ffa_flag;
	uint8_t fcoe_mac_addr[ETHERADDRL]; /* For HP FFA */
	uint16_t vlan_id;	/* VLAN ID on which it was disc */
	uint8_t	 vlan_tag_pcp;	/* Priority this FCF used */
	uint8_t	 vlan_tag_cfi;	/* should be zero */
	uint16_t  flags_sp:1;	/* Flags, SP, FP, A, S & F */
	uint16_t  flags_fp:1;
	uint16_t  flags_a:1;
	uint16_t  flags_s:1;
	uint16_t  flags_f:1;
	uint16_t  reserved:11;
	uint8_t	  priority;	/* pri described */
	uint8_t	name[BNXEF_WWN_LEN];	/* NAme described */
	uint8_t	fabric_name[BNXEF_WWN_LEN]; /* Fabric name described */
	uint16_t	vf_id;	/* Virtual Fabric ID */
	uint32_t	fc_map;	/* FC Map described */
	uint32_t	fka_adv_period;	/* In msec */
	uint8_t		max_fcoe_size_verified;   /* set to 0, changed to 1 */
	uint8_t	available_for_login;	/* Available for login A bit */
	uint16_t	d_bit_info;	/* Indicates the d_bit only */
} bnxef_fcf_info_t;

/*
 * The following are the flags that can be supported
 */
#define	FIP_F_BIT	0x0001
#define	FIP_S_BIT	0x0002	/*  Solicted Unicast Disc Advertisement */
#define	FIP_A_BIT	0x0004	/* Available for login, only from FCF */
#define	FIP_FP_BIT	0x8000	/* FP bit indicating FPMA request or Support */
#define	FIP_SP_BIT	0x4000	/* SP Bit for SPMA request or support */

/*
 *  List of FCF identified using this ENODE MAC
 */
typedef  struct __bnxef_fcf_list__ {
	kmutex_t	fcflist_mutex; /* Guard for traversal */
	bnxef_fcf_info_t *head; /* This will point to first item */
	uint32_t	num_items;	/* Num items in the list */
	bnxef_fcf_info_t *tail;	/* Gives the last items, for linking */
} bnxef_fcf_list_t;


#endif
