/*
 * The information contained in this file is confidential and proprietary to
 * Broadcom Corporation.  No part of this file may be reproduced or
 * distributed, in any form or by any means for any purpose, without the
 * express written permission of Broadcom Corporation.
 *
 * (c) COPYRIGHT 2010-2012 Broadcom Corporation, ALL RIGHTS RESERVED.
 */

#pragma ident "(C) 2010-2012 Broadcom Corporation"
#pragma ident "$File$$Revision$$Change$$DateTime$"

#ifndef __BNXEF_FCOE_SESSION_H__
#define	__BNXEF_FCOE_SESSION_H__

#include <sys/types.h>
#include <sys/conf.h>
#include <sys/stream.h>
#include <sys/stropts.h>
#include <sys/errno.h>
#include <sys/cred.h>
#include <sys/poll.h>
#include <sys/modctl.h>
#include <sys/stat.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>
#include <sys/ddifm.h>
#include <sys/fm/protocol.h>
#include <sys/fm/util.h>
#include <sys/fm/io/ddi.h>
#include <sys/pattr.h>
#include <sys/sysmacros.h>
#include <sys/strsun.h>
#include <sys/ksynch.h>
#include <sys/kstat.h>
#include <sys/strsubr.h>
#include <sys/ksynch.h>
#include <sys/ethernet.h>

/*
 * The following provides the information about the fc_unsol_buf_t allocated
 * for leadville stack. This structure includes the meta data as well as the
 * fc_unsol_buf_t structure pointer. The pointer to this meta data structure
 * will be returned as the token when the fca_ub_alloc is called.
 */
typedef struct __bnxef_unsol_buf__ {
	uint8_t	type;				/* FC-4 type */
	uint8_t	 flags;				/* ALLOC, RELEASED */
	kmutex_t  unsol_mutex;			/* To protect flags update */
	struct __bnxef_unsol_buf__ *nextp;	/* Next such buffer */
	fc_unsol_buf_t	 *unsol_bufp;	/* allocated based on size req */
	void	 *portp;		/* BAck pointer to port */
} bnxef_unsol_buf_t;

#define	BNXEF_UB_ALLOCATED		0x01
#define	BNXEF_UB_RELEASED		0x02

/*
 *  This provides the list of unsolicited buffers allocated until freed.
 */
typedef struct __bnxefc_unsol_buf_list__ {
	kmutex_t	unsol_buf_list_mutex; /* Guard for traversal */
	bnxef_unsol_buf_t *head; /* This will point to first session */
	uint32_t	num_items;	/* Num items in the list */
	bnxef_unsol_buf_t *tail;	/* Gives the last items, for linking */
} bnxef_unsol_buf_list_t;

/* Different stages of SQ allocation */
#define	BNXEF_SQ_DMA_MEM_BOUND		0x01
#define	BNXEF_SQ_DMA_MEM_ALLOCATED	0x02
#define	BNXEF_SQ_RES_DMA_HANDLE_ALLOCATED  0x04
#define	BNXEF_SQ_RES_ALL		 0x7

/* Different stages for CQ allocation */
#define	BNXEF_CQ_DMA_MEM_BOUND		0x01
#define	BNXEF_CQ_DMA_MEM_ALLOCATED	0x02
#define	BNXEF_CQ_RES_DMA_HANDLE_ALLOCATED  0x04
#define	BNXEF_CQ_RES_ALL		0x07

/* Different stages for XFER Q allocation */
#define	BNXEF_XFERQ_DMA_MEM_BOUND		0x01
#define	BNXEF_XFERQ_DMA_MEM_ALLOCATED	0x02
#define	BNXEF_XFERQ_RES_DMA_HANDLE_ALLOCATED  0x04
#define	BNXEF_XFERQ_RES_ALL		0x07


/*
 * The following are the different stages for RQ allocation
 */
#define	BNXEF_RQ_PBL_MEM_BOUND			0x01
#define	BNXEF_RQ_PBL_MEM_ALLOCATED		0x02
#define	BNXEF_RQ_PBL_HANDLE_ALLOCATED		0x04
#define	BNXEF_RQ_PBL_SETUP			0x08
#define	BNXEF_RQ_DMA_MEM_BOUND			0x10
#define	BNXEF_RQ_DMA_MEM_ALLOCATED		0x20
#define	BNXEF_RQ_DMA_HANDLE_ALLOCATED		0x40
#define	BNXEF_RQ_ALL				0x7F
/*
 * The following are the different stages for CONFQ allocation
 */
#define	BNXEF_CONFQ_PBL_MEM_BOUND			0x01
#define	BNXEF_CONFQ_PBL_MEM_ALLOCATED			0x02
#define	BNXEF_CONFQ_PBL_HANDLE_ALLOCATED		0x04
#define	BNXEF_CONFQ_PBL_SETUP				0x08
#define	BNXEF_CONFQ_DMA_MEM_BOUND			0x10
#define	BNXEF_CONFQ_DMA_MEM_ALLOCATED			0x20
#define	BNXEF_CONFQ_DMA_HANDLE_ALLOCATED		0x40
#define	BNXEF_CONFQ_ALL					0x7F

/*
 * The following are the stages for LCQ resource allocation,
 */
#define	BNXEF_LCQ_DMA_MEM_BOUND				0x01
#define	BNXEF_LCQ_DMA_MEM_ALLOCATED			0x02
#define	BNXEF_LCQ_RES_DMA_HANDLE_ALLOCATED		0x04
#define	BNXEF_LCQ_RES_ALL				0x07

/*
 * The following are the stages for LCQ resource allocation,
 */
#define	BNXEF_CONN_DB_DMA_MEM_BOUND			0x01
#define	BNXEF_CONN_DB_DMA_MEM_ALLOCATED			0x02
#define	BNXEF_CONN_DB_RES_DMA_HANDLE_ALLOCATED		0x04
#define	BNXEF_CONN_DB_RES_ALL				0x07

/*
 * The following are the session resource allocation stages
 */
#define	BNXEF_SESSION_RESOURCE_LCQ_ALLOC_COMPLETED	0x01
#define	BNXEF_SESSION_RESOURCE_CONFQ_ALLOC_COMPLETED	0x02
#define	BNXEF_SESSION_RESOURCE_XFERQ_ALLOC_COMPLETED	0x04
#define	BNXEF_SESSION_RESOURCE_RQ_ALLOC_COMPLETED	0x08
#define	BNXEF_SESSION_RESOURCE_CQ_ALLOC_COMPLETED	0x10
#define	BNXEF_SESSION_RESOURCE_SQ_ALLOC_COMPLETED	0x20
#define	BNXEF_SESSION_RESOURCE_CONN_DB_ALLOC_COMPLETED	0x40
#define	BNXEF_SESSION_RESOURCE_ALL			0x7F

/*
 * Needed for mapping the door bell register.
 */
#define	BNXEF_571X_DB_PAGE_SIZE		128
#define	BNXEF_DPM_TRIGER_TYPE		0x40

#define	BNXEF_SLCP_MAX_IDX		200
#define	BNXEF_SESSION_OFFLOAD_TYPE	0x01
#define	BNXEF_SESSION_ENABLE_TYPE	0x02
#define	BNXEF_SESSION_DISABLE_TYPE	0x03
#define	BNXEF_SESSION_DESTROY_TYPE	0x04

typedef struct __bnxef_completion_log__ {
	uint16_t	 x_id;
	uint32_t	task_flags;
	void		 *pktp;
	clock_t		start_lbolt;
	clock_t		end_lbolt;
	void		*sessionp;
	uint8_t		session_state;
	uint8_t		pkt_state;
	uint8_t		pkt_reason;
	uint8_t		pkt_expln;
	uint8_t		pkt_action;
} bnxef_completion_log_t;

#define	BNXEF_MAX_LOG_ENTRIES	1024


/*
 * This specifies the complete session information and associated rings
 * allocated.
 */
typedef struct __bnxef_session_info__ {
	uint8_t		state; /* Indicates offloaded or not */
#define	BNXEF_SESSION_OFFLOADED		0x01
#define	BNXEF_SESSION_ENABLED		0x02
#define	BNXEF_SESSION_LPORT_SHUTDOWN_COMPLETE 0x04
#define	BNXEF_SESSION_CLEANUP_INPROGRESS	0x08
#define	BNXEF_TM_CLEANUP_IN_PROGRESS		0x10
#define	BNXEF_SESSION_SEQ_RETRY_ENABLED		0x20
#define	BNXEF_SESSION_CQ_PROCESSING_INPROGRESS 0x40
#define	BNXEF_SESSION_CQ_ARMING_REQUIRED	0x80
	uint8_t		sec_state;
#define	BNXEF_SESSION_PLOGI_COMPLETE	0x01
#define	BNXEF_SESSION_PRLI_COMPLETE	0x02
#define	BNXEF_UNSOL_LOGO_INPROGRESS	0x04
#define BNXEF_SESSION_PRLI_INPROGRESS	0x08
	struct  __bnxef_session_info__ *next;
	uint32_t	con_id;		/* Connection Id allocated at offload */
	uint32_t	context_id;	/* Context ID for Offloaded Session */
	void		 *lport;	/* local port ptr */
	void		  *rport;	/* remote port ptr */
	bnxef_task_list_t  sessiontasklist; /* Tasks belong to this session */
	kmutex_t	  session_mutex; /* Used to Guard all rings per conn */
	uint32_t	  session_req_error; /* Fw Session Req error */
#define	BNXEF_CONTEXT_ID_MISMATCH	0x01
	kcondvar_t	  session_cv;	 /* Wait for offload and enable */
	kcondvar_t	  session_els;	 /* Wait for Prev ELS to complete */
	uint32_t	  *doorbell_reg; /* Door Bell for this connection */
	ddi_acc_handle_t  doorbell_acc;	 /* Access handle for door bell bar */

	/* Now get the list of HST structures allocated */
	uint32_t		max_sqes; /* MAx entries possible in SQ */
	ddi_dma_handle_t	sq_dma_handle; /* dma handle for SQ */
	ddi_dma_cookie_t	sq_cookie;	/* Sq dma Cookie */
	caddr_t			sq_addr;	/* Virtual address */
	uint64_t		sq_length;	/* Actual length allocated */
	uint16_t		sq_prod_idx;	/* producer index */
	uint32_t		sq_alloc_sz;	/* Allocated mem size */
	uint8_t			sq_toggle;	/* Toggle Bit */
	ddi_acc_handle_t	sq_acc_handle; /* access handle */
	uint_t			sq_cookie_cnt;  /* Number of cookies returned */

	/* CQ */
	uint32_t		max_cqes;		/* MAximum CQ entries */
	ddi_dma_handle_t	cq_dma_handle;	/* Cq dma handle */
	ddi_dma_cookie_t	cq_cookie;		/* Cq dma cookie */
	uint_t			cq_cookie_cnt;	/* Cq cookie count */
	ddi_acc_handle_t	cq_acc_handle;	/* access handle */
	caddr_t			cq_addr;		/* Cq Virtual address */
	uint64_t		cq_length;	/* CQ actual length allocated */
	uint16_t		cq_cons_idx;		/* Cq Consumer Index */
	uint16_t		prev_cq_cons_idx;
	uint8_t			prev_cq_toggle;
	uint16_t		iterate_cnt;
	uint8_t			cq_toggle;		/* CQ Toggle Bit */

	/* RQ */
	uint32_t		max_rqes;		/* Max RQ entries */
	ddi_dma_handle_t	rq_dma_handle;	/* RQ dma handle */
	ddi_dma_cookie_t	rq_cookie;		/* RQ Dma Cookie */
	uint_t			rq_cookie_cnt;	/* RQ Cookie count */
	ddi_acc_handle_t	rq_acc_handle;	/* RQ Access handle */
	caddr_t			rq_addr;		/* RQ Start address */
	uint64_t		rq_length;		/* RQ Actual length */
	uint16_t		rq_prod_idx;		/* RQ prod idx */
	uint16_t		rq_cons_idx;		/* RQ Cons idx */

	/* RQ Page Buffer List */
	caddr_t			rq_pbl_addr;	/* RQ Page Buffer List Vaddr */
	uint64_t		rq_pbl_length;	/* Actual length allocated */
	ddi_dma_handle_t	rq_pbl_dma_handle;	/* RQ PBL dma handle */
	ddi_dma_cookie_t	rq_pbl_dma_cookie;	/* RQ PBL dma cookie */
	uint32_t		rq_pbl_cookie_cnt; /* RQ PBL cookie count */
	ddi_acc_handle_t	rq_pbl_acc_handle;	/* RQ access handle */


	/* XFERQ */
	ddi_dma_handle_t	xferq_dma_handle;	/* XFERQ dma handle */
	ddi_dma_cookie_t	xferq_cookie;	/* XFERQ dma cookie */
	uint32_t		xferq_cookie_cnt;	/* XFERQ Cookie count */
	caddr_t			xferq_addr;		/* XFERQ virtual addr */
	uint64_t		xferq_length;	/* XFERQ Actual length */
	ddi_acc_handle_t	xferq_acc_handle;	/* XFERQ Acc handle */
	uint32_t		xferq_sz;		/* XFERQ Alloc size */

	/* CONFQ */
	ddi_dma_handle_t	confq_dma_handle;	/* CONFQ dma handle */
	ddi_dma_cookie_t	confq_cookie;	/* CONFQ dma Cookie */
	uint32_t		confq_cookie_cnt;	/* ConfQ cookie count */
	caddr_t			confq_addr;		/* CONFQ Address */
	uint64_t		confq_length;	/* CONFQ length allocated */
	uint64_t		confq_sz;		/* CONFQ alloc Size */
	ddi_acc_handle_t	confq_acc_handle;	/* CONFQ acc handle */

	/* CONFQ PBL */
	ddi_dma_handle_t	confq_pbl_dma_handle; /* CONFQ PBL Dma handle */
	ddi_dma_cookie_t	confq_pbl_cookie; /* CONFQ PBL Dma Cookie */
	uint32_t		confq_pbl_cookie_cnt; /* CONFQPBL cookie cnt */
	caddr_t			confq_pbl_addr;	 /* CONFQ PBL Vaddr */
	uint64_t		confq_pbl_length; /* CONFQ PBL Actual length */
	uint64_t		confq_pbl_sz;	 /* CONFQ PBL size allocated */
	ddi_acc_handle_t	confq_pbl_acc_handle; /* Acchandle for memory */

	/* LCQ */
	ddi_dma_handle_t	lcq_dma_handle;	 /* LCQ Dma handle */
	ddi_dma_cookie_t	lcq_cookie;		 /* LCQ DMA cookie  */
	uint32_t		lcq_cookie_cnt;	 /* LCQ Cookie count */
	caddr_t			lcq_addr;	/* LCQ Address */
	uint64_t		lcq_length;		 /* LCQ Actual length */
	uint64_t		lcq_sz;		 /* Local Queue Size */
	ddi_acc_handle_t	lcq_acc_handle;	 /* Local Queue acc handle */

	/* Connection db */
	ddi_dma_handle_t	conn_db_dma_handle; /* Conn DB DMA handle */
	ddi_dma_cookie_t	conn_db_cookie;	/* Connection DB DMA cookie */
	uint32_t		conn_db_cookie_cnt; /* Conn DB cookie count */
	caddr_t			conn_db_addr;	/* Conection DB address */
	uint64_t		conn_db_length;	/* Connection DB length */
	ddi_acc_handle_t	conn_db_acc_handle; /* Conn DB acc handle */
	union  fcoe_kwqe ofld_req1;
	union  fcoe_kwqe ofld_req2;
	union  fcoe_kwqe ofld_req3;
	union  fcoe_kwqe ofld_req4;
	union  fcoe_kwqe enbl_req;
	union  fcoe_kwqe disable_req;
	union  fcoe_kwqe destroy_req;
	struct kwqe *kwqe_arr[4];
	    uint8_t els_pending;
	bnxef_completion_log_t c_log[BNXEF_MAX_LOG_ENTRIES];
	uint16_t log_cur_idx;
} bnxef_session_t;

/*
 * The following is the session list maintained for all active sessions.
 */
typedef struct __bnxef_session_list__ {
	kmutex_t	sessionlist_mutex; /* Guard for traversal */
	bnxef_session_t *head; /* This will point to first session */
	uint32_t	num_items;	/* Num items in the list */
	bnxef_session_t *tail;	/* Gives the last items, for linking */
} bnxef_session_list_t;

/*
 * This is the argument sent for each keep alive - Indicates whether
 * normal KA or vnport ka to be sent.
 */
typedef struct ka_arg	{
	void *portp;
	int	flags;
} ka_arg_t;

#define	VN_PORT_FLAG			1
#define	BNXEF_MAX_HASH_BUCKETS		256

typedef struct __bnxef_session_life_cycle__ {
	bnxef_session_t *sessionp;
	uint8_t		type;
	uint64_t	lbolt64;
} bnxef_session_life_cycle_t;

/*
 * This describes a FC Port, either local N_PORT or VN_Port or a remote port
 * identified during discovery. The initiator ports instantiated on this HBA
 * (N_port or VN_port), will have the Solaris specific Info if it is already
 * bound. If not it will have all the information 0 or NULL.
 *
 *  For remote port the only link back to the HBA structure is by identifying
 *  the session and then we can get to driver's state structure using the
 *  initiator port which established this nexus.
 *
 *  Typically most of the remote ports will be a SCSI Target port, howeer,
 *  there is a possibility some nexus is formed with remote ports which need
 *  be a target, For example Name Server.
 *
 *  Also not all information may not be populated, for example for remote ports
 *  we may not populate the rnid Information.
 */
typedef struct __bnxef_port__ {
	struct __bnxef_port__ 	*next;	/* Pointer to next port */
	uint32_t	id;	/* To start with this will be zero */
	uint8_t		mac_addr[ETHERADDRL]; /* mac address to be used */
	uint16_t	vlan_id;	/* Vlan on which ID acquired */
	uint32_t	previd;	/* Previous Id  */
	uint8_t		prev_mac_addr[ETHERADDRL];
	uint32_t	prevvlanid;	/* Prev Vlan ID */
	kmutex_t	port_mutex;	/* Mutex to protect and update state */
	uint32_t	flags; /* Indicate whether it is local or remote */
#define	 BNXEF_FC_PORT_ID_NOT_ACQUIRED   0x01
#define	 BNXEF_FC_PORT_ID_ACQUIRED	0x02
#define	 BNXEF_FC_PORT_N_PORT		0x04
#define	 BNXEF_FC_PORT_VN_PORT		0x08
#define	 BNXEF_FC_PORT_INITIATOR	0x10
#define	 BNXEF_FC_PORT_TARGET		0x20
#define	 BNXEF_FC_PORT_WELL_KNOWN	0x40	/* Need not be Offloaded */
#define	 BNXEF_FC_PORT_SOL_BOUND	0x80
#define	 BNXE_FC_PORT_RNID_VALID	0x100	/* Remote port RNID Invalid */
#define	BNXEF_FC_PORT_REMOTE		0x200	/* Remote port */
#define	BNXEF_FC_PORT_SHUTDOWN_IN_PROGRESS	0x400
#define	BNXEF_FC_PORT_SHUTDOWN_COMPLETE	0x800
#define	BNXEF_FC_PORT_CLONE		0x1000  /* A copy - No mutex needed */
#define	BNXEF_FC_PORT_ACQUISITION_INPROGRESS 0x2000
#define	BNXEF_FC_PORT_GENERATE_FABRIC_RSCN	0x4000
#define	BNXEF_FC_PORT_RSCN_TIMEOUT_VALID	0x8000
#define	BNXEF_FC_PORT_LOGO_TO_FABRIC_CTLR	0x10000

	la_els_logi_t	plogi_info;		/* Plogi Information */
	la_els_logi_t	fcf_logi_info;		/* FCF returned LOGI info */
	fc_fca_rnid_t	rnid_info;		/* rnid information */
	/* Sessions established */
	bnxef_session_list_t sessionhash[BNXEF_MAX_HASH_BUCKETS];

	/* BAck pointer to Driver state structure */
		void *statep;

	/*
	 * Solaris Specific Information, valid if bound, else not valid.
	 * Also, this information is not valid of if it is remote port
	 * This contains information about callback to be called when state
	 * changes on the initiator ports.
	 */
	fc_fca_bind_info_t  pbind;
	bnxef_unsol_buf_list_t	 unsol_buf_header;
	/*
	 * Provides Information About FCF, where a virtual link is established.
	 */
	bnxef_fcf_info_t		*fcf;

	uint32_t	fip_state;	/* Fip State with regard to this port */
#define	FIP_VIRTUAL_LINK_MAINTENANCE_IN_PROGRESS  0x01
#define	FIP_VIRTUAL_LINK_CLEAR_IN_PROGRESS	  0x02
#define	FIP_KEEP_ALIVE_ARMED			  0x04
#define	FIP_FLOGI_SENT				  0x08
#define	FIP_FLOGI_LS_ACC_RECEIVED		  0x10
#define	FIP_FLOGI_LS_RJT_RECEIVED		  0x20
#define	FIP_FDISC_SENT				  0x40
#define	FIP_FDISC_LS_ACC_RECEIVED		  0x80
#define	FIP_FDISC_LS_RJT_RECEIVED		  0x100
#define	FIP_LOGO_SENT				  0x200
#define	FIP_LOGO_LS_ACC_RECEIVED		  0x400
#define	FIP_LOGO_LS_RJT_RECEIVED		  0x800
#define	FIP_KEEP_ALIVE_TIMER2_ARMED		  0x1000

	bnxef_fcoe_task_t  *fip_task;
	uint8_t		fip_action;		/* If LS_RJT received */
	uint8_t		fip_reason;		/* if LS_RJT is received */
	kcondvar_t	id_cv;			/* Used to wait for ID */
	kcondvar_t	shutdown_cv;		/* Syncs Shutdown request */

	uint32_t	vn_period;	/* Vn period used for KEEP Alive */
	timeout_id_t	keep_alive_tmr_hdl; /* Keep alive tmr handle */
	timeout_id_t	keep_alive_tmr2_hdl; /* USed for first port (N_PORT) */
	uint8_t		count;		/* for testing CVL, will drop FKA */
	ka_arg_t	kaarg1;
	ka_arg_t	kaarg2;
	uint64_t	num_pkts_inited;
	timeout_id_t	rscn_tid;
	uint64_t	rscn_rx_lbolt;
	bnxef_session_life_cycle_t  slcp[BNXEF_SLCP_MAX_IDX];
	uint8_t		slcpidx;
} bnxef_port_t;

typedef bnxef_port_t bnxef_lport_t;
typedef bnxef_port_t bnxef_rport_t;




/*
 * The following is the port list maintained of all local ports
 */
typedef struct __bnxef_port_list__ {
	kmutex_t	portlist_mutex; /* Guard for traversal */
	bnxef_port_t *head; /* This will point to first port */
	uint32_t	num_items;	/* Num items in the list */
	bnxef_port_t *tail;	/* Gives the last items, for linking */
} bnxef_port_list_t;



#endif
