/*
 * The information contained in this file is confidential and proprietary to
 * Broadcom Corporation.  No part of this file may be reproduced or
 * distributed, in any form or by any means for any purpose, without the
 * express written permission of Broadcom Corporation.
 *
 * (c) COPYRIGHT 2010-2012 Broadcom Corporation, ALL RIGHTS RESERVED.
 */

#pragma ident "(C) 2010-2012 Broadcom Corporation"
#pragma ident " $File$$Revision$$Change$$DateTime$"

#ifndef __BNXEF_FCOE_TASK_H__
#define	__BNXEF_FCOE_TASK_H__

#include <sys/types.h>
#include <sys/conf.h>
#include <sys/stream.h>
#include <sys/stropts.h>
#include <sys/errno.h>
#include <sys/cred.h>
#include <sys/poll.h>
#include <sys/modctl.h>
#include <sys/stat.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>
#include <sys/ddifm.h>
#include <sys/fm/protocol.h>
#include <sys/fm/util.h>
#include <sys/fm/io/ddi.h>
#include <sys/pattr.h>
#include <sys/sysmacros.h>
#include <sys/strsun.h>
#include <sys/ksynch.h>
#include <sys/kstat.h>
#include <sys/strsubr.h>
#include <sys/ksynch.h>
#include <sys/ethernet.h>

/* Stages for TASK BD table allocation */
#define	BNXEF_TASK_BD_TABLE_MEM_BOUND		0x01
#define	BNXEF_TASK_BD_TABLE_MEM_ALLOCATED	0x02
#define	BNXEF_TASK_BD_TABLE_DMA_HANDLE_ALLOCATED 0x04
#define	BNXEF_TASK_BD_TABLE_ALL			0x07

#define	BNXEF_PKT_CMD_DMA_MEM_BOUND		0x01
#define	BNXEF_PKT_CMD_DMA_MEM_ALLOCATED		0x02
#define	BNXEF_PKT_CMD_DMA_HANDLE_ALLOCATED	0x04
#define	BNXEF_PKT_CMD_ALL			0x07

#define	BNXEF_PKT_RESP_DMA_MEM_BOUND		0x01
#define	BNXEF_PKT_RESP_DMA_MEM_ALLOCATED	0x02
#define	BNXEF_PKT_RESP_DMA_HANDLE_ALLOCATED	0x04
#define	BNXEF_PKT_RESP_ALL			0x07


/* This give the meta data for the HSI defined task information */
typedef struct __bnxef_fcoetask__ {
	uint16_t	x_id;		/* Task Id, could be OX_ID or RX_ID */
	uint32_t	task_flags;	/* Indicates allocated or free */
	kmutex_t	task_mutex;	/* Use to guard flags update and CV */
	kcondvar_t task_cv;		/* Completion in polled mode */
	struct __bnxef_fcoetask__ *nextp; /* Next ptr in Free list */
	fc_packet_t	*pktp;		/* Reference to Command which uses it */
	struct fcoe_task_ctx_entry *fwtaskp; /* Dma'able task ptr */
	uint8_t		dma_handle_idx;	/* Index to access fwtaskp via dMA */
	void		 *sessionp;	/* Ptr to session it belongs */
	struct __bnxef_fcoetask__ *snextp; /* The next task in the session */
	struct __bnxef_fcoetask__ *sprevp; /* Doubly liked - back ptr */
	timeout_id_t	timerhdl;	/* timer handle for pkt_timeout arm */
	bnxef_tq_params_t tq_params;		/* Tq params */
	ddi_dma_handle_t  bd_dma_handle; /* BD DMA handle */
	ddi_acc_handle_t  bd_dev_acc;
	caddr_t		  bd_addr;	/* Bd address */
	uint64_t	  bd_length;	/* actual length allocated */
	ddi_dma_cookie_t  bd_cookie;	 /* This the BD cookie */
	uint_t		  bd_cookie_cnt;  /* The number of cookies */
	uint32_t	  actual_bd_entries;	/* num entries in bd table */
	uint16_t	  secondary_x_id;  /* used for ABTS handling or RRQ */
	uint8_t	 *frame_hdrp;
	uint8_t		*payloadp;
	uint8_t		*orig_payloadp;
	uint32_t	payloadsz;
	frtn_t		mpfree;
	ddi_dma_cookie_t  rrq_cmd_cookie; /*  used when Recovery in progress */
	ddi_dma_cookie_t  rrq_resp_cookie;
	ddi_dma_cookie_t  *data_cookie;	/* Copied Data cookie ptr */
	void		  *dsp;		/* State Struct ptr */
	uint32_t	 sec_flags;	/* Secondary flags */
	uint8_t		 rec_timeout_cnt;    /* Used only for Rec failures */
	struct fcoe_err_report_entry error; /* Copy of Error report */
	uint8_t	state;  /* State to be set- Cleanup completion */
	uint8_t	action; /* Action tobe set - cleanup completion */
	uint32_t	reason;
	uint32_t	expln;
	clock_t	start_lbolt;
	clock_t	end_lbolt;
	clock_t	post_lbolt;
	clock_t	indicate_lbolt;
	int8_t	ref_cnt;
} bnxef_fcoe_task_t;

/* Task_flags defined */
#define	BNXEF_FCOE_TASK_ALLOCATED	0x01
#define	BNXEF_FCOE_TASK_FREE		0x00
#define	BNXEF_FCOE_TASK_TIMER_ARMED	0x02
#define	BNXEF_FCOE_TASK_TIMEDOUT	0x04
#define	BNXEF_FCOE_RECOVERY_IN_PROGRESS	0x08
#define	BNXEF_FCOE_SUBMITTED_TO_FIRMWARE 0x10
#define	BNXEF_FCOE_TASK_SUBMITTED_L2	 0x10
#define	BNXEF_FCOE_TASK_COMPLETED	0x20
#define	BNXEF_FCOE_RETURNED_NO_RESOURCE  0x40	/* No Resource for Queuing */
#define	BNXEF_FCOE_TASK_MGMT_CMD	0x80
#define	BNXEF_FCOE_TASK_ELS		0x100
#define	BNXEF_FCOE_TASK_ABTS		0x200
#define	BNXEF_FCOE_TASK_SCSI		0x400
// #define	BNXEF_FCOE_TASK_SUBMITTED_L2	0x800
#define	BNXEF_FCOE_TASK_BD_LIST_ALLOC	0x800	/* BD List allocated */
#define	BNXEF_FCOE_SEC_X_ID_VALID	0x1000	/* RRQ or ABTS X_ID valid */
#define	BNXEF_FCOE_TASK_RRQ		0x2000 /* Means DMA res etc alloced */
#define	BNXEF_FCOE_TASK_CLEANUP		0x4000
#define	BNXEF_FCOE_TASK_FRAME_HDR_ALLOC	0x8000
#define	BNXEF_FCOE_TASK_PAYLOAD_ALLOC	0x10000
#define	BNXEF_FCOE_TASK_ORIG_PAYLOAD_ALLOC 0x20000
#define	BNXEF_FCOE_TASK_UNSOLICITED_ELS	0x40000
#define	BNXEF_FCOE_TASK_UNSOL_LOCAL_RESP 0x80000
#define	BNXEF_FCOE_TASK_REC		0x100000
#define	BNXEF_FCOE_TASK_FCPLS		0x200000
#define	BNXEF_FCOE_TASK_SRR		0x400000
#define	BNXEF_FCOE_TASK_SEQ_CLEANUP	0x800000
#define	BNXEF_FCOE_TASK_NO_PKT_COMP 0x1000000 /* USed only in REC case */
#define	BNXEF_FCOE_TASK_OFFLOADED	0x2000000 /* Used for debugging */


/* sec flags definitions */
#define	BNXEF_ABTS_INITIATED		0x01
#define	BNXEF_ABTS_TIMEDOUT		0x02
#define	BNXEF_RRQ_SENT			0x04
#define	BNXEF_RRQ_TIMEDOUT		0x08
#define	BNXEF_BA_ACC_RCVD		0x10
#define	BNXEF_BA_RJT_RCVD		0x20
#define	BNXEF_ABTS_ERROR		0x40
#define	BNXEF_REC_SENT			0x80
#define	BNXEF_SRR_SENT			0x100
#define	BNXEF_FCOE_DROP_RRQ_PROCESSING	0x200

/*
 * The following defines the list head for free list of task and alloc list
 * At init time, the free list is initialized with available OX_ID and taskp
 * dma'able addresses for the task. Then when an I/O starts we will get the
 * first free item from the list. This way we donot have to have a separate
 * OX_ID allocation scheme.
 *
 * When the I/O completes an item will move from the alloc list to  free list.
 */
typedef  struct __bnxe_task_list__ {
	kmutex_t	tasklist_mutex; /* Guard for traversal */
	bnxef_fcoe_task_t *head; /* This will point to first item */
	bnxef_fcoe_task_t *tail;	/* Gives the last items, for linking */
	kcondvar_t	tasklist_cv;	/* Wait for resources to be available */
	uint32_t	num_items;	/* Num items in the list */
} bnxef_task_list_t;


/* List of routines to be defined for operating on the free list */

/*
 * This will wait for the resources to be available before it allocates
 * the task. This will wait on the free list cv untill resources become
 * available if the sleep flag indicates sleeping is allowed, if no
 * sleeping is allowed this will task or NULL when no resources are available.
 */

bnxef_fcoe_task_t *bnxe_task_alloc(void *dsp, int sleepflags);

/* This will free the given task to free list at the end */
void
bnxef_task_free(void *dsp, bnxef_fcoe_task_t *taskp);

#endif
