/*
 * The information contained in this file is confidential and proprietary to
 * Broadcom Corporation.  No part of this file may be reproduced or
 * distributed, in any form or by any means for any purpose, without the
 * express written permission of Broadcom Corporation.
 *
 * (c) COPYRIGHT 2010-2012 Broadcom Corporation, ALL RIGHTS RESERVED.
 */

#pragma ident "(C) 2010-2012 Broadcom Corporation"
#pragma ident " $File$$Revision$$Change$$DateTime$"
#include <sys/conf.h>
#include <sys/ddi.h>
#include <sys/stat.h>
#include <sys/pci.h>
#include <sys/sunddi.h>
#include <sys/modctl.h>
#include <sys/cmn_err.h>
#include <sys/file.h>
#include <sys/cred.h>
#include <sys/byteorder.h>
#include <sys/atomic.h>
#include <sys/scsi/scsi.h>
#include <sys/modhash.h>
#include <sys/sunldi.h>
#include <sys/kstat.h>
#include <sys/kmem.h>
#include <sys/ethernet.h>
#include <sys/stream.h>
#include <sys/sunldi.h>
#include <note.h>
/*
 * leadville header files
 */
#include <sys/fibre-channel/fc_types.h>
#include <sys/fibre-channel/impl/fc_fcaif.h>
#include <bnxe_binding.h>
#include <bnxef_tq_param.h>
#include <bnxef_fcf_info.h>
#include <bnxef_hsi.h>
#include <bnxef_fcoe_task.h>
#include <bnxef_fcoe_session.h>
#include <bnxef_state.h>
#include <bnxef_fip.h>
#include <bnxef_fcoe.h>
#include <bnxef_debug.h>

typedef uint8_t mac_addr_t[ETHERADDRL];
extern bnxef_driver_state_t *global_state[];
extern int bnxef_get_recv_size(bnxef_driver_state_t *dsp);


/* prototypes */
static void bnxef_prep_mac_addr_desc(uint8_t *bufp, mac_addr_t *macaddrp);
static void bnxef_prep_name_identifier_desc(uint8_t *bufp, uint8_t *namep);
static void bnxef_prep_max_fcoe_size_desc(uint8_t *bufp, uint16_t maxfcoesize);
static void bnxef_prep_vx_port_desc(uint8_t *bufp, bnxef_lport_t *portp);
static void bnxef_fip_prep_vlan_request(uint8_t *bufp, mac_addr_t *macp,
			uint8_t vlan_pri, uint16_t vlan_id);
static void bnxef_prep_disc_solicitation(uint8_t *bufp, mac_addr_t *srcp,
	uint8_t pri, uint16_t vlan_id, uint8_t fpspmaflag, uint16_t size);
static void bnxef_prep_fip_flogi_fdisc_request(uint8_t *bufp,
		bnxef_driver_state_t *dsp, bnxef_fcf_info_t *fcfp,
		uint8_t opcode, uint16_t ox_id, uint8_t *port_name,
		uint8_t *node_name);
static void bnxef_prep_fip_logo_request(uint8_t *bufp,
	bnxef_driver_state_t *dsp,
	bnxef_fcf_info_t *fcfp, bnxef_lport_t *portp);
static void bnxef_prep_fip_keep_alive_request(uint8_t *bufp,
	bnxef_driver_state_t *dsp, bnxef_fcf_info_t *fcfp,
	bnxef_lport_t *portp);

static void
bnxef_prep_fip_ls_rjt_response(uint8_t *bufp, bnxef_driver_state_t *dsp,
	bnxef_fcf_info_t *fcfp, bnxef_lport_t *portp, uint16_t ox_id,
	uint32_t s_id);

static void bnxef_prep_fip_ls_acc_response(uint8_t *bufp,
	bnxef_driver_state_t *dsp,
	bnxef_fcf_info_t *fcfp, bnxef_lport_t *portp, uint16_t ox_id);

mac_addr_t all_fcf_mac = { 0x01, 0x10, 0x18, 0x01, 0x00, 0x02 };
#if 0
static mac_addr_t all_enode_mac = { 0x01, 0x10, 0x18, 0x01, 0x00, 0x01 };
static mac_addr_t all_fcoe_mac = { 0x01, 0x10, 0x18, 0x01, 0x00, 0x00 };
#endif

void
bnxef_big_endian(void *bufp, int size)
{
#ifdef LITTLE_ENDIAN
	uint32_t x;
	uint32_t *bp = (uint32_t *)bufp;
	int i;

	if (size % sizeof (uint32_t) != 0)
		ASSERT(0);

	for (i = 0; i < size/sizeof (uint32_t); i++)  {
		x = *bp;
		*bp = htonl(x);
		bp++;
	}
#else

#endif
}


/*
 * The following are the utility routines to prepare FIP Frame descriptors
 */
static void
bnxef_prep_mac_addr_desc(uint8_t *bufp, mac_addr_t *macaddrp)
{
	ASSERT(bufp != NULL);
	ASSERT(macaddrp != NULL);

	/* LINTED: alignment */
	fip_mac_addr_desc_fmt_t *fp = (fip_mac_addr_desc_fmt_t *)bufp;
	fp->type = FIP_DESC_TYPE_MAC_ADDR;
	fp->length = 2;
	fp->mac_addr_msb1 = ((uint8_t *)macaddrp)[0];
	fp->mac_addr_msb0 = ((uint8_t *)macaddrp)[1];
	bcopy(&((uint8_t *)macaddrp)[2], (uint8_t *)&fp->mac_addr_lsb, 4);
}

/* prepare name identifier descriptor. */
static void
bnxef_prep_name_identifier_desc(uint8_t *bufp, uint8_t *namep)
{
	ASSERT(bufp != NULL);
	ASSERT(namep != NULL);

	/* LINTED: alignment */
	fip_name_id_desc_fmt_t *fp = (fip_name_id_desc_fmt_t *)bufp;

#ifdef TESTING
	bnxef_driver_state_t *dsp;
	dsp = global_state[0];
#endif
	fp->type = FIP_DESC_NAME_IDENTIFIER;
		fp->length = 0x03;
	fp->name[0] = 0x10;
	fp->name[1] = 0x00;
	bcopy(namep, &fp->name[2], 6);

}

/* prepare max fcoe size descriptor */
static void
bnxef_prep_max_fcoe_size_desc(uint8_t *bufp, uint16_t maxfcoesize)
{
	ASSERT(bufp != NULL);

	fip_max_fcoe_size_desc_fmt_t *fp =
		/* LINTED: alignment */
		(fip_max_fcoe_size_desc_fmt_t *)bufp;

#ifdef TESTING
	bnxef_driver_state_t *dsp;
	dsp = global_state[0];
#endif

	fp->type = FIP_DESC_MAX_FCOE_SIZE;
	fp->length = 0x01;
	fp->max_fcoe_size = htons(maxfcoesize);

}

static void
bnxef_prep_vx_port_desc(uint8_t *bufp, bnxef_lport_t *portp)
{
	ASSERT(bufp != NULL);
	ASSERT(portp != NULL);

	/* LINTED: alignment */
	fip_vx_port_desc_fmt_t *fp = (fip_vx_port_desc_fmt_t *)bufp;
	fp->type = FIP_DESC_VX_PORT_ID;
	fp->length = sizeof (fip_vx_port_desc_fmt_t)/(sizeof (uint32_t));

	if ((portp->flags & BNXEF_FC_PORT_CLONE) == 0)
		mutex_enter(&portp->port_mutex);
	bcopy(portp->mac_addr, fp->macaddr, 6);
	bcopy(portp->plogi_info.nport_ww_name.raw_wwn, fp->port_name,
		sizeof (la_wwn_t));
	fp->address_id = htonl(portp->id);
	if ((portp->flags & BNXEF_FC_PORT_CLONE) == 0)
		mutex_exit(&portp->port_mutex);
}



/*
 *  This will always be sent to ALL FCF MAC, and the local address will be
 *  used for the source address.
 */
/*ARGSUSED*/
static void
bnxef_fip_prep_vlan_request(uint8_t *bufp, mac_addr_t *macp, uint8_t vlan_pri,
		uint16_t vlan_id)
{
	ASSERT(bufp != NULL);
	ASSERT(macp != NULL);

	/* LINTED: alignment */
	struct ether_header *ep = (struct ether_header *)bufp;
	fip_pdu_hdr_t	*fipp	= (fip_pdu_hdr_t *)(ep +1);
	/* LINTED: alignment */
	fip_operation_fmt_hdr_t *ophp = (fip_operation_fmt_hdr_t *)(fipp + 1);

	/* set the destination address  to all FCF MAC */
	bcopy((uint8_t *)all_fcf_mac, (uint8_t *)&(ep->ether_dhost), 6);

	/* Set the local address */
	bcopy((uint8_t *)macp, (uint8_t *)&(ep->ether_shost), 6);

	ep->ether_type = htons(FIP_ETHERTYPE);

	/* Set the FIP version */
	fipp->version = FIP_FRAME_VER;
	fipp->reserved = 0;

	/* Set protoco code/subcode and other flags */
	ophp->fip_prot_code = htons(FIP_PROT_VLAN);
	ophp->fip_subcode = FIP_PROT_SUB_VLAN_REQUEST;
	ophp->flags = 0;

	/* This only carries the MAC Address descriptor */
	ophp->desc_len = htons(sizeof (fip_mac_addr_desc_fmt_t)/4);

	bnxef_prep_mac_addr_desc((uint8_t *)(ophp + 1), macp);
}


/*
 * This will prepare the Discovery solicitation frame for transmission.
 * This will always send the frame to ALL-FCF-MACS.
 */
/*ARGSUSED*/
static void
bnxef_prep_disc_solicitation(uint8_t *bufp, mac_addr_t *srcp,
	uint8_t pri, uint16_t vlan_id, uint8_t fpspmaflag, uint16_t size)
{
	NOTE(ARGUNUSED(vlan_id, pri))
	ASSERT(bufp != NULL);
	uint8_t	*bp;
	/* LINTED: alignment */
	struct ether_header *ep = (struct ether_header *)bufp;
		fip_pdu_hdr_t	*fipp   = (fip_pdu_hdr_t *)(ep +1);
	/* LINTED: alignment */
	fip_operation_fmt_hdr_t *ophp = (fip_operation_fmt_hdr_t *)(fipp + 1);

	ASSERT(srcp != NULL);
	/* set the destination address  to all FCF MAC */
	bcopy((uint8_t *)all_fcf_mac, (uint8_t *)&(ep->ether_dhost), 6);

	/* Set the local address */
	bcopy((uint8_t *)srcp, (uint8_t *)&(ep->ether_shost), 6);

	ep->ether_type = htons(FIP_ETHERTYPE);

	/* Set the FIP version */
	fipp->version = FIP_FRAME_VER;
	fipp->reserved = 0;

	/* Set protoco code/subcode and other flags */
	ophp->fip_prot_code = htons(FIP_PROT_DISCOVERY);
	ophp->fip_subcode = FIP_PROT_SUB_DISC_SOLICIT;
	ophp->flags = fpspmaflag?  htons(BNXEF_FIP_FLAGS_FP):
		htons(BNXEF_FIP_FLAGS_SP);

	/* This only carries the MAC Address descriptor */
	ophp->desc_len = htons((sizeof (fip_mac_addr_desc_fmt_t) +
		sizeof (fip_name_id_desc_fmt_t) +
		sizeof (fip_max_fcoe_size_desc_fmt_t))/sizeof (uint32_t));

	bp = (uint8_t *)(ophp + 1);
	/* Prepare the MAC address descriptor for use */
	bnxef_prep_mac_addr_desc(bp, srcp);

	bp += sizeof (fip_mac_addr_desc_fmt_t);

	/* Prepare the Name Identifier Descriptor */
	bnxef_prep_name_identifier_desc(bp, (uint8_t *)srcp);

	bp += sizeof (fip_name_id_desc_fmt_t);

	/* prepare the MAx FCoE Size descriptor */
	bnxef_prep_max_fcoe_size_desc(bp, size);
}

static void
bnxef_prep_fip_flogi_fdisc_request(uint8_t *bufp, bnxef_driver_state_t *dsp,
		bnxef_fcf_info_t *fcfp, uint8_t opcode, uint16_t ox_id,
		uint8_t *port_name, uint8_t *node_name)
{
	ASSERT(bufp != NULL);
	uint8_t	*bp;
	/* LINTED: alignment */
	struct ether_header *ep = (struct ether_header *)bufp;
		fip_pdu_hdr_t	*fipp   = (fip_pdu_hdr_t *)(ep +1);
	/* LINTED: alignment */
	fip_operation_fmt_hdr_t *ophp = (fip_operation_fmt_hdr_t *)(fipp + 1);
	fip_fc_desc_fmt_t	*fipfcp = (fip_fc_desc_fmt_t *)(ophp + 1);
	fc_frame_hdr_t *framep;
	la_els_logi_t *loginp;
	svc_param_t *class3_param;
	uint8_t mac_addr[ETHERADDRL];

	ASSERT(dsp != NULL);
	ASSERT(fcfp != NULL);

	/* set the destination address  to all FCF MAC */
	bcopy(fcfp->mac_addr, (uint8_t *)&(ep->ether_dhost), 6);

	/* Set the local address */
	bcopy(dsp->mac_add, (uint8_t *)&(ep->ether_shost), 6);

	ep->ether_type = htons(FIP_ETHERTYPE);

	/* Set the FIP version */
	fipp->version = FIP_FRAME_VER;
	fipp->reserved = 0;

	/* Set protoco code/subcode and other flags */
	ophp->fip_prot_code = htons(FIP_PROT_VLINK);
	ophp->fip_subcode = FIP_PROT_SUB_VLINK_INIT_REQ;
	/* If set to 1 FPMA supported */
	ophp->flags = dsp->flags_fpma_spma? htons(BNXEF_FIP_FLAGS_FP):
			htons(BNXEF_FIP_FLAGS_SP);

	ophp->desc_len = htons(38);

	/*
	 * htons((sizeof (fip_mac_addr_desc_fmt_t) + 1 +
	 *	sizeof (fc_frame_hdr_t) +
	 *	sizeof (la_els_logi_t))/sizeof (uint32_t));
	 */
	ASSERT((opcode == LA_ELS_FLOGI) || (opcode == LA_ELS_FDISC));
	fipfcp->type = (opcode == LA_ELS_FLOGI)?FIP_DESC_FLOGI:
				FIP_DESC_NPIV_FDISC;
	fipfcp->length = 36; /* Shall be set to 36 for FLOGI */
	/* LINTED: alignment */
	framep = (fc_frame_hdr_t *)fipfcp->fc_info;
	framep->r_ctl = R_CTL_ELS_REQ;
	framep->d_id =  FS_FABRIC_F_PORT;
	framep->type =  FC_TYPE_EXTENDED_LS;
	framep->f_ctl = F_CTL_FIRST_SEQ | F_CTL_END_SEQ | F_CTL_SEQ_INITIATIVE;
	framep->ox_id = ox_id;
	framep->rx_id = 0xffff;
	bnxef_big_endian(framep, sizeof (fc_frame_hdr_t));

	loginp = (la_els_logi_t *)(framep + 1);
	loginp->ls_code.ls_code = opcode;
	bnxef_big_endian(loginp, 4);
	bcopy(port_name, &loginp->nport_ww_name.raw_wwn[0],  8);
	bcopy(node_name, &loginp->node_ww_name.raw_wwn[0], 8);
	loginp->common_service.fcph_version = htons(0x2020);
	loginp->common_service.btob_credit = htons(8);
	loginp->common_service.cmn_features = htons(0x8000);
	loginp->common_service.conc_sequences = htons(0xff);
	loginp->common_service.relative_offset = htons(3);
	loginp->common_service.e_d_tov = htons(0x07d0);
	class3_param = (svc_param_t *)&loginp->class_3;
	class3_param->class_opt = htons(0x8800);

	class3_param->rcv_size = loginp->common_service.rx_bufsize =
		htons((dsp->mtu_size >= 2500) ? 2048 :
				bnxef_get_recv_size(dsp));
	class3_param->conc_sequences = htons(0xff);
	class3_param->open_seq_per_xchng = htons(1);
	bp = (uint8_t *)(loginp + 1);

	/* Prepare the MAC address descriptor for use */
	bzero(mac_addr, 6);
	bnxef_prep_mac_addr_desc(bp, (mac_addr_t *)mac_addr);

	bp += sizeof (fip_mac_addr_desc_fmt_t);

}


static void
bnxef_prep_fip_logo_request(uint8_t *bufp, bnxef_driver_state_t *dsp,
	bnxef_fcf_info_t *fcfp, bnxef_lport_t *portp)
{

	ASSERT(bufp != NULL);
	uint8_t	*bp;
	/* LINTED: alignment */
	struct ether_header *ep = (struct ether_header *)bufp;
		fip_pdu_hdr_t	*fipp   = (fip_pdu_hdr_t *)(ep +1);
	/* LINTED: alignment */
	fip_operation_fmt_hdr_t *ophp = (fip_operation_fmt_hdr_t *)(fipp + 1);
	fip_fc_desc_fmt_t	*fipfcp = (fip_fc_desc_fmt_t *)(ophp + 1);
	fc_frame_hdr_t *framep;
	la_els_logo_t *logop;

	ASSERT(dsp != NULL);
	ASSERT(fcfp != NULL);
	ASSERT(portp != NULL);

	/* set the destination address  to all FCF MAC */
	bcopy(fcfp->mac_addr, (uint8_t *)&(ep->ether_dhost), 6);

	/* Set the local address */
	bcopy(dsp->mac_add, (uint8_t *)&(ep->ether_shost), 6);

	ep->ether_type = htons(FIP_ETHERTYPE);

	/* Set the FIP version */
	fipp->version = FIP_FRAME_VER;
	fipp->reserved = 0;

	/* Set protoco code/subcode and other flags */
	ophp->fip_prot_code = htons(FIP_PROT_VLINK);
	ophp->fip_subcode = FIP_PROT_SUB_VLINK_INIT_REQ;
	/* If set to 1 FPMA supported */
	ophp->flags = dsp->flags_fpma_spma? htons(BNXEF_FIP_FLAGS_FP):
			htons(BNXEF_FIP_FLAGS_SP);

	ophp->desc_len = htons(13);

	//		htons((sizeof (fip_mac_addr_desc_fmt_t) +
	//		sizeof (fip_fc_desc_fmt_t) +
	//		sizeof (fc_frame_hdr_t) +
	//		sizeof (la_els_logo_t))/sizeof (uint32_t));
	fipfcp->type = FIP_DESC_LOGO;
	fipfcp->length = 11; /* Shall be set to 11 for LOGO */

	/* LINTED: alignment */
	framep = (fc_frame_hdr_t *)fipfcp->fc_info;
	framep->r_ctl = R_CTL_ELS_REQ;
	framep->d_id =  FS_FABRIC_F_PORT;
	framep->s_id =  portp->id;
	framep->f_ctl = F_CTL_FIRST_SEQ | F_CTL_END_SEQ | F_CTL_SEQ_INITIATIVE;
	framep->type =  FC_TYPE_EXTENDED_LS;
	framep->ox_id = portp->fip_task->x_id;
	framep->rx_id = 0xffff;
	bnxef_big_endian(framep, sizeof (fc_frame_hdr_t));

	mutex_enter(&portp->port_mutex);

	logop = (la_els_logo_t *)(framep + 1);
	logop->ls_code.ls_code = LA_ELS_LOGO;
	logop->nport_id.port_id = portp->id;
	bnxef_big_endian((uint8_t *)logop, 8);
	bcopy(portp->plogi_info.nport_ww_name.raw_wwn,
		logop->nport_ww_name.raw_wwn,
		sizeof (la_wwn_t));
	mutex_exit(&portp->port_mutex);

	bp = (uint8_t *)(logop + 1);

	/* Prepare the MAC address descriptor for use */
	bnxef_prep_mac_addr_desc(bp, (mac_addr_t *)portp->mac_addr);

	bp += sizeof (fip_mac_addr_desc_fmt_t);
}

static void
bnxef_prep_fip_ls_rjt_response(uint8_t *bufp, bnxef_driver_state_t *dsp,
	bnxef_fcf_info_t *fcfp, bnxef_lport_t *portp, uint16_t ox_id,
	uint32_t s_id)
{
	ASSERT(bufp != NULL);
	uint8_t	*bp;
	/* LINTED: alignment */
	struct ether_header *ep = (struct ether_header *)bufp;
		fip_pdu_hdr_t	*fipp   = (fip_pdu_hdr_t *)(ep +1);
	/* LINTED: alignment */
	fip_operation_fmt_hdr_t *ophp = (fip_operation_fmt_hdr_t *)(fipp + 1);
	fip_fc_desc_fmt_t	*fipfcp = (fip_fc_desc_fmt_t *)(ophp + 1);
	fc_frame_hdr_t *framep;
	la_els_rjt_t *lsrjtp;

	ASSERT(dsp != NULL);
	ASSERT(fcfp != NULL);
	ASSERT(portp != NULL);

	/* set the destination address  to all FCF MAC */
	bcopy(fcfp->mac_addr, (uint8_t *)&(ep->ether_dhost), 6);

	/* Set the local address */
	bcopy(dsp->mac_add, (uint8_t *)&(ep->ether_shost), 6);

	ep->ether_type = htons(FIP_ETHERTYPE);

	/* Set the FIP version */
	fipp->version = FIP_FRAME_VER;
	fipp->reserved = 0;

	/* Set protoco code/subcode and other flags */
	ophp->fip_prot_code = htons(FIP_PROT_VLINK);
	ophp->fip_subcode = FIP_PROT_SUB_VLINK_RESP;
	/* If set to 1 FPMA supported */
	ophp->flags = dsp->flags_fpma_spma? htons(BNXEF_FIP_FLAGS_FP):
			htons(BNXEF_FIP_FLAGS_SP);

	ophp->desc_len = htons(11);  /* 11 for LOGO LS_RJT */

	fipfcp->type = FIP_DESC_LOGO;
	fipfcp->length = 9; /* Shall be set to 9 for LOGO LS_RJT */

	/* LINTED: alignment */
	framep = (fc_frame_hdr_t *)fipfcp->fc_info;
	framep->r_ctl = R_CTL_ELS_RSP;
	framep->d_id =  s_id;
	framep->s_id =  portp->id;
	framep->f_ctl = F_CTL_LAST_SEQ | F_CTL_END_SEQ | F_CTL_SEQ_INITIATIVE;
	framep->type =  FC_TYPE_EXTENDED_LS;
	framep->rx_id = portp->fip_task->x_id;
	framep->ox_id = ox_id;
	bnxef_big_endian(framep, sizeof (fc_frame_hdr_t));

	mutex_enter(&portp->port_mutex);

	lsrjtp = (la_els_rjt_t *)(framep + 1);
	lsrjtp->ls_code.ls_code = LA_ELS_RJT;
	lsrjtp->ls_code.mbz = 0;
	lsrjtp->reason =  0x0b; /* FC_REASON_CMD_UNSUPPORTED */


	bnxef_big_endian((uint8_t *)lsrjtp, 4);
	mutex_exit(&portp->port_mutex);

	bp = (uint8_t *)(lsrjtp + 1);

	/* Prepare the MAC address descriptor for use */
	bnxef_prep_mac_addr_desc(bp, (mac_addr_t *)dsp->mac_add);

	bp += sizeof (fip_mac_addr_desc_fmt_t);


}
static void
bnxef_prep_fip_ls_acc_response(uint8_t *bufp, bnxef_driver_state_t *dsp,
	bnxef_fcf_info_t *fcfp, bnxef_lport_t *portp, uint16_t ox_id)
{

	ASSERT(bufp != NULL);
	uint8_t	*bp;
	/* LINTED: alignment */
	struct ether_header *ep = (struct ether_header *)bufp;
		fip_pdu_hdr_t	*fipp   = (fip_pdu_hdr_t *)(ep +1);
	/* LINTED: alignment */
	fip_operation_fmt_hdr_t *ophp = (fip_operation_fmt_hdr_t *)(fipp + 1);
	fip_fc_desc_fmt_t	*fipfcp = (fip_fc_desc_fmt_t *)(ophp + 1);
	fc_frame_hdr_t *framep;
	ls_code_t *lsaccp;

	ASSERT(dsp != NULL);
	ASSERT(fcfp != NULL);
	ASSERT(portp != NULL);

	/* set the destination address  to all FCF MAC */
	bcopy(fcfp->mac_addr, (uint8_t *)&(ep->ether_dhost), 6);

	/* Set the local address */
	bcopy(dsp->mac_add, (uint8_t *)&(ep->ether_shost), 6);

	ep->ether_type = htons(FIP_ETHERTYPE);

	/* Set the FIP version */
	fipp->version = FIP_FRAME_VER;
	fipp->reserved = 0;

	/* Set protoco code/subcode and other flags */
	ophp->fip_prot_code = htons(FIP_PROT_VLINK);
	ophp->fip_subcode = FIP_PROT_SUB_VLINK_RESP;
	/* If set to 1 FPMA supported */
	ophp->flags = dsp->flags_fpma_spma? htons(BNXEF_FIP_FLAGS_FP):
			htons(BNXEF_FIP_FLAGS_SP);

	ophp->desc_len = htons(10);  /* 10 for LOGO LS_ACC */

	fipfcp->type = FIP_DESC_LOGO;
	fipfcp->length = 8; /* Shall be set to 11 for LOGO */

	/* LINTED: alignment */
	framep = (fc_frame_hdr_t *)fipfcp->fc_info;
	framep->r_ctl = R_CTL_ELS_RSP;
	framep->d_id =  FS_FABRIC_F_PORT;
	framep->s_id =  portp->id;
	framep->f_ctl = F_CTL_LAST_SEQ | F_CTL_END_SEQ | F_CTL_SEQ_INITIATIVE;
	framep->type =  FC_TYPE_EXTENDED_LS;
	framep->rx_id = portp->fip_task->x_id;
	framep->ox_id = ox_id;
	bnxef_big_endian(framep, sizeof (fc_frame_hdr_t));

	mutex_enter(&portp->port_mutex);

	lsaccp = (ls_code_t *)(framep + 1);
	lsaccp->ls_code = LA_ELS_ACC;
	lsaccp->mbz = 0;

	bnxef_big_endian((uint8_t *)lsaccp, 4);
	mutex_exit(&portp->port_mutex);

	bp = (uint8_t *)(lsaccp + 1);

	/* Prepare the MAC address descriptor for use */
	bnxef_prep_mac_addr_desc(bp, (mac_addr_t *)portp->mac_addr);

	bp += sizeof (fip_mac_addr_desc_fmt_t);
}

static void
bnxef_prep_fip_keep_alive_request(uint8_t *bufp, bnxef_driver_state_t *dsp,
	bnxef_fcf_info_t *fcfp, bnxef_lport_t *portp)
{
	ASSERT(bufp != NULL);
	uint8_t	*bp;
	/* LINTED: alignment */
	struct ether_header *ep = (struct ether_header *)bufp;
		fip_pdu_hdr_t	*fipp   = (fip_pdu_hdr_t *)(ep +1);
	/* LINTED: alignment */
	fip_operation_fmt_hdr_t *ophp = (fip_operation_fmt_hdr_t *)(fipp + 1);
	fip_mac_addr_desc_fmt_t	*macp = (fip_mac_addr_desc_fmt_t *)(ophp + 1);
	uint32_t	flags;

	ASSERT(dsp != NULL);
	ASSERT(fcfp != NULL);
	ASSERT(portp != NULL);

	if ((portp->flags & BNXEF_FC_PORT_CLONE) == 0) {
		mutex_enter(&portp->port_mutex);
		flags = portp->flags;
		mutex_exit(&portp->port_mutex);
	} else
		flags = portp->flags;

	/* set the destination address  to all FCF MAC */
	bcopy(fcfp->mac_addr, (uint8_t *)&(ep->ether_dhost), 6);

	if (flags & BNXEF_FC_PORT_N_PORT) {
		/* Set the local address */
		bcopy(dsp->mac_add, (uint8_t *)&(ep->ether_shost), 6);
	} else {
		ASSERT((flags & BNXEF_FC_PORT_VN_PORT) ==
			BNXEF_FC_PORT_VN_PORT);
		bcopy(portp->mac_addr, (uint8_t *)&(ep->ether_shost), 8);
	}

	ep->ether_type = htons(FIP_ETHERTYPE);


	/* Set the FIP version */
	fipp->version = FIP_FRAME_VER;
	fipp->reserved = 0;

	/* Set protoco code/subcode and other flags */
	ophp->fip_prot_code = htons(FIP_PROT_VLINK_MAIN);
	ophp->fip_subcode = FIP_PROT_SUB_KEEP_ALIVE;
	//  ophp->flags = 0;
	ophp->flags = dsp->flags_fpma_spma? htons(BNXEF_FIP_FLAGS_FP):
			htons(BNXEF_FIP_FLAGS_SP);


	if (flags & BNXEF_FC_PORT_N_PORT)
		ophp->desc_len =
		htons(sizeof (fip_mac_addr_desc_fmt_t)/ sizeof (uint32_t));
	if (flags & BNXEF_FC_PORT_VN_PORT)
		    ophp->desc_len =
			htons(sizeof (fip_mac_addr_desc_fmt_t) +
			sizeof (fip_vx_port_desc_fmt_t))/ sizeof (uint32_t);


	bp = (uint8_t *)macp;
	/* Prepare the MAC address descriptor for use */
	bnxef_prep_mac_addr_desc(bp, (mac_addr_t *)dsp->mac_add);

	bp += sizeof (fip_mac_addr_desc_fmt_t);

	if (flags & BNXEF_FC_PORT_VN_PORT) {
		/* prepare the Vx port Descripor */
		bnxef_prep_vx_port_desc(bp, portp);
	}
}



/*
 * The following are exported routines which will be used from other files.
 */

/*
 * This will allocate an mblk_t structure and return that pointer with the
 * VLAN request filled in completely in the format to be sent.
 */
mblk_t *
bnxef_prep_mblk_fip_vlan_request(bnxef_driver_state_t *dsp)
{
	mblk_t  *mp;
	int size;
	uint8_t *bufp;

	ASSERT(dsp != NULL);

	/* calculate the size of VLAN request, extra 4 for FCS, TBD */
	size = sizeof (struct ether_header) +
		sizeof (fip_pdu_hdr_t) + sizeof (fip_operation_fmt_hdr_t) +
		sizeof (fip_mac_addr_desc_fmt_t);

	BNXEF_DEBUG(dsp, FIP, -1, "%s%d: Size calculated for VLAN Request 0x%x",
		BNXEF_DRIVER_NAME, dsp->instance, size);
	mp = allocb(size, 0);
	if (mp == NULL) {
		cmn_err(CE_NOTE, "%s%d: Unable to allocate mblk_t structure\n",
				BNXEF_DRIVER_NAME, dsp->instance);
		return (NULL);
	}

	BNXEF_DEBUG(dsp, FIP, -1, "%s%d: Mblk allocated",
		BNXEF_DRIVER_NAME, dsp->instance);
	bufp = mp->b_wptr;

	bzero(bufp, size);

	/* Hold the Driver state mutex */
	MUTEX_ENTER(dsp, &dsp->state_mutex, __func__, __LINE__);

	/*
	 * In this case prepare the VLAn Request to be sent on default
	 * vlan ID = 0
	 */
	bnxef_fip_prep_vlan_request(bufp, (mac_addr_t *)dsp->mac_add,
		dsp->pri, 0);

	mp->b_wptr += size;

	/* Release the mutex */
	mutex_exit(&dsp->state_mutex);

	BNXEF_DEBUG(dsp, FIP, -1, "%s%d: Prepared the Mblk required",
		BNXEF_DRIVER_NAME, dsp->instance);

	return (mp);
}

/*
 * This prepares an mblk_t structure with the FIP Discovery Solicitation for
 * transmission
 */
mblk_t *
bnxef_prep_mblk_fip_disc_sol(bnxef_driver_state_t *dsp)
{
	mblk_t  *mp;
	int size;
	uint8_t *bufp;

	ASSERT(dsp != NULL);

	/* calculate the size of discovery Solicitation request */
	size = sizeof (struct ether_header) +
		sizeof (fip_pdu_hdr_t) + sizeof (fip_operation_fmt_hdr_t) +
		sizeof (fip_mac_addr_desc_fmt_t) +
		sizeof (fip_name_id_desc_fmt_t) +
		sizeof (fip_max_fcoe_size_desc_fmt_t);
	BNXEF_DEBUG(dsp, FIP, -1, "%s%d: Size calculated for Disc Sol 0x%x",
		BNXEF_DRIVER_NAME, dsp->instance, size);

	mp = allocb(size, 0);
	if (mp == NULL) {
		cmn_err(CE_NOTE, "%s%d: Unable to allocate mblk_t structure\n",
				BNXEF_DRIVER_NAME, dsp->instance);
		return (NULL);
	}

	bufp = mp->b_wptr;

	bzero(bufp, size);

	/* Hold the driver state mutex */
	MUTEX_ENTER(dsp, &dsp->state_mutex, __func__, __LINE__);

	/* Prepare Discovery Solicitation */
	bnxef_prep_disc_solicitation(bufp, (mac_addr_t *)dsp->mac_add,
		dsp->pri, dsp->fcf_ident->vlan_id, dsp->flags_fpma_spma,
		dsp->max_fcoe_size);

	mp->b_wptr += size;

	/* Release the mutex */
	mutex_exit(&dsp->state_mutex);

	return (mp);
}

/*
 * This prepares a mblk_t with the fip flogi descriptor included
 */
mblk_t *
bnxef_prep_mblk_fip_flogi_fdisc(bnxef_driver_state_t *dsp, uint8_t opcode,
		uint16_t ox_id, uint8_t *port_name, uint8_t *node_name)
{
	mblk_t  *mp;
	int size;
	uint8_t *bufp;
	bnxef_fcf_info_t *fcfp;

	ASSERT(dsp != NULL);

	/*
	 * calculate the size of FIP FLOGI request
	 * minus 4 is adjustment for padding in fip_fc_desc_fmt_t
	 */
	size = sizeof (struct ether_header) +
		sizeof (fip_pdu_hdr_t) + sizeof (fip_operation_fmt_hdr_t) +
		sizeof (fip_fc_desc_fmt_t) + sizeof (fc_frame_hdr_t) +
		sizeof (la_els_logi_t) + sizeof (fip_mac_addr_desc_fmt_t) - 4;

	mp = allocb(size, 0);
	if (mp == NULL) {
		cmn_err(CE_NOTE, "%s%d: Unable to allocate mblk_t structure\n",
				BNXEF_DRIVER_NAME, dsp->instance);
		return (NULL);
	}

	bufp = mp->b_wptr;

	bzero(bufp, size);

	/* Hold the State mutex */
	MUTEX_ENTER(dsp, &dsp->state_mutex, __func__, __LINE__);
	fcfp = dsp->fcf_ident;

	mutex_enter(&fcfp->fcf_mutex);
	ASSERT((fcfp->fcf_state & FCF_SELECTED_FOR_LOGIN));
	bnxef_prep_fip_flogi_fdisc_request(bufp, dsp, fcfp, opcode, ox_id,
		port_name, node_name);
	mutex_exit(&fcfp->fcf_mutex);

	mp->b_wptr += size;

	/* Release the state mutex */
	mutex_exit(&dsp->state_mutex);

	return (mp);
}

/*
 * This prepares a mblk_t structure to prepare a FIP LOGO frame
 * with appropriate descriptors
 */
mblk_t *
bnxef_prep_mblk_fip_logo(bnxef_driver_state_t *dsp, bnxef_lport_t *portp)
{
	mblk_t  *mp;
	int size;
	uint8_t *bufp;
	bnxef_fcf_info_t *fcfp;

	ASSERT(dsp != NULL);
	ASSERT(portp != NULL);

	/*
	 * calculate the size of FIP FLOGI request
	 * minus 4 is adjustment for padding in fip_fc_desc_fmt_t
	 */
	size = sizeof (struct ether_header) +
		sizeof (fip_pdu_hdr_t) + sizeof (fip_operation_fmt_hdr_t) +
		sizeof (fip_fc_desc_fmt_t) + sizeof (fc_frame_hdr_t) +
		sizeof (la_els_logo_t) +
		sizeof (fip_mac_addr_desc_fmt_t) -4;


	mp = allocb(size, 0);
	if (mp == NULL) {
		cmn_err(CE_NOTE, "%s%d: Unable to allocate mblk_t structure\n",
				BNXEF_DRIVER_NAME, dsp->instance);
		return (NULL);
	}

	bufp = mp->b_wptr;

	bzero(bufp, size);

	/* Hold the State mutex */
	MUTEX_ENTER(dsp, &dsp->state_mutex, __func__, __LINE__);

	fcfp = dsp->fcf_ident;
	if (fcfp == NULL) {
		mutex_exit(&dsp->state_mutex);
		freemsg(mp);
		return (NULL);
	}

	/* Make sure the task is allocated for FIP FLOGI Frame */
	ASSERT(portp->fip_task != NULL);

	mutex_enter(&fcfp->fcf_mutex);
	ASSERT((fcfp->fcf_state & FCF_SELECTED_FOR_LOGIN));
	if (fcfp->fcf_state & FCF_SELECTED_FOR_LOGIN) {
		bnxef_prep_fip_logo_request(bufp, dsp, fcfp, portp);
	} else {
		mutex_exit(&fcfp->fcf_mutex);
		mutex_exit(&dsp->state_mutex);
		freemsg(mp);
		return (NULL);
	}

	mutex_exit(&fcfp->fcf_mutex);
	mp->b_wptr += size;

	/* Release the state mutex */
	mutex_exit(&dsp->state_mutex);

	return (mp);
}

mblk_t *
bnxef_prep_mblk_fip_ls_rjt(bnxef_driver_state_t *dsp, bnxef_lport_t *portp,
	uint16_t ox_id, uint32_t s_id)
{
	mblk_t  *mp;
	int size;
	uint8_t *bufp;
	bnxef_fcf_info_t *fcfp;

	ASSERT(dsp != NULL);
	ASSERT(portp != NULL);

	/* calculate the size of FIP request */
	size = sizeof (struct ether_header) +
		sizeof (fip_pdu_hdr_t) + sizeof (fip_operation_fmt_hdr_t) +
		sizeof (fip_fc_desc_fmt_t) + sizeof (fc_frame_hdr_t) +
		sizeof (la_els_rjt_t) +
		sizeof (fip_mac_addr_desc_fmt_t);

	mp = allocb(size, 0);
	if (mp == NULL) {
		cmn_err(CE_NOTE, "%s%d: Unable to allocate mblk_t structure\n",
				BNXEF_DRIVER_NAME, dsp->instance);
		return (NULL);
	}

	bufp = mp->b_wptr;

	bzero(bufp, size);

	/* Hold the State mutex */
	MUTEX_ENTER(dsp, &dsp->state_mutex, __func__, __LINE__);

	fcfp = dsp->fcf_ident;

	/* Make sure the task is allocated for FIP FLOGI Frame */
	ASSERT(portp->fip_task != NULL);

	mutex_enter(&fcfp->fcf_mutex);
	ASSERT((fcfp->fcf_state & FCF_SELECTED_FOR_LOGIN));
	bnxef_prep_fip_ls_rjt_response(bufp, dsp, fcfp, portp, ox_id, s_id);
	mutex_exit(&fcfp->fcf_mutex);
	mp->b_wptr += size;

	/* Release the state mutex */
	mutex_exit(&dsp->state_mutex);

	return (mp);
}

mblk_t *
bnxef_prep_mblk_fip_ls_acc(bnxef_driver_state_t *dsp, bnxef_lport_t *portp,
		uint16_t ox_id)
{
	mblk_t  *mp;
	int size;
	uint8_t *bufp;
	bnxef_fcf_info_t *fcfp;

	ASSERT(dsp != NULL);
	ASSERT(portp != NULL);

	/* calculate the size of FIP FLOGI request */
	size = sizeof (struct ether_header) +
		sizeof (fip_pdu_hdr_t) + sizeof (fip_operation_fmt_hdr_t) +
		sizeof (fip_fc_desc_fmt_t) + sizeof (fc_frame_hdr_t) +
		sizeof (ls_code_t) +
		sizeof (fip_mac_addr_desc_fmt_t);

	mp = allocb(size, 0);
	if (mp == NULL) {
		cmn_err(CE_NOTE, "%s%d: Unable to allocate mblk_t structure\n",
				BNXEF_DRIVER_NAME, dsp->instance);
		return (NULL);
	}

	bufp = mp->b_wptr;

	bzero(bufp, size);

	/* Hold the State mutex */
	MUTEX_ENTER(dsp, &dsp->state_mutex, __func__, __LINE__);

	fcfp = dsp->fcf_ident;

	/* Make sure the task is allocated for FIP FLOGI Frame */
	ASSERT(portp->fip_task != NULL);

	mutex_enter(&fcfp->fcf_mutex);
	ASSERT((fcfp->fcf_state & FCF_SELECTED_FOR_LOGIN));
	bnxef_prep_fip_ls_acc_response(bufp, dsp, fcfp, portp, ox_id);
	mutex_exit(&fcfp->fcf_mutex);
	mp->b_wptr += size;

	/* Release the state mutex */
	mutex_exit(&dsp->state_mutex);

	return (mp);
}


/*
 * This generates the appropriate Keep Alive Frame depending on the port
 * information.
 */
mblk_t *
bnxef_prep_mblk_fip_keep_alive(bnxef_driver_state_t *dsp, bnxef_lport_t *portp)
{
	mblk_t  *mp;
	int size;
	uint8_t *bufp;
	bnxef_fcf_info_t *fcfp;

	ASSERT(dsp != NULL);
	ASSERT(portp != NULL);

	/* calculate the size of FIP Keep Alive request */
	size = sizeof (struct ether_header) +
		sizeof (fip_pdu_hdr_t) + sizeof (fip_operation_fmt_hdr_t) +
		sizeof (fip_mac_addr_desc_fmt_t) +
		sizeof (fip_vx_port_desc_fmt_t);

	mp = allocb(size, 0);
	if (mp == NULL) {
		cmn_err(CE_NOTE, "%s%d: Unable to allocate mblk_t structure\n",
				BNXEF_DRIVER_NAME, dsp->instance);
		return (NULL);
	}

	bufp = mp->b_wptr;

	bzero(bufp, size);

	/* Hold the State mutex */
	MUTEX_ENTER(dsp, &dsp->state_mutex, __func__, __LINE__);

	fcfp = dsp->fcf_ident;
	if (fcfp == NULL) {
		mutex_exit(&dsp->state_mutex);
		return (NULL);
	}

	mutex_exit(&dsp->state_mutex);
	/* Make sure the task is allocated for FIP FLOGI Frame */
	ASSERT(portp->fip_task != NULL);

	mutex_enter(&fcfp->fcf_mutex);
	ASSERT((fcfp->fcf_state & FCF_SELECTED_FOR_LOGIN));
	mutex_exit(&fcfp->fcf_mutex);
	bnxef_prep_fip_keep_alive_request(bufp, dsp, fcfp, portp);

	mp->b_wptr += size;
#if 0
	/* Release the state mutex */
	mutex_exit(&dsp->state_mutex);
#endif

	return (mp);
}
