/*
 * The information contained in this file is confidential and proprietary to
 * Broadcom Corporation.  No part of this file may be reproduced or
 * distributed, in any form or by any means for any purpose, without the
 * express written permission of Broadcom Corporation.
 *
 * (c) COPYRIGHT 2010-2012 Broadcom Corporation, ALL RIGHTS RESERVED.
 */

#pragma ident "(C) 2010-2012 Broadcom Corporation"
#pragma ident " $File$$Revision$$Change$$DateTime$"
#include <sys/conf.h>
#include <sys/ddi.h>
#include <sys/stat.h>
#include <sys/pci.h>
#include <sys/sunddi.h>
#include <sys/modctl.h>
#include <sys/cmn_err.h>
#include <sys/file.h>
#include <sys/cred.h>
#include <sys/byteorder.h>
#include <sys/atomic.h>
#include <sys/scsi/scsi.h>
#include <sys/modhash.h>
#include <sys/sunldi.h>
#include <sys/kstat.h>
#include <sys/kmem.h>
#include <sys/ethernet.h>
#include <sys/stream.h>
#include <sys/sunldi.h>
/*
 * leadville header files
 */
#include <sys/fibre-channel/fc_types.h>
#include <sys/fibre-channel/impl/fc_fcaif.h>
#include <bnxe_binding.h>
#include <bnxef_tq_param.h>
#include <bnxef_fcf_info.h>
#include <bnxef_hsi.h>
#include <bnxef_fcoe_task.h>
#include <bnxef_fcoe_session.h>
#include <bnxef_state.h>
#include <bnxef_fip.h>
#include <bnxef_fcoe.h>
#include <bnxef_debug.h>

#define	BNXEF_RQ_WQE_SIZE 1024

/* Prototypes */
uint8_t bnxef_session_gen_hash(int16_t vlan_id, uint32_t d_id,
	uint8_t *macaddr);
int bnxef_session_conn_db_res_alloc(bnxef_session_t *sessionp);
int bnxef_session_conn_db_res_free(bnxef_session_t *sessionp, uint32_t stage);
int bnxef_session_sq_res_alloc(bnxef_session_t *sessionp);
int bnxef_sq_res_free(bnxef_session_t *sessionp, uint32_t stage);
int bnxef_session_cq_res_alloc(bnxef_session_t *sessionp);
int bnxef_cq_res_free(bnxef_session_t *sessionp, uint32_t stage);
int bnxef_session_rq_res_alloc(bnxef_session_t *sessionp);
int bnxef_rq_res_free(bnxef_session_t *sessionp, uint32_t stage);
int bnxef_xferq_res_free(bnxef_session_t *sessionp, uint32_t stage);
int bnxef_xferq_res_free(bnxef_session_t *sessionp, uint32_t stage);
int bnxef_session_confq_res_alloc(bnxef_session_t *sessionp);
void bnxef_session_confq_res_free(bnxef_session_t *sessionp, uint32_t stage);
int bnxef_session_lcq_res_alloc(bnxef_session_t *sessionp);
int bnxef_session_lcq_res_free(bnxef_session_t *sessionp, uint32_t stage);
int bnxef_free_session_resources(bnxef_session_t *sessionp, uint32_t stage);
int bnxef_session_xferq_res_alloc(bnxef_session_t *sessionp);
int bnxef_find_session(bnxef_port_t *portp, bnxef_session_t *sp);

/* External Prototypes/ definition */
extern ddi_dma_attr_t bnxef_dma_attr;
extern ddi_device_acc_attr_t bnxef_dev_acc_attr;


/*
 * This file provides the list of routines which works with session object
 */
/*
 * The following is the cache constructor for the session cache allocated.
 */

/*ARGSUSED*/
int
bnxef_session_cache_constructor(void *bufp, void *args, int kmflags)
{
	char  name[KSTAT_STRLEN];
	bnxef_session_t *sessionp = (bnxef_session_t *)bufp;
	bnxef_driver_state_t *dsp = (bnxef_driver_state_t *)args;
	uint32_t	refcnt;

	NOTE(ARGUNUSED(kmflags))
	MUTEX_ENTER(dsp, &dsp->state_mutex, __func__, __LINE__);
	refcnt = dsp->session_ref_cnt;
	dsp->session_ref_cnt ++;
	mutex_exit(&dsp->state_mutex);


	bzero((uint8_t *)bufp, sizeof (bnxef_session_t));
	(void) snprintf(name, KSTAT_STRLEN, "session%d-%d_mutex",
		dsp->instance, refcnt);
	mutex_init(&sessionp->session_mutex, name, MUTEX_DRIVER, NULL);
	(void) snprintf(name, KSTAT_STRLEN, "session%d-%d_cv",
		dsp->instance, refcnt);
	cv_init(&sessionp->session_cv, name, CV_DRIVER, NULL);
	(void) snprintf(name, KSTAT_STRLEN, "session%d-%d_els",
		dsp->instance, refcnt);
	cv_init(&sessionp->session_els, name, CV_DRIVER, NULL);


	return (0);
}

/*
 * Session Cache destructor
 */
void
bnxef_session_cache_destructor(void *bufp, void *args)
{
	bnxef_session_t	*sessionp = (bnxef_session_t *)bufp;
	bnxef_driver_state_t *dsp = (bnxef_driver_state_t *)args;

	cv_destroy(&sessionp->session_els);
	cv_destroy(&sessionp->session_cv);
	mutex_destroy(&sessionp->session_mutex);

	MUTEX_ENTER(dsp, &dsp->state_mutex, __func__, __LINE__);
	dsp->session_ref_cnt--;
	mutex_exit(&dsp->state_mutex);

}


/*
 * This creates the cache for use at attach time.
 */
int
bnxef_session_cache_create(bnxef_driver_state_t *dsp)
{
	char  buf[KSTAT_STRLEN];

	ASSERT(dsp != NULL);

	(void) snprintf(buf, KSTAT_STRLEN, "session%d_cache", dsp->instance);
	dsp->session_cache = kmem_cache_create(buf, sizeof (bnxef_session_t),
		8, bnxef_session_cache_constructor,
		bnxef_session_cache_destructor, NULL,
		dsp, NULL,  0);

	if (dsp->session_cache == NULL) {
		cmn_err(CE_PANIC, "%s%d Unable to allocate port cache\n",
			BNXEF_DRIVER_NAME, dsp->instance);
			return (DDI_FAILURE);
	}

	return (DDI_SUCCESS);
}


/*
 * This destroys the cache at detach time.
 */
int
bnxef_session_cache_destroy(bnxef_driver_state_t *dsp)
{
	ASSERT(dsp != NULL);

	kmem_cache_destroy(dsp->session_cache);

	return (DDI_SUCCESS);
}


/*
 * This initializes one of the hash buckets out of 256 hash buckets.
 */
int
bnxef_session_list_init(bnxef_driver_state_t *dsp, bnxef_port_t *portp,
	uint8_t hashbucket)
{
	bnxef_session_list_t *bucketp;
	char  name[KSTAT_STRLEN];

	ASSERT(dsp != NULL);
	ASSERT(portp != NULL);

	mutex_enter(&portp->port_mutex);
	bucketp = &portp->sessionhash[hashbucket];
	mutex_exit(&portp->port_mutex);

	bzero((uint8_t *)bucketp, sizeof (bnxef_session_list_t));
	snprintf(name, KSTAT_STRLEN, "session%d-hashbucket%d-mutex",
		dsp->instance, hashbucket);
	mutex_init(&bucketp->sessionlist_mutex, name, MUTEX_DRIVER, NULL);

	return (BNXEF_SUCCESS);
}

int
bnxef_session_list_fini(bnxef_driver_state_t *dsp, bnxef_port_t *portp,
	uint8_t hashbucket)
{
	bnxef_session_list_t *bucketp;

	ASSERT(dsp != NULL);
	ASSERT(portp != NULL);

	mutex_enter(&portp->port_mutex);
	bucketp = &portp->sessionhash[hashbucket];
	mutex_exit(&portp->port_mutex);

	mutex_destroy(&bucketp->sessionlist_mutex);

	return (BNXEF_SUCCESS);
}

/*
 * This is a simple hash function which folds all bytes into one byte to get
 * the hash number.
 */
uint8_t
bnxef_session_gen_hash(int16_t vlan_id, uint32_t d_id, uint8_t *macaddr)
{
	uint8_t hash = 0;
	uint8_t  i;

	ASSERT(macaddr != NULL);

	for (i = 0; i < 6; i++)
		hash += *macaddr++;

	hash += (vlan_id & 0xff);
	hash += ((vlan_id >> 8) & 0xff);

	hash += (d_id & 0xff);
	hash += (d_id >> 8) & 0xff;
	hash += (d_id >> 16) & 0xff;

	return (hash);
}

/*
 * This will find the remote port from the list of sessions and identify the
 * session existing on this local port identified by the portp.
 *
 * This routine will do the ashing and find the right bucket to search and
 * provide the session if it exists, else it will return NULL.
 */
bnxef_session_t *
bnxef_session_find_remote_port(bnxef_port_t *portp, int16_t vlan_id,
		uint32_t d_id, uint8_t *macaddr)
{
	uint8_t	 hash;
	bnxef_session_list_t *bucketp;
	bnxef_session_t	*sessionp;
	int i, ret;
	bnxef_port_t	*rportp;
	int found = 0;
	bnxef_driver_state_t *dsp;

	ASSERT(macaddr != NULL);
	ASSERT(portp != NULL);

	dsp = portp->statep;
	ASSERT(dsp != NULL);

	hash = bnxef_session_gen_hash(vlan_id, d_id, macaddr);
	BNXEF_AUTO_DEBUG("%s%d: Hash value 0x%x",
		BNXEF_DRIVER_NAME, dsp->instance, hash);
	bucketp = &portp->sessionhash[hash];

	ASSERT(bucketp != NULL);

	mutex_enter(&bucketp->sessionlist_mutex);
	sessionp = bucketp->head;
	if (sessionp == NULL) {
		mutex_exit(&bucketp->sessionlist_mutex);
		return (NULL);
	}

	ASSERT(sessionp != NULL);
	BNXEF_AUTO_DEBUG("%s%d: Num Items in this bucket 0x%x\n",
		BNXEF_DRIVER_NAME, dsp->instance, bucketp->num_items);
	for (i = 0; i < bucketp->num_items; i++) {
		rportp = sessionp->rport;
		ASSERT(rportp != NULL);
		ret = bcmp(rportp->mac_addr, macaddr, 6);
		if ((rportp->vlan_id == vlan_id) &&
			(rportp->id  == d_id) && (ret == 0)) {
			BNXEF_AUTO_DEBUG(
			"%s%d: Found the right port vlanId 0x%x, port_id 0x%x",
				BNXEF_DRIVER_NAME, dsp->instance,
					vlan_id, d_id);
			found++;
			break;
		}
		sessionp = sessionp->next;
	// 	ASSERT(sessionp != NULL);
	}
	if (found) {
		mutex_enter(&sessionp->session_mutex);
		if (sessionp->state & BNXEF_SESSION_CLEANUP_INPROGRESS) {
			found = 0;
		}
		mutex_exit(&sessionp->session_mutex);
	}

	mutex_exit(&bucketp->sessionlist_mutex);
	return (found? sessionp: NULL);
}

int
bnxef_find_session(bnxef_port_t *portp, bnxef_session_t *sp)
{
	uint8_t	 hash;
	bnxef_session_list_t *bucketp;
	bnxef_session_t	*sessionp;
	int i;
	int found = 0;
	bnxef_driver_state_t *dsp;

	ASSERT(portp != NULL);

	dsp = portp->statep;
	ASSERT(dsp != NULL);

	for (hash = 0; hash < BNXEF_MAX_HASH_BUCKETS; hash++) {
		BNXEF_AUTO_DEBUG("%s%d: Hash value 0x%x",
			BNXEF_DRIVER_NAME, dsp->instance, hash);
		cmn_err(CE_NOTE, "%s%d: Hash value 0x%x",
			BNXEF_DRIVER_NAME, dsp->instance, hash);
		bucketp = &portp->sessionhash[hash];

		ASSERT(bucketp != NULL);

		mutex_enter(&bucketp->sessionlist_mutex);
		sessionp = bucketp->head;
		if (sessionp == NULL) {
			mutex_exit(&bucketp->sessionlist_mutex);
				continue;
		}

		ASSERT(sessionp != NULL);
		BNXEF_AUTO_DEBUG("%s%d: Num Items in this bucket 0x%x\n",
			BNXEF_DRIVER_NAME, dsp->instance, bucketp->num_items);
		for (i = 0; i < bucketp->num_items; i++) {
			if (sp == sessionp)  {
				BNXEF_AUTO_DEBUG(
				"%s%d: Found the session 0x%p",
					BNXEF_DRIVER_NAME, dsp->instance,
						sp);
				cmn_err(CE_NOTE, "%s%d: Found the session 0x%p",
					BNXEF_DRIVER_NAME, dsp->instance, sp);
				found++;
				break;
			}
			sessionp = sessionp->next;
		}
		mutex_exit(&bucketp->sessionlist_mutex);
		if (found)
			break;
	}

	return (found? BNXEF_SUCCESS: BNXEF_FAILURE);
}

/*
 * THis will return the first session in the given hash bucket, if there is
 * no session available it will return NULL. Explict call should be made to
 * remove that session from the hash bucket.
 */
bnxef_session_t *
bnxef_get_first_session(bnxef_port_t *portp, int bucket)
{
	bnxef_session_list_t *bucketp;
	bnxef_session_t	*sessionp;

	ASSERT(portp != NULL);

	bucketp = &portp->sessionhash[bucket];

	ASSERT(bucketp != NULL);

	mutex_enter(&bucketp->sessionlist_mutex);
	sessionp = bucketp->head;
	mutex_exit(&bucketp->sessionlist_mutex);
	return (sessionp);
}

bnxef_session_t *
bnxef_get_next_session(bnxef_port_t *portp, int bucket,
		bnxef_session_t *sessionp)
{
	bnxef_session_list_t *bucketp;
	bnxef_session_t *sessp;

	ASSERT(portp != NULL);

	bucketp = &portp->sessionhash[bucket];

	ASSERT(bucketp != NULL);
	ASSERT(sessionp != NULL);

	mutex_enter(&bucketp->sessionlist_mutex);
	sessp = sessionp->next;
	mutex_exit(&bucketp->sessionlist_mutex);

	return (sessp);
}



/*
 * This will add the given session pointer at the end of the list identified
 * by the hash bucket.
 */
int
bnxef_session_list_add(bnxef_port_t *portp, bnxef_session_t *sessionp,
	uint8_t hashbucket)
{

	bnxef_session_list_t *sessionlstp;
	bnxef_driver_state_t *dsp;
	ASSERT(portp != NULL);
	ASSERT(sessionp != NULL);

	sessionlstp = &portp->sessionhash[hashbucket];
	dsp = portp->statep;
	ASSERT(dsp != NULL);

	BNXEF_DEBUG(dsp, GENERAL, -1,
		"%s%d: Session 0x%p added to hash bucket 0x%x on port 0x%p",
		BNXEF_DRIVER_NAME, dsp->instance, sessionp, hashbucket, portp);
	ASSERT(sessionlstp != NULL);
	mutex_enter(&sessionlstp->sessionlist_mutex);

	if ((sessionlstp ->head == NULL) && (sessionlstp->tail == NULL)) {
		/* This list is empty, insert this element as first element */
		sessionlstp-> head = sessionp;
		sessionlstp->tail = sessionp;
	} else {
		ASSERT(sessionlstp->tail != NULL);
		ASSERT(sessionlstp->tail->next == NULL);
		sessionlstp->tail->next = sessionp;
		sessionlstp->tail = sessionp;
	}
	sessionlstp->num_items ++;
	mutex_exit(&sessionlstp->sessionlist_mutex);

	return (BNXEF_SUCCESS);
}


/*
 * This removes the given session from the hash bucket.
 */
int
bnxef_session_list_remove(bnxef_port_t *portp, bnxef_session_t *sessionp,
	uint8_t hashbucket)
{
	bnxef_session_list_t *sessionlstp;
	bnxef_session_t *sessioncp;
	bnxef_session_t *sessionprevp;
	int i, found = 0;
	int ret = BNXEF_FAILURE;
	bnxef_driver_state_t *dsp;

	ASSERT(portp != NULL);
	ASSERT(sessionp != NULL);

	dsp = portp->statep;
	ASSERT(dsp != NULL);

	sessionlstp = &portp->sessionhash[hashbucket];

	BNXEF_DEBUG(dsp, GENERAL, -1,
		"%s%d: Removin session 0x%p, hashbucket 0x%x, port 0x%p",
		BNXEF_DRIVER_NAME, dsp->instance, sessionp, hashbucket, portp);
	ASSERT(sessionlstp != NULL);

	ASSERT(sessionlstp->num_items > 0);
	mutex_enter(&sessionlstp->sessionlist_mutex);
	sessioncp = sessionlstp->head;
	ASSERT(sessioncp != NULL);
	sessionprevp = NULL;
	for (i = 0; i < sessionlstp->num_items; i++) {
		if (sessioncp == sessionp) {
			found++;
			break;
		}
		sessionprevp = sessioncp;
		sessioncp = sessioncp->next;
	}

	if (found) {
		if (sessionlstp->num_items == 1) {
			/* There is only one item */
			sessionlstp->head = NULL;
			sessionlstp->tail = NULL;
		} else {
			if (sessionprevp != NULL)
				sessionprevp->next = sessionp->next;
			else
				sessionlstp->head = sessionp->next;

			if (sessionlstp->tail == sessionp)
				sessionlstp->tail = sessionprevp;
		}
		sessionlstp->num_items--;
		ret = BNXEF_SUCCESS;
	}
	mutex_exit(&sessionlstp->sessionlist_mutex);

	sessionp->next = NULL;
	return (ret);
}


/*
 * This will allocate the session specific resources needed for offloading.
 */
int
bnxef_alloc_session_resources(bnxef_session_t *sessionp)
{
	bnxef_port_t *lportp;
	bnxef_port_t *rportp;
	bnxef_driver_state_t *dsp;
	int ret;
	uint32_t stage = 0;

	ASSERT(sessionp != NULL);

	lportp = sessionp->lport;
	rportp = sessionp->rport;

	if (lportp == NULL)
		return (BNXEF_FAILURE);


	dsp = lportp->statep;
	if ((dsp == NULL) || (rportp == NULL))
		return (BNXEF_FAILURE);
	ASSERT(dsp != NULL);
	ASSERT(lportp != NULL);
	ASSERT(rportp != NULL);

	BNXEF_DEBUG(dsp, GENERAL, -1,
		"%s%d: Connection specific resources allocated",
		BNXEF_DRIVER_NAME, dsp->instance);
	/* This will allocate the connection database resources */
	ret = bnxef_session_conn_db_res_alloc(sessionp);
	if (ret != BNXEF_SUCCESS)  {
		BNXEF_DEBUG(dsp, ERROR, lportp->id,
			"%s%d: Unable to allocate connection database resource\
			for session between 0x%x and 0x%x\n",
			BNXEF_DRIVER_NAME, dsp->instance, lportp->id,
				rportp->id);
		goto bnxef_session_res_alloc_error;
	}

	stage |= BNXEF_SESSION_RESOURCE_CONN_DB_ALLOC_COMPLETED;
	BNXEF_DEBUG(dsp, GENERAL, -1, "%s%d: Connection DB resource allocated",
		BNXEF_DRIVER_NAME, dsp->instance);

	ret = bnxef_session_sq_res_alloc(sessionp);
	if (ret != BNXEF_SUCCESS) {
		BNXEF_DEBUG(dsp, ERROR, lportp->id,
			"%s%d: Unable to allocate SQ resources for session\
				between 0x%x and 0x%x\n",
			BNXEF_DRIVER_NAME, dsp->instance, lportp->id,
			rportp->id);
		goto bnxef_session_res_alloc_error;
	}
	stage |= BNXEF_SESSION_RESOURCE_SQ_ALLOC_COMPLETED;
	BNXEF_DEBUG(dsp, GENERAL, -1,
		"%s%d: Send Queue Specific Resources allocated",
		BNXEF_DRIVER_NAME, dsp->instance);

	ret = bnxef_session_cq_res_alloc(sessionp);
	if (ret != BNXEF_SUCCESS) {
		BNXEF_DEBUG(dsp, ERROR, lportp->id,
			"%s%d: Unable to allocate CQ resources for session\
				between 0x%x and 0x%x\n",
			BNXEF_DRIVER_NAME, dsp->instance, lportp->id,
				rportp->id);
		goto bnxef_session_res_alloc_error;
	}

	stage |= BNXEF_SESSION_RESOURCE_CQ_ALLOC_COMPLETED;
	BNXEF_DEBUG(dsp, GENERAL, -1,
		"%s%d: CQ Specific resources Allocated",
		BNXEF_DRIVER_NAME, dsp->instance);

	ret = bnxef_session_rq_res_alloc(sessionp);
	if (ret != BNXEF_SUCCESS) {
		BNXEF_DEBUG(dsp, ERROR, lportp->id,
		"%s%d: Unable to allocate RQ resources for session\
			between 0x%x and 0x%x\n",
		BNXEF_DRIVER_NAME, dsp->instance, lportp->id,
			rportp->id);
		goto bnxef_session_res_alloc_error;
	}

	stage |= BNXEF_SESSION_RESOURCE_RQ_ALLOC_COMPLETED;
	BNXEF_DEBUG(dsp, GENERAL, -1,
		"%s%d: RQ specific resources allocated",
		BNXEF_DRIVER_NAME, dsp->instance);

	ret = bnxef_session_xferq_res_alloc(sessionp);
	if (ret != BNXEF_SUCCESS) {
		BNXEF_DEBUG(dsp, ERROR, lportp->id,
		"%s%d: Unable to allocate XFERQ resources for session\
			between 0x%x and 0x%x\n",
		BNXEF_DRIVER_NAME, dsp->instance, lportp->id,
			rportp->id);
		goto bnxef_session_res_alloc_error;
	}

	stage |= BNXEF_SESSION_RESOURCE_XFERQ_ALLOC_COMPLETED;
	BNXEF_DEBUG(dsp, GENERAL, -1,
		"%s%d: XFER QUeue specific resources allocated",
		BNXEF_DRIVER_NAME, dsp->instance);

	ret = bnxef_session_confq_res_alloc(sessionp);
	if (ret != BNXEF_SUCCESS) {
		BNXEF_DEBUG(dsp, ERROR, lportp->id,
		"%s%d: Unable to allocate CONFQ resources for session\
			between 0x%x and 0x%x\n",
		BNXEF_DRIVER_NAME, dsp->instance, lportp->id,
			rportp->id);
		goto bnxef_session_res_alloc_error;
	}

	stage |= BNXEF_SESSION_RESOURCE_CONFQ_ALLOC_COMPLETED;
	BNXEF_DEBUG(dsp, GENERAL, -1,
		"%s%d: CONF Queue specific resources allocated",
		BNXEF_DRIVER_NAME, dsp->instance);

	ret = bnxef_session_lcq_res_alloc(sessionp);
	if (ret != BNXEF_SUCCESS) {
		BNXEF_DEBUG(dsp, ERROR, lportp->id,
		"%s%d: Unable to allocate LCQ resources for session\
			between 0x%x and 0x%x\n",
		BNXEF_DRIVER_NAME, dsp->instance, lportp->id,
				rportp->id);
		goto bnxef_session_res_alloc_error;
	}

	stage |= BNXEF_SESSION_RESOURCE_LCQ_ALLOC_COMPLETED;
	BNXEF_DEBUG(dsp, GENERAL, -1,
		"%s%d: LCQ specific resources allocated",
		BNXEF_DRIVER_NAME, dsp->instance);

	return (BNXEF_SUCCESS);

bnxef_session_res_alloc_error:
	(void) bnxef_free_session_resources(sessionp, stage);
	return (BNXEF_FAILURE);
}

/*
 * The following will free the session resources allocated.
 */
int
bnxef_free_session_resources(bnxef_session_t *sessionp, uint32_t stage)
{
	int ret = BNXEF_FAILURE;
	ASSERT(sessionp != NULL);

	if ((stage & BNXEF_SESSION_RESOURCE_LCQ_ALLOC_COMPLETED) ==
		BNXEF_SESSION_RESOURCE_LCQ_ALLOC_COMPLETED) {
		ret = bnxef_session_lcq_res_free(sessionp, BNXEF_LCQ_RES_ALL);
	}

	if ((stage & BNXEF_SESSION_RESOURCE_CONFQ_ALLOC_COMPLETED) ==
		BNXEF_SESSION_RESOURCE_CONFQ_ALLOC_COMPLETED) {
		bnxef_session_confq_res_free(sessionp, BNXEF_CONFQ_ALL);

	}

	if ((stage & BNXEF_SESSION_RESOURCE_XFERQ_ALLOC_COMPLETED) ==
		BNXEF_SESSION_RESOURCE_XFERQ_ALLOC_COMPLETED) {
		ret = bnxef_xferq_res_free(sessionp, BNXEF_XFERQ_RES_ALL);

	}

	if ((stage & BNXEF_SESSION_RESOURCE_RQ_ALLOC_COMPLETED) ==
		BNXEF_SESSION_RESOURCE_RQ_ALLOC_COMPLETED) {
		ret = bnxef_rq_res_free(sessionp, BNXEF_RQ_ALL);

	}

	if ((stage & BNXEF_SESSION_RESOURCE_CQ_ALLOC_COMPLETED) ==
		BNXEF_SESSION_RESOURCE_CQ_ALLOC_COMPLETED) {
		ret = bnxef_cq_res_free(sessionp, BNXEF_CQ_RES_ALL);
	}

	if ((stage & BNXEF_SESSION_RESOURCE_SQ_ALLOC_COMPLETED) ==
		BNXEF_SESSION_RESOURCE_SQ_ALLOC_COMPLETED) {
		ret = bnxef_sq_res_free(sessionp, BNXEF_SQ_RES_ALL);
	}
	if ((stage & BNXEF_SESSION_RESOURCE_CONN_DB_ALLOC_COMPLETED) ==
		BNXEF_SESSION_RESOURCE_CONN_DB_ALLOC_COMPLETED) {
		ret = bnxef_session_conn_db_res_free(sessionp,
			BNXEF_CONN_DB_RES_ALL);
	}

	return (ret);

}
/*
 * This will allocate all SQ related session resources.
 */
int
bnxef_session_sq_res_alloc(bnxef_session_t *sessionp)
{
	bnxef_port_t *lportp;
	bnxef_port_t *rportp;
	bnxef_driver_state_t *dsp;
	ddi_dma_attr_t *sqdmaattr  = &bnxef_dma_attr;
	ddi_device_acc_attr_t *sqaccattr = &bnxef_dev_acc_attr;
	int stage = 0, ret;
	    uint32_t req_size;

	ASSERT(sessionp != NULL);
	lportp = sessionp->lport;
	rportp = sessionp->rport;

	if (lportp == NULL)
		return (BNXEF_FAILURE);

	dsp = lportp->statep;
	if ((dsp == NULL) || (rportp == NULL))
		return (BNXEF_FAILURE);
	ASSERT(dsp != NULL);
	ASSERT(lportp != NULL);
	ASSERT(rportp != NULL);

	BNXEF_DEBUG(dsp, GENERAL, lportp->id,
		"%s%d: SQ resources Allocatedd for Session between 0x%x-->0x%x",
		BNXEF_DRIVER_NAME, dsp->instance, lportp->id, rportp->id);

	ret = ddi_dma_alloc_handle(dsp->dip, sqdmaattr, DDI_DMA_SLEEP, NULL,
		&sessionp->sq_dma_handle);
	if (ret != DDI_SUCCESS) {
		BNXEF_DEBUG(dsp, ERROR, lportp->id,
			"%s%d: Unable to allocate SQ dma handle\n",
			BNXEF_DRIVER_NAME, dsp->instance);
		goto bnxef_sq_res_alloc_error;
	}

	stage |= BNXEF_SQ_RES_DMA_HANDLE_ALLOCATED;
	BNXEF_DEBUG(dsp, GENERAL, lportp->id,
		"%s%d: SQ DMA handle allocated for  Sessio between 0x%x-->0x%x",
			BNXEF_DRIVER_NAME, dsp->instance, lportp->id,
				rportp->id);

	req_size = (sessionp->max_sqes + 8) * sizeof (struct fcoe_sqe);
	req_size = (req_size + (BNXEF_PAGE_SIZE -1)) & BNXEF_PAGE_ROUND_MASK;

	ret = ddi_dma_mem_alloc(sessionp->sq_dma_handle,
			req_size,
			sqaccattr, DDI_DMA_CONSISTENT, DDI_DMA_SLEEP, NULL,
			&sessionp->sq_addr,
			(unsigned long *)&sessionp->sq_length,
			&sessionp->sq_acc_handle);
	if (ret != DDI_SUCCESS) {
		BNXEF_DEBUG(dsp, ERROR, lportp->id,
			"%s%d: Unable to allocate dma memory\n",
				BNXEF_DRIVER_NAME, dsp->instance);
		goto bnxef_sq_res_alloc_error;
	}

	stage |=  BNXEF_SQ_DMA_MEM_ALLOCATED;

	memset(sessionp->sq_addr, 0, sessionp->sq_length);
	BNXEF_DEBUG(dsp, GENERAL, lportp->id,
		"%s%d: SQ DMA memory allocated for Session between 0x%x-->0x%x",
		BNXEF_DRIVER_NAME, dsp->instance, lportp->id, rportp->id);

	/* Now map the buffer for use with firmware */
	ret = ddi_dma_addr_bind_handle(sessionp->sq_dma_handle,
		NULL, sessionp->sq_addr, sessionp->sq_length,
		DDI_DMA_CONSISTENT | DDI_DMA_RDWR, DDI_DMA_SLEEP, NULL,
		&sessionp->sq_cookie, &sessionp->sq_cookie_cnt);
	if (ret != DDI_DMA_MAPPED) {
		BNXEF_DEBUG(dsp, ERROR, lportp->id,
				"%s%d: Unable to bind dma mem with handle\n",
			BNXEF_DRIVER_NAME, dsp->instance);
		goto bnxef_sq_res_alloc_error;
	}
	ddi_dma_sync(sessionp->sq_dma_handle, 0, sessionp->sq_length,
		DDI_DMA_SYNC_FORDEV);
	/* We have indicated no scatter gather allowed */
	ASSERT(sessionp->sq_cookie_cnt == 1);
	stage |= BNXEF_SQ_DMA_MEM_BOUND;
	BNXEF_DEBUG(dsp, GENERAL, lportp->id,
		"%s%d: SQ DMA memory bound for Session between 0x%x-->0x%x",
		BNXEF_DRIVER_NAME, dsp->instance, lportp->id, rportp->id);

	return (BNXEF_SUCCESS);

bnxef_sq_res_alloc_error:
	(void) bnxef_sq_res_free(sessionp, stage);
	return (BNXEF_FAILURE);
}


/*
 * This will free the SQ resources
 */

int
bnxef_sq_res_free(bnxef_session_t *sessionp, uint32_t stage)
{
	int ret = BNXEF_FAILURE;

	ASSERT(sessionp != NULL);

	if ((stage & BNXEF_SQ_DMA_MEM_BOUND) == BNXEF_SQ_DMA_MEM_BOUND) {
		ret = ddi_dma_unbind_handle(sessionp->sq_dma_handle);
	}

	if ((stage & BNXEF_SQ_DMA_MEM_ALLOCATED)
			== BNXEF_SQ_DMA_MEM_ALLOCATED) {
		ddi_dma_mem_free(&sessionp->sq_acc_handle);
	}

	if ((stage & BNXEF_SQ_RES_DMA_HANDLE_ALLOCATED) ==
			BNXEF_SQ_RES_DMA_HANDLE_ALLOCATED) {
		(void) ddi_dma_free_handle(&sessionp->sq_dma_handle);
	}

	return (ret);
}


/*
 * The following allocates the CQ related resources
 */
int
bnxef_session_cq_res_alloc(bnxef_session_t *sessionp)
{
	bnxef_port_t *lportp;
	bnxef_port_t *rportp;
	bnxef_driver_state_t *dsp;
	ddi_dma_attr_t *cqdmaattr  = &bnxef_dma_attr;
	ddi_device_acc_attr_t *cqaccattr = &bnxef_dev_acc_attr;
	int stage = 0, ret;
	    uint32_t req_size;

	ASSERT(sessionp != NULL);
	lportp = sessionp->lport;
	rportp = sessionp->rport;

	if (lportp == NULL)
		return (BNXEF_FAILURE);

	dsp = lportp->statep;
	if ((dsp == NULL) || (rportp == NULL))
		return (BNXEF_FAILURE);

	ASSERT(dsp != NULL);
	ASSERT(lportp != NULL);
	ASSERT(rportp != NULL);
	BNXEF_DEBUG(dsp, GENERAL, lportp->id,
	"%s%d: CQ resources Allocatedd for Session between 0x%x-->0x%x",
		BNXEF_DRIVER_NAME, dsp->instance, lportp->id, rportp->id);

	ret = ddi_dma_alloc_handle(dsp->dip, cqdmaattr, DDI_DMA_SLEEP, NULL,
		&sessionp->cq_dma_handle);
	if (ret != DDI_SUCCESS) {
		BNXEF_DEBUG(dsp, ERROR, lportp->id,
			"%s%d: Unable to allocate SQ dma handle\n",
			BNXEF_DRIVER_NAME, dsp->instance);
		goto bnxef_cq_res_alloc_error;
	}

	stage |= BNXEF_CQ_RES_DMA_HANDLE_ALLOCATED;

	BNXEF_DEBUG(dsp, GENERAL, lportp->id,
		"%s%d; CQ DMA handle allocated for session between 0x%x-->0x%x",
		BNXEF_DRIVER_NAME, dsp->instance, lportp->id, rportp->id);

	req_size = (sessionp->max_cqes + 8) * sizeof (struct fcoe_cqe);
	req_size = (req_size + (BNXEF_PAGE_SIZE -1)) & BNXEF_PAGE_ROUND_MASK;

	ret = ddi_dma_mem_alloc(sessionp->cq_dma_handle,
			req_size,
			cqaccattr, DDI_DMA_CONSISTENT, DDI_DMA_SLEEP, NULL,
			&sessionp->cq_addr,
			(unsigned long *)&sessionp->cq_length,
			&sessionp->cq_acc_handle);
	if (ret != DDI_SUCCESS) {
		BNXEF_DEBUG(dsp, ERROR, lportp->id,
			"%s%d: Unable to allocate dma memory\n",
				BNXEF_DRIVER_NAME, dsp->instance);
		goto bnxef_cq_res_alloc_error;
	}

	stage |=  BNXEF_CQ_DMA_MEM_ALLOCATED;

	memset(sessionp->cq_addr, 0, sessionp->cq_length);
	BNXEF_DEBUG(dsp, GENERAL, lportp->id,
		"%s%d; CQ DMA mem allocated for session between 0x%x-->0x%x",
		BNXEF_DRIVER_NAME, dsp->instance, lportp->id, rportp->id);

	/* Now map the buffer for use with firmware */
	ret = ddi_dma_addr_bind_handle(sessionp->cq_dma_handle,
		NULL, sessionp->cq_addr, sessionp->cq_length,
		DDI_DMA_CONSISTENT | DDI_DMA_RDWR, DDI_DMA_SLEEP, NULL,
		&sessionp->cq_cookie, &sessionp->cq_cookie_cnt);
	if (ret != DDI_DMA_MAPPED) {
		BNXEF_DEBUG(dsp, ERROR, lportp->id,
				"%s%d: Unable to bind dma mem with handle\n",
			BNXEF_DRIVER_NAME, dsp->instance);
		goto bnxef_cq_res_alloc_error;
	}
	/* We have indicated no scatter gather allowed */
	ASSERT(sessionp->cq_cookie_cnt == 1);
	ddi_dma_sync(sessionp->cq_dma_handle, 0, sessionp->cq_length,
		DDI_DMA_SYNC_FORDEV);
	stage |= BNXEF_CQ_DMA_MEM_BOUND;
	BNXEF_DEBUG(dsp, GENERAL, lportp->id,
	"%s%d; CQ DMA mem bound allocated for session between 0x%x-->0x%x",
		BNXEF_DRIVER_NAME, dsp->instance, lportp->id, rportp->id);

	return (BNXEF_SUCCESS);

bnxef_cq_res_alloc_error:
	(void) bnxef_cq_res_free(sessionp, stage);
	return (BNXEF_FAILURE);
}


/*
 * This will free the CQ resources
 */
int
bnxef_cq_res_free(bnxef_session_t *sessionp, uint32_t stage)
{
	int ret = BNXEF_FAILURE;

	ASSERT(sessionp != NULL);

	if ((stage & BNXEF_CQ_DMA_MEM_BOUND) == BNXEF_CQ_DMA_MEM_BOUND) {
		ret = ddi_dma_unbind_handle(sessionp->cq_dma_handle);
	}

	if ((stage & BNXEF_CQ_DMA_MEM_ALLOCATED)
			== BNXEF_CQ_DMA_MEM_ALLOCATED) {
		ddi_dma_mem_free(&sessionp->cq_acc_handle);
	}

	if ((stage & BNXEF_CQ_RES_DMA_HANDLE_ALLOCATED) ==
			BNXEF_CQ_RES_DMA_HANDLE_ALLOCATED) {
		(void) ddi_dma_free_handle(&sessionp->cq_dma_handle);
	}

	return (ret);
}


/*
 * This will allocate the RQ and RQ PBL and set the page buffer list
 */
int
bnxef_session_rq_res_alloc(bnxef_session_t *sessionp)
{
	bnxef_port_t *lportp;
	bnxef_port_t *rportp;
	bnxef_driver_state_t *dsp;
	ddi_dma_attr_t rqdmaattr =  {
			DMA_ATTR_V0,		/* Version number */
			(uint64_t)0x0,		/* low address */
			(uint64_t)0xffffffffffffffff, /* High address */
			(uint64_t)0xffffffff,		/* DMA counter max */
			(uint64_t)0x1,		/* byte aligned  */
			0x100, /* 256 bytes of bursts */
			0x1,	/* Minimum transfer size */
			(uint64_t)0xffffffff, /* MAximum transfer size */
			(uint64_t)0xffffffff, /* Donot cross 4 GB boundary */
			1,	/* no scatter gather list allowed */
			128,	/* Granularity */
			0	/* DMA flags */
	};
	ddi_dma_attr_t rqpbldmaattr = {
			DMA_ATTR_V0,		/* Version number */
			(uint64_t)0x0,		/* low address */
			(uint64_t)0xffffffffffffffff, /* High address */
			(uint64_t)0xffffffff,		/* DMA counter max */
			(uint64_t)0x1,		/* 4K alignment */
			0x8,	 /* 8 bytes of bursts */
			0x1,	/* Minimum transfer size */
			(uint64_t)0xffffffff, /* MAximum transfer size */
			(uint64_t)0xFFFFFfff,  /* Donot cross 4GB Boundary */
			1,	/* no scatter gather list supported  */
			8,	/* Granularity */
			0	/* DMA flags */
	};

	ddi_device_acc_attr_t *rqaccattr = &bnxef_dev_acc_attr;
	ddi_device_acc_attr_t *rqpblaccattr = &bnxef_dev_acc_attr;
	int stage = 0, ret, i;
	int pblsize;
	uint32_t dma_size;
	struct regpair *pbl;

	ASSERT(sessionp != NULL);
	lportp = sessionp->lport;
	rportp = sessionp->rport;
	if (lportp == NULL)
		return (BNXEF_FAILURE);

	dsp = lportp->statep;
	if ((dsp == NULL) || (rportp == NULL))
		return (BNXEF_FAILURE);

	ASSERT(dsp != NULL);
	ASSERT(lportp != NULL);
	ASSERT(rportp != NULL);
	BNXEF_DEBUG(dsp, GENERAL, lportp->id,
	"%s%d: RQ resources Allocated for Session between 0x%x-->0x%x",
		BNXEF_DRIVER_NAME, dsp->instance, lportp->id, rportp->id);
	ret = ddi_dma_alloc_handle(dsp->dip, &rqdmaattr, DDI_DMA_SLEEP, NULL,
			&sessionp->rq_dma_handle);
	if (ret != DDI_SUCCESS) {
		BNXEF_DEBUG(dsp, ERROR, lportp->id,
				"%s%d: Unable allocate DMA handle\n",
			BNXEF_DRIVER_NAME, dsp->instance);
		goto bnxef_rq_res_alloc_error;
	}

	stage |= BNXEF_RQ_DMA_HANDLE_ALLOCATED;
	BNXEF_DEBUG(dsp, GENERAL, lportp->id,
		"%s%d: RQ DMA handle allocated for session between 0x%x-->0x%x",
		BNXEF_DRIVER_NAME, dsp->instance, lportp->id, rportp->id);

	BNXEF_DEBUG(dsp, GENERAL, lportp->id,
		"%s%d: Requested Size to allocated 0x%x",
		BNXEF_DRIVER_NAME, dsp->instance,
			sessionp->max_rqes * BNXEF_RQ_WQE_SIZE);
	/*
	 * This will allocate the required memory, which is physically
	 * contiguous
	 */
	ret = ddi_dma_mem_alloc(sessionp->rq_dma_handle,
		sessionp->max_rqes * BNXEF_RQ_WQE_SIZE,
		rqaccattr,
		DDI_DMA_CONSISTENT, DDI_DMA_SLEEP, NULL,
		&sessionp->rq_addr, (unsigned long *)&sessionp->rq_length,
		&sessionp->rq_acc_handle);
	if (ret != DDI_SUCCESS) {
		BNXEF_DEBUG(dsp, ERROR, lportp->id,
				"%s%d:Unable to allocate RQ DMA mem\n",
			BNXEF_DRIVER_NAME, dsp->instance);

		goto bnxef_rq_res_alloc_error;
	}

	stage |= BNXEF_RQ_DMA_MEM_ALLOCATED;
	memset(sessionp->rq_addr, 0, sessionp->rq_length);
	BNXEF_DEBUG(dsp, GENERAL, lportp->id,
		"%s%d: RQ DMA memory allocated for session between 0x%x-->0x%x",
		BNXEF_DRIVER_NAME, dsp->instance, lportp->id, rportp->id);

	/* Now map or pin this buffer for subsequent DMA based data movement */
	ret = ddi_dma_addr_bind_handle(sessionp->rq_dma_handle,
		NULL, sessionp->rq_addr, sessionp->rq_length,
		DDI_DMA_CONSISTENT | DDI_DMA_RDWR, DDI_DMA_SLEEP, NULL,
		&sessionp->rq_cookie, &sessionp->rq_cookie_cnt);

	if (ret != DDI_DMA_MAPPED) {
		BNXEF_DEBUG(dsp, ERROR, lportp->id,
				"%s%d: Unable to bind dma mem with handle\n",
			BNXEF_DRIVER_NAME, dsp->instance);
		goto bnxef_rq_res_alloc_error;
	}
	/* We have indicated no scatter gather allowed */
	ASSERT(sessionp->rq_cookie_cnt == 1);
	ddi_dma_sync(sessionp->rq_dma_handle, 0,
		sessionp->rq_length, DDI_DMA_SYNC_FORDEV);
	stage |= BNXEF_RQ_DMA_MEM_BOUND;
	BNXEF_DEBUG(dsp, GENERAL, lportp->id,
		"%s%d: RQ DMA memory bound for session between 0x%x-->0x%x",
		BNXEF_DRIVER_NAME, dsp->instance, lportp->id, rportp->id);

	/*
	 * Now allocate a page size worth of page descriptor to describe the
	 * allocated buffer, just above
	 */
	ret = ddi_dma_alloc_handle(dsp->dip, &rqpbldmaattr, DDI_DMA_SLEEP,
			NULL, &sessionp->rq_pbl_dma_handle);
	if (ret != DDI_SUCCESS) {
		BNXEF_DEBUG(dsp, ERROR, lportp->id,
			"%s%d: Unable allocate task ctx handle\n",
			BNXEF_DRIVER_NAME, dsp->instance);
		goto bnxef_rq_res_alloc_error;
	}
	stage |= BNXEF_RQ_PBL_HANDLE_ALLOCATED;
	BNXEF_DEBUG(dsp, GENERAL, lportp->id,
		"%s%d: RQ PBL DMA handle for session between 0x%x-->0x%x",
		BNXEF_DRIVER_NAME, dsp->instance, lportp->id, rportp->id);

	pblsize = (sessionp->rq_length / BNXEF_PAGE_SIZE) * sizeof (void *);
	pblsize = (pblsize + (BNXEF_PAGE_SIZE -1)) & BNXEF_PAGE_ROUND_MASK;


	/* Allocate a page size worth of DMA memory */
	ret = ddi_dma_mem_alloc(sessionp->rq_pbl_dma_handle, pblsize,
		rqpblaccattr, DDI_DMA_CONSISTENT, DDI_DMA_SLEEP, NULL,
		&sessionp->rq_pbl_addr,
		(unsigned long *)&sessionp->rq_pbl_length,
		&sessionp->rq_pbl_acc_handle);
	if (ret != DDI_SUCCESS) {
		BNXEF_DEBUG(dsp, ERROR, lportp->id,
				"%s%d:Unable to allocate RQ PBL DMA mem\n",
			BNXEF_DRIVER_NAME, dsp->instance);
		goto bnxef_rq_res_alloc_error;
	}
	stage |= BNXEF_RQ_PBL_MEM_ALLOCATED;

	memset((uint8_t *)sessionp->rq_pbl_addr, 0, sessionp->rq_pbl_length);
	BNXEF_DEBUG(dsp, GENERAL, lportp->id,
	"%s%d: RQ PBL DMA memory allocated for session between 0x%x-->0x%x",
		BNXEF_DRIVER_NAME, dsp->instance, lportp->id, rportp->id);


	/* MAp this page descriptor table */
	ret = ddi_dma_addr_bind_handle(sessionp->rq_pbl_dma_handle,
		NULL, sessionp->rq_pbl_addr, sessionp->rq_pbl_length,
		DDI_DMA_CONSISTENT | DDI_DMA_RDWR, DDI_DMA_SLEEP, NULL,
		&sessionp->rq_pbl_dma_cookie, &sessionp->rq_pbl_cookie_cnt);
	if (ret != DDI_DMA_MAPPED) {
		BNXEF_DEBUG(dsp, ERROR, lportp->id,
			"%s%d: Unable to bind RQ PBL dma mem with handle\n",
			BNXEF_DRIVER_NAME, dsp->instance);
		goto bnxef_rq_res_alloc_error;
	}
	ASSERT(sessionp->rq_pbl_cookie_cnt == 1);
	stage |= BNXEF_RQ_PBL_MEM_BOUND;
	BNXEF_DEBUG(dsp, GENERAL, lportp->id,
	"%s%d: RQ PBL DMA memory bound for session between 0x%x-->0x%x",
		BNXEF_DRIVER_NAME, dsp->instance, lportp->id, rportp->id);

	/* LINTED: alignment */
	pbl = (struct regpair *)sessionp->rq_pbl_addr;
	dma_size = sessionp->rq_cookie.dmac_size;
	i = 0;
	while (dma_size) {
		uint64_t addr;
		addr = sessionp->rq_cookie.dmac_laddress + i * BNXEF_PAGE_SIZE;
		pbl->hi = CPU_TO_LE32((addr >> 32));
		pbl->lo = CPU_TO_LE32((addr & 0xffffffff));
		BNXEF_DEBUG(dsp, GENERAL, lportp->id,
			"%s%d: Pbl Address 0x%x%x",
			BNXEF_DRIVER_NAME, dsp->instance, pbl->hi, pbl->lo);
		if (dma_size < BNXEF_PAGE_SIZE) break;
		dma_size -= BNXEF_PAGE_SIZE;
		i++;
		pbl++;
	}

	ddi_dma_sync(sessionp->rq_pbl_dma_handle, 0,
		sessionp->rq_pbl_length, DDI_DMA_SYNC_FORDEV);

	stage |= BNXEF_RQ_PBL_SETUP;
	BNXEF_DEBUG(dsp, GENERAL, lportp->id,
		"%s%d: RQ PBL Setup for Session between 0x%x-->0x%x",
		BNXEF_DRIVER_NAME, dsp->instance, lportp->id, rportp->id);

	return (BNXEF_SUCCESS);
bnxef_rq_res_alloc_error:
	(void) bnxef_rq_res_free(sessionp, stage);
	return (BNXEF_FAILURE);
}

/*
 * This will free the rq resources allocated
 */
int
bnxef_rq_res_free(bnxef_session_t *sessionp, uint32_t stage)
{
	int ret = BNXEF_FAILURE;

	ASSERT(sessionp != NULL);

	/*
	 * Freeing of DMA resources should go in this order.
	 * This is reverse of what is done when we allocated the resources
	 * in the alloc routine.
	 */
	if ((stage & BNXEF_RQ_PBL_MEM_BOUND) ==
		BNXEF_RQ_PBL_MEM_BOUND) {
		ret = ddi_dma_unbind_handle(sessionp->rq_pbl_dma_handle);
	}

	if ((stage & BNXEF_RQ_PBL_MEM_ALLOCATED) ==
		BNXEF_RQ_PBL_MEM_ALLOCATED) {
		ddi_dma_mem_free(&sessionp->rq_pbl_acc_handle);
	}

	if ((stage & BNXEF_RQ_PBL_HANDLE_ALLOCATED) ==
		BNXEF_RQ_PBL_HANDLE_ALLOCATED) {
		ddi_dma_free_handle(&sessionp->rq_pbl_dma_handle);
	}

	/* Now release all the resources allocated with RQ DMA */
	if ((stage & BNXEF_RQ_DMA_MEM_BOUND) ==
		BNXEF_RQ_DMA_MEM_BOUND) {
		ret = ddi_dma_unbind_handle(sessionp->rq_dma_handle);
	}

	if ((stage & BNXEF_RQ_DMA_MEM_ALLOCATED) ==
		BNXEF_RQ_DMA_MEM_ALLOCATED) {
		ddi_dma_mem_free(&sessionp->rq_acc_handle);
	}

	if ((stage & BNXEF_RQ_DMA_HANDLE_ALLOCATED) ==
		BNXEF_RQ_DMA_HANDLE_ALLOCATED) {
		ddi_dma_free_handle(&sessionp->rq_dma_handle);
	}

	return (ret);
}

/*
 * THis will free the confq resources
 */
void
bnxef_session_confq_res_free(bnxef_session_t *sessionp, uint32_t stage)
{

	ASSERT(sessionp != NULL);

	if ((stage & BNXEF_CONFQ_DMA_HANDLE_ALLOCATED) ==
		BNXEF_CONFQ_DMA_HANDLE_ALLOCATED) {
		(void) ddi_dma_unbind_handle(sessionp->confq_pbl_dma_handle);

	}

	if ((stage & BNXEF_CONFQ_DMA_MEM_ALLOCATED) ==
		BNXEF_CONFQ_DMA_MEM_ALLOCATED) {
		ddi_dma_mem_free(&sessionp->confq_pbl_acc_handle);

	}

	if ((stage & BNXEF_CONFQ_DMA_MEM_BOUND) ==
		BNXEF_CONFQ_DMA_MEM_BOUND) {
		ddi_dma_free_handle(&sessionp->confq_pbl_dma_handle);

	}

	if ((stage & BNXEF_CONFQ_PBL_HANDLE_ALLOCATED) ==
		BNXEF_CONFQ_PBL_HANDLE_ALLOCATED) {
		(void) ddi_dma_unbind_handle(sessionp->confq_dma_handle);

	}

	if ((stage &  BNXEF_CONFQ_PBL_MEM_ALLOCATED) ==
			BNXEF_CONFQ_PBL_MEM_ALLOCATED) {
		ddi_dma_mem_free(&sessionp->confq_acc_handle);

	}

	if ((stage & BNXEF_CONFQ_PBL_MEM_BOUND) ==
		BNXEF_CONFQ_PBL_MEM_BOUND) {
		ddi_dma_free_handle(&sessionp->confq_dma_handle);

	}

}

/*
 * This will allocate the conf q resources.
 */
int
bnxef_session_confq_res_alloc(bnxef_session_t *sessionp)
{
	bnxef_port_t *lportp;
	bnxef_port_t *rportp;
	bnxef_driver_state_t *dsp;
	ddi_dma_attr_t confqdmaattr =  {
			DMA_ATTR_V0,		/* Version number */
			(uint64_t)0x0,		/* low address */
			(uint64_t)0xffffffffffffffff, /* High address */
			(uint64_t)0xffffffff,		/* DMA counter max */
			(uint64_t)0x1,		/* byte aligned  */
			0x100, /* 256 bytes of bursts */
			0x1,	/* Minimum transfer size */
			(uint64_t)0xffffffff, /* MAximum transfer size */
			(uint64_t)0xffffffff, /* Donot cross 4 GB boundary */
			1,	/* no scatter gather list allowed */
			128,	/* Granularity */
			0	/* DMA flags */
	};
	ddi_dma_attr_t confqpbldmaattr = {
			DMA_ATTR_V0,		/* Version number */
			(uint64_t)0x0,		/* low address */
			(uint64_t)0xffffffffffffffff, /* High address */
			(uint64_t)0xffffffff,		/* DMA counter max */
			(uint64_t)0x1,		/* byte alignment */
			0x8,	 /* 8 bytes of bursts */
			0x1,	/* Minimum transfer size */
			(uint64_t)0xffffffff, /* MAximum transfer size */
			(uint64_t)0xffffffff, /* Donot cross 4K Boundary */
			1,	/* no scatter gather list supported  */
			8,	/* Granularity */
			0	/* DMA flags */
	};

	ddi_device_acc_attr_t *confqaccattr = &bnxef_dev_acc_attr;
	ddi_device_acc_attr_t *confqpblaccattr = &bnxef_dev_acc_attr;
	int stage = 0, ret;
	struct regpair *pbl;
	uint32_t pblsize, dma_size;
	int i;
	    uint32_t req_size;

	ASSERT(sessionp != NULL);
	lportp = sessionp->lport;
	rportp = sessionp->rport;

	if (lportp == NULL)
		return (BNXEF_FAILURE);

	dsp = lportp->statep;
	if ((dsp == NULL) || (rportp == NULL))
		return (BNXEF_FAILURE);
	ASSERT(dsp != NULL);
	ASSERT(lportp != NULL);
	ASSERT(rportp != NULL);
	BNXEF_DEBUG(dsp, GENERAL, lportp->id,
	"%s%d: CONFQ resources Allocated for Session between 0x%x-->0x%x",
		BNXEF_DRIVER_NAME, dsp->instance, lportp->id, rportp->id);
	ret = ddi_dma_alloc_handle(dsp->dip, &confqdmaattr, DDI_DMA_SLEEP,
			NULL, &sessionp->confq_dma_handle);
	if (ret != DDI_SUCCESS) {
		BNXEF_DEBUG(dsp, ERROR, lportp->id,
				"%s%d: Unable allocate DMA handle\n",
			BNXEF_DRIVER_NAME, dsp->instance);
		goto bnxef_confq_res_alloc_error;
	}

	stage |= BNXEF_CONFQ_DMA_HANDLE_ALLOCATED;
	BNXEF_DEBUG(dsp, GENERAL, lportp->id,
	"%s%d: CONFQ DMA handle allocated for session between 0x%x-->0x%x",
		BNXEF_DRIVER_NAME, dsp->instance, lportp->id, rportp->id);
	req_size = (sessionp->max_sqes  + 8)* sizeof (struct fcoe_confqe);
	req_size = (req_size + (BNXEF_PAGE_SIZE -1)) & BNXEF_PAGE_ROUND_MASK;
	/*
	 * This will allocate the required memory, which is physically
	 * contiguous
	 */
	ret = ddi_dma_mem_alloc(sessionp->confq_dma_handle,
		req_size,
		confqaccattr,
		DDI_DMA_CONSISTENT, DDI_DMA_SLEEP, NULL,
		&sessionp->confq_addr,
		(unsigned long *)&sessionp->confq_length,
		&sessionp->confq_acc_handle);
	if (ret != DDI_SUCCESS) {
		BNXEF_DEBUG(dsp, ERROR, lportp->id,
				"%s%d:Unable to allocate CONFQ DMA mem\n",
			BNXEF_DRIVER_NAME, dsp->instance);
		goto bnxef_confq_res_alloc_error;
	}

	stage |= BNXEF_CONFQ_DMA_MEM_ALLOCATED;
	memset(sessionp->confq_addr, 0, sessionp->confq_length);
	BNXEF_DEBUG(dsp, GENERAL, lportp->id,
	"%s%d: CONFQ DMA memory allocated for session between 0x%x-->0x%x",
		BNXEF_DRIVER_NAME, dsp->instance, lportp->id, rportp->id);

	/* Now map or pin this buffer for subsequent DMA based data movement */
	ret = ddi_dma_addr_bind_handle(sessionp->confq_dma_handle,
		NULL, sessionp->confq_addr, sessionp->confq_length,
		DDI_DMA_CONSISTENT | DDI_DMA_RDWR, DDI_DMA_SLEEP, NULL,
		&sessionp->confq_cookie, &sessionp->confq_cookie_cnt);

	if (ret != DDI_DMA_MAPPED) {
		BNXEF_DEBUG(dsp, ERROR, lportp->id,
				"%s%d: Unable to bind dma mem with handle\n",
			BNXEF_DRIVER_NAME, dsp->instance);
		goto bnxef_confq_res_alloc_error;
	}
	/* We have indicated no scatter gather allowed */
	ASSERT(sessionp->confq_cookie_cnt == 1);

	ddi_dma_sync(sessionp->confq_dma_handle, 0,
		sessionp->confq_length, DDI_DMA_SYNC_FORDEV);

	stage |= BNXEF_CONFQ_DMA_MEM_BOUND;

	BNXEF_DEBUG(dsp, GENERAL, lportp->id,
		"%s%d: CONFQ DMA memory bound for session between 0x%x-->0x%x",
		BNXEF_DRIVER_NAME, dsp->instance, lportp->id, rportp->id);
	/*
	 * Now allocate a page size worth of page descriptor to describe the
	 * allocated buffer, just above
	 */
	ret = ddi_dma_alloc_handle(dsp->dip, &confqpbldmaattr, DDI_DMA_SLEEP,
			NULL, &sessionp->confq_pbl_dma_handle);
	if (ret != DDI_SUCCESS) {
		BNXEF_DEBUG(dsp, ERROR, lportp->id,
			"%s%d: Unable allocate task ctx handle\n",
			BNXEF_DRIVER_NAME, dsp->instance);
		goto bnxef_confq_res_alloc_error;
	}
	stage |= BNXEF_CONFQ_PBL_HANDLE_ALLOCATED;

	BNXEF_DEBUG(dsp, GENERAL, lportp->id,
		"%s%d: CONFQ PBL DMA handle for session between 0x%x-->0x%x",
		BNXEF_DRIVER_NAME, dsp->instance, lportp->id, rportp->id);
	pblsize = (sessionp->confq_length / BNXEF_PAGE_SIZE) * sizeof (void *);
	pblsize = ((pblsize + (BNXEF_PAGE_SIZE -1)) & BNXEF_PAGE_MASK) + 1;
	BNXEF_DEBUG(dsp, GENERAL, lportp->id,
		"%s%d: Pblsize requested 0x%x\n", BNXEF_DRIVER_NAME,
		dsp->instance, pblsize);

	/* Allocate a page size worth of DMA memory */
	ret = ddi_dma_mem_alloc(sessionp->confq_pbl_dma_handle, pblsize,
		confqpblaccattr, DDI_DMA_CONSISTENT, DDI_DMA_SLEEP, NULL,
		&sessionp->confq_pbl_addr,
		(unsigned long *)&sessionp->confq_pbl_length,
		&sessionp->confq_pbl_acc_handle);
	if (ret != DDI_SUCCESS) {
		BNXEF_DEBUG(dsp, ERROR, lportp->id,
			"%s%d:Unable to allocate CONFQ PBL DMA mem\n",
			BNXEF_DRIVER_NAME, dsp->instance);
		goto bnxef_confq_res_alloc_error;
	}
	stage |= BNXEF_CONFQ_PBL_MEM_ALLOCATED;

	memset((uint8_t *)sessionp->confq_pbl_addr, 0,
			sessionp->confq_pbl_length);
	BNXEF_DEBUG(dsp, GENERAL, lportp->id,
	"%s%d: CONFQ PBL DMA memory allocated for session between 0x%x-->0x%x",
		BNXEF_DRIVER_NAME, dsp->instance, lportp->id, rportp->id);

	/* MAp this page descriptor table */
	ret = ddi_dma_addr_bind_handle(sessionp->confq_pbl_dma_handle,
		NULL, sessionp->confq_pbl_addr, sessionp->confq_pbl_length,
		DDI_DMA_CONSISTENT | DDI_DMA_RDWR, DDI_DMA_SLEEP, NULL,
		&sessionp->confq_pbl_cookie, &sessionp->confq_pbl_cookie_cnt);
	if (ret != DDI_DMA_MAPPED) {
		BNXEF_DEBUG(dsp, ERROR, lportp->id,
		"%s%d: Unable to bind CONFQ PBL dma mem with handle %d\n",
			BNXEF_DRIVER_NAME, dsp->instance, ret);
		goto bnxef_confq_res_alloc_error;
	}
	ASSERT(sessionp->confq_pbl_cookie_cnt == 1);
	stage |= BNXEF_CONFQ_PBL_MEM_BOUND;
	BNXEF_DEBUG(dsp, GENERAL, lportp->id,
	"%s%d: CONFQ PBL DMA memory bound for session between 0x%x-->0x%x",
		BNXEF_DRIVER_NAME, dsp->instance, lportp->id, rportp->id);

	/* LINTED: alignment */
	pbl = (struct regpair *)sessionp->confq_pbl_addr;
	dma_size = sessionp->confq_cookie.dmac_size;
	i = 0;
	while (dma_size) {
		uint64_t addr;
		addr =
		(sessionp->confq_cookie.dmac_laddress + i * BNXEF_PAGE_SIZE);
		pbl->hi = CPU_TO_LE32((addr >> 32));
		pbl->lo = CPU_TO_LE32((addr & 0xFFFFFFFF));
		BNXEF_DEBUG(dsp, GENERAL, lportp->id,
			"%s%d: Pbl Address 0x%x%x",
			BNXEF_DRIVER_NAME, dsp->instance, pbl->hi, pbl->lo);
		if (dma_size < BNXEF_PAGE_SIZE) break;
		dma_size -= BNXEF_PAGE_SIZE;
		i++;
		pbl++;
	}
	ddi_dma_sync(sessionp->confq_pbl_dma_handle, 0,
		sessionp->confq_pbl_length, DDI_DMA_SYNC_FORDEV);
	stage |= BNXEF_CONFQ_PBL_SETUP;
	BNXEF_DEBUG(dsp, GENERAL, lportp->id,
		"%s%d: CONFQ PBL Setup for Session between 0x%x-->0x%x",
		BNXEF_DRIVER_NAME, dsp->instance, lportp->id, rportp->id);

	return (BNXEF_SUCCESS);
bnxef_confq_res_alloc_error:
	(void) bnxef_session_confq_res_free(sessionp, stage);
	return (BNXEF_FAILURE);
}

/*
 * The following allocates the resources needed for XFERQ.
 */
int
bnxef_session_xferq_res_alloc(bnxef_session_t *sessionp)
{
	bnxef_port_t *lportp;
	bnxef_port_t *rportp;
	bnxef_driver_state_t *dsp;
	ddi_dma_attr_t *xferqdmaattr  = &bnxef_dma_attr;
	ddi_device_acc_attr_t *xferqaccattr = &bnxef_dev_acc_attr;
	int stage = 0, ret;
	    uint32_t req_size;

	ASSERT(sessionp != NULL);
	lportp = sessionp->lport;
	rportp = sessionp->rport;
	if (lportp == NULL)
		return (BNXEF_FAILURE);

	dsp = lportp->statep;
	if ((dsp == NULL) || (rportp == NULL))
		return (BNXEF_FAILURE);

	ASSERT(dsp != NULL);
	ASSERT(lportp != NULL);
	ASSERT(rportp != NULL);


	BNXEF_DEBUG(dsp, GENERAL, lportp->id,
	"%s%d: XFERQ resources Allocated for Session between 0x%x-->0x%x",
		BNXEF_DRIVER_NAME, dsp->instance, lportp->id, rportp->id);
	ret = ddi_dma_alloc_handle(dsp->dip, xferqdmaattr, DDI_DMA_SLEEP, NULL,
		&sessionp->xferq_dma_handle);
	if (ret != DDI_SUCCESS) {
		BNXEF_DEBUG(dsp, ERROR, lportp->id,
			"%s%d: Unable to allocate XFERQ dma handle\n",
			BNXEF_DRIVER_NAME, dsp->instance);
		goto bnxef_xferq_res_alloc_error;
	}

	stage |= BNXEF_XFERQ_RES_DMA_HANDLE_ALLOCATED;
	BNXEF_DEBUG(dsp, GENERAL, lportp->id,
	"%s%d: XFERQ DMA handle allocated for session between 0x%x-->0x%x",
		BNXEF_DRIVER_NAME, dsp->instance, lportp->id, rportp->id);

	req_size = (sessionp->max_sqes + 8) * sizeof (struct fcoe_xfrqe);
	req_size = (req_size + (BNXEF_PAGE_SIZE -1))  & BNXEF_PAGE_ROUND_MASK;
	ret = ddi_dma_mem_alloc(sessionp->xferq_dma_handle,
			req_size,
			xferqaccattr, DDI_DMA_CONSISTENT, DDI_DMA_SLEEP, NULL,
			&sessionp->xferq_addr,
			(unsigned long *)&sessionp->xferq_length,
			&sessionp->xferq_acc_handle);
	if (ret != DDI_SUCCESS) {
		BNXEF_DEBUG(dsp, ERROR, lportp->id,
			"%s%d: Unable to allocate dma memory\n",
				BNXEF_DRIVER_NAME, dsp->instance);
		goto bnxef_xferq_res_alloc_error;
	}

	stage |=  BNXEF_XFERQ_DMA_MEM_ALLOCATED;

	BNXEF_DEBUG(dsp, GENERAL, lportp->id,
	"%s%d: XFERQ DMA memory allocated for session between 0x%x-->0x%x",
		BNXEF_DRIVER_NAME, dsp->instance, lportp->id, rportp->id);
	memset(sessionp->xferq_addr, 0, sessionp->xferq_length);

	/* Now map the buffer for use with firmware */
	ret = ddi_dma_addr_bind_handle(sessionp->xferq_dma_handle,
		NULL, sessionp->xferq_addr, sessionp->xferq_length,
		DDI_DMA_CONSISTENT | DDI_DMA_RDWR, DDI_DMA_SLEEP, NULL,
		&sessionp->xferq_cookie, &sessionp->xferq_cookie_cnt);
	if (ret != DDI_DMA_MAPPED) {
		BNXEF_DEBUG(dsp, ERROR, lportp->id,
				"%s%d: Unable to bind dma mem with handle\n",
			BNXEF_DRIVER_NAME, dsp->instance);
		goto bnxef_xferq_res_alloc_error;
	}
	/* We have indicated no scatter gather allowed */
	ASSERT(sessionp->xferq_cookie_cnt == 1);
	ddi_dma_sync(sessionp->xferq_dma_handle, 0,
		sessionp->xferq_length, DDI_DMA_SYNC_FORDEV);
	BNXEF_DEBUG(dsp, GENERAL, lportp->id,
		"%s%d: XFERQ DMA memory bound for session between 0x%x-->0x%x",
		BNXEF_DRIVER_NAME, dsp->instance, lportp->id, rportp->id);
	stage |= BNXEF_XFERQ_DMA_MEM_BOUND;

	return (BNXEF_SUCCESS);

bnxef_xferq_res_alloc_error:
	(void) bnxef_xferq_res_free(sessionp, stage);
	return (BNXEF_FAILURE);
}

/*
 * This will free the XFERQ resources
 */
int
bnxef_xferq_res_free(bnxef_session_t *sessionp, uint32_t stage)
{
	int ret = BNXEF_FAILURE;

	ASSERT(sessionp != NULL);

	if ((stage & BNXEF_XFERQ_DMA_MEM_BOUND) == BNXEF_XFERQ_DMA_MEM_BOUND) {
		ret = ddi_dma_unbind_handle(sessionp->xferq_dma_handle);
	}

	if ((stage & BNXEF_XFERQ_DMA_MEM_ALLOCATED)
			== BNXEF_XFERQ_DMA_MEM_ALLOCATED) {
		ddi_dma_mem_free(&sessionp->xferq_acc_handle);
	}

	if ((stage & BNXEF_XFERQ_RES_DMA_HANDLE_ALLOCATED) ==
			BNXEF_XFERQ_RES_DMA_HANDLE_ALLOCATED) {
		(void) ddi_dma_free_handle(&sessionp->xferq_dma_handle);
	}

	return (ret);
}


/*
 * This will allocate the resources needed for LCQ
 */
int
bnxef_session_lcq_res_alloc(bnxef_session_t *sessionp)
{
	bnxef_port_t *lportp;
	bnxef_port_t *rportp;
	bnxef_driver_state_t *dsp;
	ddi_dma_attr_t *lcqdmaattr  = &bnxef_dma_attr;
	ddi_device_acc_attr_t *lcqaccattr = &bnxef_dev_acc_attr;
	int stage = 0, ret;
	uint32_t req_size;

	ASSERT(sessionp != NULL);
	lportp = sessionp->lport;
	rportp = sessionp->rport;
	if (lportp == NULL)
		return (BNXEF_FAILURE);

	dsp = lportp->statep;
	if ((dsp == NULL) || (rportp == NULL))
		return (BNXEF_FAILURE);

	ASSERT(dsp != NULL);
	ASSERT(lportp != NULL);
	ASSERT(rportp != NULL);


	BNXEF_DEBUG(dsp, GENERAL, lportp->id,
	"%s%d: LCQ resources Allocated for Session between 0x%x-->0x%x",
		BNXEF_DRIVER_NAME, dsp->instance, lportp->id, rportp->id);
	ret = ddi_dma_alloc_handle(dsp->dip, lcqdmaattr, DDI_DMA_SLEEP, NULL,
		&sessionp->lcq_dma_handle);
	if (ret != DDI_SUCCESS) {
		BNXEF_DEBUG(dsp, ERROR, lportp->id,
			"%s%d: Unable to allocate LCQ dma handle\n",
			BNXEF_DRIVER_NAME, dsp->instance);
		goto bnxef_lcq_res_alloc_error;
	}

	stage |= BNXEF_LCQ_RES_DMA_HANDLE_ALLOCATED;

	BNXEF_DEBUG(dsp, GENERAL, lportp->id,
	"%s%d: LCQ DMA handle allocated for session between 0x%x-->0x%x",
		BNXEF_DRIVER_NAME, dsp->instance, lportp->id, rportp->id);
	req_size = (sessionp->max_sqes + 8) * sizeof (struct fcoe_lcqe);
	req_size = (req_size + (BNXEF_PAGE_SIZE -1)) & BNXEF_PAGE_ROUND_MASK;
	ret = ddi_dma_mem_alloc(sessionp->lcq_dma_handle,
			req_size,
			lcqaccattr, DDI_DMA_CONSISTENT, DDI_DMA_SLEEP, NULL,
			&sessionp->lcq_addr,
			(unsigned long *)&sessionp->lcq_length,
			&sessionp->lcq_acc_handle);
	if (ret != DDI_SUCCESS) {
		BNXEF_DEBUG(dsp, ERROR, lportp->id,
			"%s%d: Unable to allocate dma memory\n",
				BNXEF_DRIVER_NAME, dsp->instance);
		goto bnxef_lcq_res_alloc_error;
	}

	stage |=  BNXEF_LCQ_DMA_MEM_ALLOCATED;

	memset(sessionp->lcq_addr, 0, sessionp->lcq_length);
	BNXEF_DEBUG(dsp, GENERAL, lportp->id,
	"%s%d: LCQ DMA memory allocated for session between 0x%x-->0x%x",
		BNXEF_DRIVER_NAME, dsp->instance, lportp->id, rportp->id);

	/* Now map the buffer for use with firmware */
	ret = ddi_dma_addr_bind_handle(sessionp->lcq_dma_handle,
		NULL, sessionp->lcq_addr, sessionp->lcq_length,
		DDI_DMA_CONSISTENT | DDI_DMA_RDWR, DDI_DMA_SLEEP, NULL,
		&sessionp->lcq_cookie, &sessionp->lcq_cookie_cnt);
	if (ret != DDI_DMA_MAPPED) {
		BNXEF_DEBUG(dsp, ERROR, lportp->id,
		"%s%d: Unable to bind dma mem with handle\n",
		BNXEF_DRIVER_NAME, dsp->instance);
		goto bnxef_lcq_res_alloc_error;
	}
	/* We have indicated no scatter gather allowed */
	ASSERT(sessionp->lcq_cookie_cnt == 1);
	ddi_dma_sync(sessionp->lcq_dma_handle, 0, sessionp->lcq_length,
		DDI_DMA_SYNC_FORDEV);
	stage |= BNXEF_LCQ_DMA_MEM_BOUND;
	BNXEF_DEBUG(dsp, GENERAL, lportp->id,
	"%s%d: LCQ DMA memory bound for session between 0x%x-->0x%x",
	BNXEF_DRIVER_NAME, dsp->instance, lportp->id, rportp->id);

	return (BNXEF_SUCCESS);

bnxef_lcq_res_alloc_error:
	(void) bnxef_session_lcq_res_free(sessionp, stage);
	return (BNXEF_FAILURE);
}

/*
 * This will free the LCQ resources
 */
int
bnxef_session_lcq_res_free(bnxef_session_t *sessionp, uint32_t stage)
{
	int ret = BNXEF_FAILURE;

	ASSERT(sessionp != NULL);

	if ((stage & BNXEF_LCQ_DMA_MEM_BOUND) == BNXEF_LCQ_DMA_MEM_BOUND) {
		ret = ddi_dma_unbind_handle(sessionp->lcq_dma_handle);
	}

	if ((stage & BNXEF_LCQ_DMA_MEM_ALLOCATED)
			== BNXEF_LCQ_DMA_MEM_ALLOCATED) {
		ddi_dma_mem_free(&sessionp->lcq_acc_handle);
	}

	if ((stage & BNXEF_LCQ_RES_DMA_HANDLE_ALLOCATED) ==
			BNXEF_LCQ_RES_DMA_HANDLE_ALLOCATED) {
		(void) ddi_dma_free_handle(&sessionp->lcq_dma_handle);
	}

	return (ret);
}


/*
 * This will allocate the resources needed for connection DB
 */
int
bnxef_session_conn_db_res_alloc(bnxef_session_t *sessionp)
{
	bnxef_port_t *lportp;
	bnxef_port_t *rportp;
	bnxef_driver_state_t *dsp;
#if 0
	ddi_dma_attr_t conndbdmaattrd  = {
		DMA_ATTR_V0,			 /* DMA Attribute version */
		(uint64_t)BNXEF_DMA_ATTR_ADDR_LO, /* dma_attr_addr_lo */
		(uint64_t)BNXEF_DMA_ATTR_ADDR_HI,  /* dma_attr_addr_hi */
		(uint64_t)BNXEF_DMA_ATTR_COUNT_MAX, /* dma_attr_count_max */
		8,	 /* dma_attr_align */
		1, /* dma_attr_burstsizes */
		8,  /* dma_attr_minxfer */
		(uint64_t)16, /* dma_attr_maxxfer */
		(uint64_t)BNXEF_DMA_ATTR_BOUNDARY_CROSS, /* dma_attr_seg */
		BNXEF_DMA_ATTR_1_SEG_COUNT, /* dma_attr_segllen - Revisit */
		1,	 /* dma_attr_granular */
	BNXEF_DMA_ATTR_DMA_FLAGS		 /* dma_attr_flags */
	};
#endif
	ddi_dma_attr_t *conndbdmaattr = &bnxef_dma_attr;
	ddi_device_acc_attr_t *conndbaccattr = &bnxef_dev_acc_attr;
	int stage = 0, ret;

	ASSERT(sessionp != NULL);
	lportp = sessionp->lport;
	rportp = sessionp->rport;
	if (lportp == NULL)
		return (BNXEF_FAILURE);

	dsp = lportp->statep;
	if ((dsp == NULL) || (rportp == NULL))
		return (BNXEF_FAILURE);

	ASSERT(dsp != NULL);
	ASSERT(lportp != NULL);
	ASSERT(rportp != NULL);

	BNXEF_DEBUG(dsp, GENERAL, -1,
		"%s%d: Connection DB resources Allocated",
		BNXEF_DRIVER_NAME, dsp->instance);

	ret = ddi_dma_alloc_handle(dsp->dip, conndbdmaattr, DDI_DMA_SLEEP, NULL,
		&sessionp->conn_db_dma_handle);
	if (ret != DDI_SUCCESS) {
		BNXEF_DEBUG(dsp, ERROR, lportp->id,
			"%s%d: Unable to allocate LCQ dma handle\n",
			BNXEF_DRIVER_NAME, dsp->instance);
		goto bnxef_conndb_res_alloc_error;
	}

	stage |= BNXEF_CONN_DB_RES_DMA_HANDLE_ALLOCATED;
	BNXEF_DEBUG(dsp, GENERAL, -1,
		"%s%d: Connection DB DMA handle allocated",
		BNXEF_DRIVER_NAME, dsp->instance);
	ret = ddi_dma_mem_alloc(sessionp->conn_db_dma_handle,
			sizeof (struct fcoe_conn_db),
			conndbaccattr, DDI_DMA_CONSISTENT, DDI_DMA_SLEEP, NULL,
			&sessionp->conn_db_addr,
			(size_t *)&sessionp->conn_db_length,
			&sessionp->conn_db_acc_handle);
	if (ret != DDI_SUCCESS) {
		BNXEF_DEBUG(dsp, ERROR, lportp->id,
			"%s%d: Unable to allocate dma memory\n",
				BNXEF_DRIVER_NAME, dsp->instance);
		goto bnxef_conndb_res_alloc_error;
	}

	stage |=  BNXEF_CONN_DB_DMA_MEM_ALLOCATED;
	BNXEF_DEBUG(dsp, GENERAL, -1,
		"%s%d: Conn DB DMA memory allocated 0x%lx ",
		BNXEF_DRIVER_NAME, dsp->instance, sessionp->conn_db_length);
	memset(sessionp->conn_db_addr, 0, sessionp->conn_db_length);

	/* Now map the buffer for use with firmware */
	ret = ddi_dma_addr_bind_handle(sessionp->conn_db_dma_handle,
		NULL, sessionp->conn_db_addr, sessionp->conn_db_length,
		DDI_DMA_CONSISTENT | DDI_DMA_RDWR, DDI_DMA_SLEEP, NULL,
		&sessionp->conn_db_cookie, &sessionp->conn_db_cookie_cnt);
	if (ret != DDI_DMA_MAPPED) {
		BNXEF_DEBUG(dsp, ERROR, lportp->id,
			"%s%d: Unable to bind dma mem with handle 0x%x",
			BNXEF_DRIVER_NAME, dsp->instance, ret);
		goto bnxef_conndb_res_alloc_error;
	}
	/* We have indicated no scatter gather allowed */
	ASSERT(sessionp->conn_db_cookie_cnt == 1);
	stage |= BNXEF_CONN_DB_DMA_MEM_BOUND;
	BNXEF_DEBUG(dsp, GENERAL, -1,
		"%s%d: Connection DB DMA memory Bound",
		BNXEF_DRIVER_NAME, dsp->instance);

	/* LINTED: alignment */
	((struct fcoe_conn_db *)(sessionp->conn_db_addr))->rq_prod =
			CPU_TO_LE16(0x8000);
	ddi_dma_sync(sessionp->conn_db_dma_handle, 0,
		sessionp->conn_db_length, DDI_DMA_SYNC_FORDEV);

	return (BNXEF_SUCCESS);

bnxef_conndb_res_alloc_error:
	(void) bnxef_session_conn_db_res_free(sessionp, stage);
	return (BNXEF_FAILURE);
}

/*
 * This will free the XFERQ resources
 */
int
bnxef_session_conn_db_res_free(bnxef_session_t *sessionp, uint32_t stage)
{
	int ret = BNXEF_FAILURE;

	ASSERT(sessionp != NULL);

	if ((stage & BNXEF_CONN_DB_DMA_MEM_BOUND)
		== BNXEF_CONN_DB_DMA_MEM_BOUND) {
		ret = ddi_dma_unbind_handle(sessionp->conn_db_dma_handle);
	}

	if ((stage & BNXEF_CONN_DB_DMA_MEM_ALLOCATED)
			== BNXEF_CONN_DB_DMA_MEM_ALLOCATED) {
		ddi_dma_mem_free(&sessionp->conn_db_acc_handle);
	}

	if ((stage & BNXEF_CONN_DB_RES_DMA_HANDLE_ALLOCATED) ==
			BNXEF_CONN_DB_RES_DMA_HANDLE_ALLOCATED) {
		(void) ddi_dma_free_handle(&sessionp->conn_db_dma_handle);
	}

	return (ret);
}


/*
 * This connection Id generation method is used only for offloaded connection.
 * not used for other connections.
 */
int
bnxef_alloc_session_con_id(bnxef_driver_state_t *dsp, bnxef_session_t *sessionp)
{
	int i;

	ASSERT(dsp != NULL);

	MUTEX_ENTER(dsp, &dsp->state_mutex, __func__, __LINE__);
	for (i = 0; i < BNXEF_MAX_SESSIONS_SUPPORTED; i++) {
		if (dsp->sessp[i] == NULL) {
			dsp->sessp[i] = sessionp;
			dsp->num_sessions += 1;
			break;
		}
	}

	mutex_exit(&dsp->state_mutex);
	if (i >= BNXEF_MAX_SESSIONS_SUPPORTED)
		return (-1);

	return (i);
}


void
bnxef_free_session_con_id(bnxef_driver_state_t *dsp, int con_id)
{

	ASSERT(dsp != NULL);

	MUTEX_ENTER(dsp, &dsp->state_mutex, __func__, __LINE__);
	dsp->sessp[con_id] = NULL;
	ASSERT(dsp->num_sessions > 0);
	dsp->num_sessions -= 1;
	mutex_exit(&dsp->state_mutex);
}
