/*
 * The information contained in this file is confidential and proprietary to
 * Broadcom Corporation.  No part of this file may be reproduced or
 * distributed, in any form or by any means for any purpose, without the
 * express written permission of Broadcom Corporation.
 *
 * (c) COPYRIGHT 2010-2012 Broadcom Corporation, ALL RIGHTS RESERVED.
 */

#pragma ident "(C) 2010-2012 Broadcom Corporation"
#pragma ident "$File$$Revision$$Change$$DateTime$"

#ifndef __BNXEF_STATE_H__
#define	__BNXEF_STATE_H__

#include <sys/types.h>
#include <sys/conf.h>
#include <sys/kmem.h>
#include <sys/stream.h>
#include <sys/stropts.h>
#include <sys/errno.h>
#include <sys/cred.h>
#include <sys/poll.h>
#include <sys/modctl.h>
#include <sys/stat.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>
#include <sys/ddifm.h>
#include <sys/fm/protocol.h>
#include <sys/fm/util.h>
#include <sys/fm/io/ddi.h>
#include <sys/pattr.h>
#include <sys/sysmacros.h>
#include <sys/strsun.h>
#include <sys/ksynch.h>
#include <sys/kstat.h>
#include <sys/strsubr.h>
#include <sys/ksynch.h>
#include <sys/ethernet.h>
#include <bnxef_trace_events.h>

#include "version.h"

#if !(defined(__S11) || defined(__S12))
#define	 ETHERTYPE_FCOE	(0x8906)
#define	htonll(a)   (htonl(a >> 32) | ((uint64_t)htonl(a & 0xFFFFFFFF) << 32))
#endif

#ifndef STRINGIFY
#define XSTRINGIFY(x) #x
#define STRINGIFY(x) XSTRINGIFY(x)
#endif

/*
 * The Sucess or FAILURE MACROS
 */
#define	BNXEF_SUCCESS	0
#define	BNXEF_FAILURE 	-1
#define	BNXEF_DRIVER_MAGIC	0xCAFEF00D
#define	BNXEF_FCOE_VERSION	STRINGIFY(MAJVERSION) "." STRINGIFY(MINVERSION) "." STRINGIFY(REVVERSION)

#define	BNXEF_DRIVER_NAME	"bnxef"
#define	BNXEF_PRODUCT_BANNER "Broadcom NetXtreme II 10 Gigabit FCoE Driver v" BNXEF_FCOE_VERSION
#define	BNXEF_PRODUCT_INFO   "Broadcom NXII 10 GbE FCoE v" BNXEF_FCOE_VERSION

/*
 * THe following are firmware related constants used in some calculations
 */
#define	BNXEF_MAX_INSTANCES		40
#define	BNXEF_MAX_SESSIONS_SUPPORTED	2048
#define	BNXEF_MAX_FC_PORTS		255
#define	BNXEF_MAX_EXCHGS_SUPPORTED	(4*1024)
#define	BNXEF_NUM_TASKS_ALLOWED_FOR_LV_IOS  (BNXEF_MAX_EXCHGS_SUPPORTED/2)
#define	BNXEF_FPMA_SUPPORTED		1
#define	BNXEF_MAX_ROWS_IN_HASH_TBL	8
#define	BNXEF_HASH_TBL_CHUNK_SIZE	(16*1024)
#define	BNXEF_SQ_WQES_MAX		4096
#define	BNXEF_RQ_WQES_MAX		128
#define	BNXEF_CQ_WQES_MAX		(BNXEF_SQ_WQES_MAX + BNXEF_RQ_WQES_MAX)
#define	BNXEF_RQ_BUF_SIZE		256
#define	BNXEF_RQ_BUF_LOG_SZ	    8
#define	BNXEF_NUM_MAX_SESS_LOG		11
#define	BNXEF_TASK_HANDLES		256


/*
 * Kmem cache alloc stages
 */
#define	 BNXEF_SESSION_CACHE_CREATED	0x01
#define	 BNXEF_PORT_CACHE_CREATED	0x02
#define	 BNXEF_TQ_PARAM_CACHE_CREATED	0x04
#define	 BNXEF_FCF_CACHE_CREATED	0x08
#define	 BNXEF_TASK_CACHE_CREATED	0x10
#define	 BNXEF_UNSOL_BUF_CACHE_CREATED	0x20
#define	BNXEF_CACHE_ALL			0x3F

/*
 * init list stages
 */
#define	 BNXEF_FCF_LIST_INIT		0x01
#define	 BNXEF_TASK_FREE_LIST_INIT	0x02
#define	 BNXEF_PORT_LIST_INIT		0x04
#define	 BNXEF_LIST_INIT_ALL		0x07
/*
 * TAsk iresource allocation stages
 */
#define	 BNXEF_INDIV_TASK_DMA_HANDLE_ALLOCATED	0x01
#define	 BNXEF_INDIV_TASK_DMA_MEM_ALLOCATED		0x02
#define	 BNXEF_INDIV_TASK_DMA_MEM_BOUND		0x04
#define	 BNXEF_INDIV_TASK_RESOURCE_ALL			0x07
#define	 BNXEF_TASK_CTX_BD_HANDLE_ALLOCATED	0x08
#define	 BNXEF_TASK_CTX_BD_MEM_ALLOCATED	0x10
#define	 BNXEF_TASK_CTX_BD_MEM_BOUND		0x20
#define	 BNXEF_TASK_BD_PBL_SETUP		0x40
#define	 BNXEF_TASK_RESOURCE_ALL		0x7F

/*
 * Hash table resource allocation stages
 */
#define	BNXEF_HASH_TABLE_PBL_SETUP		0x01
#define	BNXEF_HASH_TABLE_PBL_MEM_BOUND		0x02
#define	 BNXEF_HASH_TABLE_PBL_MEM_ALLOCATED	0x04
#define	 BNXEF_HASH_TABLE_PBL_HANDLE_ALLOCATED	0x08
#define	 BNXEF_HASH_TABLE_BOUND			0x10
#define	 BNXEF_HASH_TABLE_DMA_MEM_ALLOCATED	0x20
#define	 BNXEF_HASH_TABLE_DMA_HANDLE_ALLOCATED	0x40
#define	 BNXEF_HASH_TABLE_ALL			0x7F

/*
 * T2 Hash table resource allocation stages
 */
#define	BNXEF_T2_HASH_TABLE_PTR_BOUND			0x01
#define	BNXEF_T2_HASH_TABLE_PTR_MEM_ALLOCATED		0x02
#define	BNXEF_T2_HASH_TABLE_PTR_HANDLE_ALLOCATED	0x04
#define	BNXEF_T2_HASH_TABLE_BOUND			0x08
#define	BNXEF_T2_HASH_TABLE_MEM_ALLOCATED		0x10
#define	BNXEF_T2_HASH_TABLE_HANDLE_ALLOCATED		0x20
#define	BNXEF_T2_HASH_TABLE_ALL				0x3F


/*
 * Dummy Buffer resource allocation stages
 */
#define	 BNXEF_DUMMY_BUF_BOUND				0x01
#define	 BNXEF_DUMMY_BUF_MEM_ALLOCATED			0x02
#define	 BNXEF_DUMMY_BUF_HANDLE_ALLOCATED		0x04
#define	 BNXEF_DUMMY_BUF_ALL				0x07

/*
 * Stats Buffer Resource allocation stages
 */
#define	BNXEF_STATS_BUF_BOUND				0x01
#define	BNXEF_STATS_BUF_MEM_ALLOCATED			0x02
#define	BNXEF_STATS_BUF_HANDLE_ALLOCATED		0x04
#define	BNXEF_STATS_BUF_ALL				0x07

/*
 * Firmware resource allocation stages
 */
#define	BNXEF_FW_DUMMY_BUFFER_COMPLETE			0x01
#define	BNXEF_FW_T2_HASH_TABLE_COMPLETE			0x02
#define	BNXEF_FW_HASH_TABLE_COMPLETE			0x04
#define	BNXEF_FW_TASK_RES_ALLOC_COMPLETE		0x08
#define	BNXEF_FW_STATS_BUFFER_COMPLETE			0x10
#define	BNXEF_FW_RES_ALL				0x1F

/*
 * Attach stages.
 */
#define	 BNXEF_FCOE_FW_STARTED				0x01
#define	 BNXEF_L2_REG_COMPLETE				0x02
#define	 BNXEF_LEADVILLE_REG_COMPLETE			0x04
#define	 BNXEF_DOOR_BELL_BAR_MAPPED			0x08
#define	 BNXEF_TASK_QUEUE_CREATED			0x10
#define	 BNXEF_KSTAT_RES_ALLOCATED			0x20
#define	 BNXEF_DRIVER_TASK_RES_ALLOCATED		0x40
#define	 BNXEF_MISC_MEM_ALLOCATED			0x80
#define	 BNXEF_FIRMWARE_RES_ALLOCATED			0x100
#define	 BNXEF_LISTS_INITIALIZED			0x200
#define	 BNXEF_CACHE_ALLOC_COMPLETED			0x400
#define	 BNXEF_L2_PARAMS_RECEIVED			0x800
#define	 BNXEF_TUNABLES_VALIDATED			0x1000
#define	 BNXEF_TRIGGER_FP_ATTACH			0x2000
#define	 BNXEF_MISC_TASK_QUEUE_CREATED			0x4000
#define	 BNXEF_ATTACH_ALL				0x7fff

#define	BNXEF_USE_OX_ID					0x01
#define	BNXEF_USE_RX_ID					0x02

/*
 * These are LS RJT Reason code. Defines in Leadville is convuluted to include.
 * Because of conflicts. Therefore only LS_RJT reason codes are defined.
 */
#define	BNXEF_LA_RJT_INVALID		0x01
#define	BNXEF_LA_RJT_LOGICAL_ERR	0x03
#define	BNXEF_LA_RJT_LOGICAL_BUSY	0x05
#define	BNXEF_LS_RJT_PROTOCOL_ERR	0x07
#define	BNXEF_LA_RJT_UNABLE_TO_PERFORM  0x09
#define	BNXEF_LA_RJT_NOT_SUPPORTED	0x0b
#define	BNXEF_LA_RJT_VENDOR		0xff

/* DMA attributes related defines */
#define	BNXEF_DMA_ATTR_ADDR_LO		0x0
#define	BNXEF_DMA_ATTR_ADDR_HI		0xffffffffffffffffull
#define	BNXEF_DMA_ATTR_COUNT_MAX	0xffffffff
#define	BNXEF_DMA_ATTR_BYTE_ALIGN	0x01
#ifdef LITTLE_ENDIAN
#define	BNXEF_DMA_ATTR_ALIGN		4096
#else
#define	BNXEF_DMA_ATTR_ALIGN		8192
#endif
#define	BNXEF_DMA_ATTR_4K_ALIGN		4096
#define	BNXEF_DMA_ATTR_8K_ALIGN		8192
#define	BNXEF_DMA_ATTR_16K_ALIGN	(16 *1024)
#define	BNXEF_DMA_ATTR_MIN_XFER		1
#define	BNXEF_DMA_ATTR_MAX_XFER		0xFFFFFFFF
#define	BNXEF_DMA_ATTR_BOUNDARY_CROSS	0xFFFFFFFFFFFFFFFFull
#define	BNXEF_DMA_ATTR_1_SEG_COUNT	1
#define	BNXEF_DMA_ATTR_16_SEG_COUNT	16
#define	BNXEF_DMA_ATTR_128_GRANULARITY	1
#define	BNXEF_DMA_ATTR_8_GRANULARITY	8
#define	BNXEF_DMA_ATTR_DMA_FLAGS	0
#define	BNXEF_DMA_ATTR_BURST_SIZE	0xff

/* BD Table Length Related defined */
#define	BNXEF_MAX_BD_SIZE		0xffff
#define	BNXEF_SPLIT_BD_SIZE		0x8000

/* Active I/Os per session allowed */
#define	BNXEF_MAX_ACTIVE_IOS_PER_SESSION  (BNXEF_SQ_WQES_MAX/2)

#define	BNXEF_STATS(sp, x)  \
	if (sp->stats) { \
		bnxef_driver_stats_t *statsp; \
		statsp = (bnxef_driver_stats_t *)sp->stats->ks_data; \
		statsp->x.value.ui64++; \
	}

#define	BNXEF_STATS_ADD(sp, y, x, val)  {\
		if (LE32_TO_CPU(sp->prevstats.y.x) <= val) {\
			val -= LE32_TO_CPU(sp->prevstats.y.x); \
		} else {\
			val = val + (0xffffffff \
			    - LE32_TO_CPU(sp->prevstats.y.x)); \
		}\
		\
		if (sp->fwstats) {\
			bnxef_fw_stats_t *statsp; \
			statsp = (bnxef_fw_stats_t *)sp->fwstats->ks_data; \
			statsp->x.value.ui64 += val; \
		}\
	}

#define	BNXEF_STATS_SET(sp, x, val) \
	if (sp->stats) { \
		bnxef_driver_stats_t *statsp; \
		statsp = (bnxef_driver_stats_t *)sp->stats->ks_data; \
		statsp->x.value.ui32 = val; \
	}

/*
 * The following is used for statistics maintenance using kstat infra structure
 */
typedef struct __bnxef_driver_stats__ {
	kstat_named_t	init_cqe_tx;
	kstat_named_t	init_comp_cqe_rx;
	kstat_named_t	ofld_cqe_tx;
	kstat_named_t	ofld_comp_cqe_rx;
	kstat_named_t	enable_cqe_tx;
	kstat_named_t	enable_comp_cqe_rx;
	kstat_named_t	fp_completion_tx;
	kstat_named_t	fp_completion_rx;
	kstat_named_t	successfull_scsi_cmds; /* successfull completion */
	kstat_named_t	errored_scsi_cmds; /* Errored scsi commands */
	kstat_named_t	driver_version;	   /* drver version */
	kstat_named_t	firmware_version;  /* firmware version used */
	kstat_named_t	l2_rx;		   /* No. of L2 msg bufs rxed */
	kstat_named_t	l2_tx;		   /* No. of L2 msg bufs txed */
	kstat_named_t	l2_rx_free;	   /* No. of L2 msg buf freed */
	kstat_named_t	l2_fip_rx;	   /* No. of FIP l2 msg buf rxed */
	kstat_named_t	l2_fip_tx;	   /* No. of FIP l2 msg buf txed */
	kstat_named_t	l2_fip_rx_free;	   /* No. of FIP L2 rx msg buf freed */
	kstat_named_t	l2_fcoe_rx;	   /* No. of FCOE L2 msg buf rxed */
	kstat_named_t	l2_fcoe_tx;	   /* No. of FCOE l2 msg buf txed */
	kstat_named_t	l2_fcoe_rx_free;   /* No. of FCOE l2 rx msg buf freed */
	kstat_named_t	l2_unknown_rx;	   /* Unknown to this driver */
	kstat_named_t	l2_unknown_rx_free; /* Unknown Frame rx freed count */
	kstat_named_t	els_dispatched;	   /* ELS tasks dispatched */
	kstat_named_t	ct_dispatched;	   /* FC-CT tasks dispatched */
	kstat_named_t	scsi_dispatched;   /* SCSI tasks dispatched */
	kstat_named_t	rscn_rx;	   /* Count of RSCN Received */
	kstat_named_t	rscn_acc_tx;	   /* Number of RSCN Local Acc sent */
	kstat_named_t	rscn_acc_lv_tx;	   /* Number of RSCN LV acc Dropped */
	kstat_named_t	timeout_tasks;	   /* Timedout tasks */
	kstat_named_t	state_hi;	   /* State ptr hi */
	kstat_named_t	state_lo;	   /* state of this instance */
	kstat_named_t	max_exchanges;	   /* MAx exchanged allowed */
	kstat_named_t	max_connections;   /* Max. ofld connection allowed */
	kstat_named_t	scsi_bsy_ret;	   /* No. of times TRAN bsy returned */
} bnxef_driver_stats_t;

typedef struct __bnxef_fw_stats__ {
	kstat_named_t	fcoe_tx_pkt_cnt; /* Cumulative fcoe tx pkts */
	kstat_named_t	fcoe_tx_byte_cnt; /* Cumulative fcoe byte count */
	kstat_named_t	fcp_tx_pkt_cnt;	   /* cumulative fcp tx pkt count */

	kstat_named_t	fcp_rx_pkt_cnt;	/* Cumulative fcp rx pkt count */
	kstat_named_t	 fcoe_rx_drop_pkt_cnt; /* Cumulative rx drop count */
	kstat_named_t	drop_seq_cnt;	/* Cumulative dropped sequence count */
	kstat_named_t	seq_timeout_cnt; /* Cumulative Seq timeout count */
	kstat_named_t	miss_frame_cnt;	/* Cumulative missing frame count */
	kstat_named_t	eofa_del_cnt;	/* Cumulative EoFa received count */
	kstat_named_t	fc_crc_cnt;	/* Cumulative CRC error count */

	kstat_named_t	fcoe_ver_cnt;	/* Cumulative version count */

	kstat_named_t	fcoe_rx_pkt_cnt; /* Cumulative rx pkt cnt */
	kstat_named_t	fcoe_rx_byte_cnt; /* Cumulative rx byte cnt */
} bnxef_fw_stats_t;

#if 0
/*
 * This will carry all debugging information only, will be removed for final
 * driver. It will go under #ifdef DEBUG or #ifdef TESTING.
 */
typedef struct __bnxef_global_debug__ {
	uint32_t	task_freed;
	uint32_t	task_allocated;
	struct ether_header *gep;
	struct ether_vlan_extinfo *gvp;
	void *gfipp;
	void *gophp;
	void *gfp;
	void *gnamep;
	void	*gfcoesizep;
	void *gfipfcp;
	void *gframep;
	void  *gloginp;
	void *glogop;
	void *portp;
	void *sessionp;
	uint8_t hashbucket;
	void *elsp;
	uint32_t  d_id;
	void *pktp;
	void *mp;
	uint32_t crc;
	uint32_t rxcrc;
	uint32_t size;
	void *taskp;
	void *unsolbp;
	void *ubp;
	void *tqp;
	struct fcoe_kwqe_conn_offload1 ofld_req1[2];
	struct fcoe_kwqe_conn_offload2 ofld_req2[2];
	struct fcoe_kwqe_conn_offload3 ofld_req3[2];
	struct fcoe_kwqe_conn_offload4 ofld_req4[2];
	struct fcoe_kwqe_conn_enable_disable enable_req[2];
	int offload_count;
	int enable_count;
	uint8_t global_task_array[BNXEF_MAX_EXCHGS_SUPPORTED];
} bnxef_global_debug_t;
#endif

typedef struct __bnxef_driver_state__ {
	uint32_t	magic;		/* magic number to validate */
	/* Solaris, Leadville Specific Information */
	dev_info_t	*dip;		/* Dip for this node */
	dev_info_t	*fp_child_dip;	/* Dip for fp node created */
	int		instance; 	/* This drivers instance */
	int		state;		/* Indicate whether attach completed */
#define	BNXEF_ATTACH_COMPLETED		0x01
#define	BNXEF_IMPLICIT_FABRIC_LOGO_IN_PROGRESS	0x04
#define	BNXEF_PORT_ONLINE_HANDLER_INPROGRESS	0x08
#define	BNXEF_PORT_CVL_PROCESSING_INPROGRESS	0x10
#define	BNXEF_PORT_OFFLINE_HANDLER_INPROGRESS	0x20
#define	BNXEF_DETACH_IN_PROGRESS		0x40
#define	BNXEF_L2_UNLOAD_CALLED			0x80
#define	BNXEF_STAT_PULL_TIMEOUT_VALID		0x100
#define	BNXEF_STAT_REPORT_TIMEOUT_VALID		0x200

	uint8_t		physical_port_state; /* Phy port state */
#define	BNXEF_PHYPORT_ONLINE		0x01
#define	BNXEF_PHYPORT_OFFLINE		0x00

	kcondvar_t	port_offline_cv; /* Wait for port ofln to complete */
	uint32_t	port_bring_up;

	ldi_ident_t	lip;		/* LAyered driver id pointer */
	ldi_handle_t	lh;		/* Layered device Handle */
	fc_fca_tran_t	tran;		/* Transport structure */
	kmutex_t	state_mutex;	/* This will guard all state */
	kmutex_t	global_offload_mutex; /* guard offload destroy */
	uint32_t	capablities;	/* Leadville exported capability */
	la_els_logi_t	logi_params;	/* Default logi Parameters supported */
	kstat_t		*stats;		/* kstat structure allocated */
	kstat_t		*fwstats;	/* Firmware related stats */
	int32_t		poll_ref_cnt;

	/* Task queue Related */
	ddi_taskq_t	*tq;		/* Task Q to handle error/misc items */
	ddi_taskq_t	*mtq;		/* Miscellaneous task Queue */
	kmem_cache_t *tq_params_cache; /* Slab for Task Queue params */

#ifdef DEBUG
	/* Debug Related */
	uint32_t	bnxef_debug;	    /* Debug Variable */
	int		*bnxef_debug_ports; /* Port IDs */
	uint32_t	port_debug_num_ports; /* Number of items in the list */
#endif

	/* Port Structure Cache */
	kmem_cache_t	*port_cache;	/* Cache for port(remote, lport) */

	/* Local Port Information */
	bnxef_port_list_t port_list;	/* List of Lport Instantiated */
	uint32_t	max_fc_ports;	/* max. supportable fc ports */
	uint32_t	num_fc_ports;	/* Number of instantiated ports */
	uint32_t	port_ref_cnt;	/* No. of times Constructor called */

	/* Session Information */
	kmem_cache_t	*session_cache;	/* Session Cache  - 2K Max */
	uint32_t	max_sessions;	/* MAx Session Supported */
	uint32_t	num_sessions;	/* num session established */

	/* You can index based on connection ID */
	bnxef_session_t	**sessp;		/* Array of session Ptrs */
	int		con_id_cnt;	/* Connection ID Count  */


	/* Task Related resources */
	kmem_cache_t	*task_cache;		/* Task Meta data */
	uint32_t	task_ref_cnt;		/* Task Reference count */
	uint16_t	num_task_dma_handles;	/* Number of hdls allocated */
	caddr_t		task_ctx[BNXEF_TASK_HANDLES];
			/* Task Ctx-Dma'able Vaddr */
	size_t		task_ctx_length[BNXEF_TASK_HANDLES];
			/* real length allocated */
	ddi_dma_attr_t	task_dma_attr;		/* TAsk DMA attributes */
	ddi_dma_handle_t task_dma_handle[BNXEF_TASK_HANDLES];
			/* Dma Handles allocated */
	ddi_dma_cookie_t task_dma_cookie[BNXEF_TASK_HANDLES];
			/* DMA cookies allocated */
	uint32_t	task_dma_cookie_cnt[BNXEF_TASK_HANDLES];
				/* Cookie count */
	ddi_acc_handle_t  task_acc_handle[BNXEF_TASK_HANDLES];
			/* device access handle */

	bnxef_task_list_t bnxef_task_free_list;	/* Free list of tasks/X_ID */
	bnxef_task_list_t bnxef_task_alloc_list; /* Allocated list of tasks */
	uint32_t	max_exchanges;		/* MAx exch supported */
	bnxef_fcoe_task_t  **taskarray;		/* Index using X_ID for taskp */
	uint32_t	task_bd_table_pre_alloc;

	/* Unsolicited buffer cache */
	kmem_cache_t	*unsol_buf_cache;	/* Unsolicited buffer cache */
	uint32_t	unsol_alloc_cnt;	/* Unsol buf alloc count */
	uint32_t	unsol_buf_ref_cnt;	/* Unsolicited refence count */
	/* TAsk Buffer descriptor Information */
	char	*task_ctx_bd_table;	/* Vaddr of bd table */
	size_t		task_ctx_bd_table_len;	/* Actual length Allocated */
	ddi_dma_handle_t task_ctx_bd_handle;	/* DMA Handle for this Page */
	ddi_dma_cookie_t	 task_ctx_bd_cookie;	/* DMA cookie  */
	uint32_t	task_ctx_bd_cookie_cnt; /* Number of cookies */
	ddi_acc_handle_t task_ctx_bd_acc_handle; /* Device access handle */

	/* HAsh Tbale Information */
	int		 hash_tbl_seg_count;	/* Segment Count */
	char		*hash_tbl;		/* Vaddr of different Segs */
	size_t		hash_tbl_len;		/* HAsh table len Allocated */
	ddi_dma_handle_t hash_tbl_dma_handle;	/* Hash Table DMA handle */
	ddi_dma_cookie_t hash_tbl_dma_cookie;	/* Hash table DMA cookie */
	uint32_t	hash_tbl_cookie_cnt;	/* cookie count */
	ddi_acc_handle_t hash_tbl_acc_handle;	/* device access handle */

	/* HAsh Table Buffer descriptors */
	char		*hash_table_pbl;	/* Hash table Buffer List */
	size_t		hash_table_pbl_len;	/* Length of PBL allcated */
	ddi_dma_handle_t hash_table_pbl_handle;	/* DMA Handle for HTBL */
	ddi_dma_cookie_t hash_table_pbl_cookie;	/* DMA Cookie for HTBL */
	uint32_t	hash_table_pbl_cookie_cnt;	/* Cookie Count */
	ddi_acc_handle_t hash_table_pbl_acc_handle; /* device access handle */

	/* Tier 2 Hash Table information */
	char *t2_hash_table; /* VAddr of T2 hash */
	size_t		t2_hash_table_len;		/* Actual length */
	ddi_dma_handle_t t2_hash_table_dma_handle;	/* DMa Hdl T2 Hash */
	ddi_dma_cookie_t t2_hash_table_dma_cookie; 	/* DMA Cookie T2 Hash */
	uint32_t	t2_hash_table_cookie_cnt;	/* Cookie Count */
	ddi_acc_handle_t t2_hash_table_acc_handle;	/* Dev access handle */

	/* TBD */
	char	 *t2_hash_tbl_ptr;	/* Pointer to T2 Hash Table */
	size_t	 t2_hash_tbl_ptr_len;	/* Actual length */
	ddi_dma_handle_t t2_hash_table_dma_ptr_handle; /* Hash table Ptr */
	ddi_dma_cookie_t t2_hash_table_dma_ptr_cookie; /* H-table ptr cookie */
	uint32_t	t2_hash_table_ptr_cookie_cnt;	/* Cookie count */
	ddi_acc_handle_t t2_hash_table_ptr_acc_handle;	/* device acc handle */

	/* Dummy Buffer */
	char	 *dummy_buffer;			/* Vaddr of Dummy Buffer */
	size_t	 dummy_size;			/* Actual size */
	ddi_dma_handle_t  dummy_dma_handle;	/* DMA handle for dummy Bufg */
	ddi_dma_cookie_t dummy_dma_cookie;	/* DMA Cookie for Dummy Buf */
	uint32_t	dummy_cookie_cnt;	/* Cookie count */
	ddi_acc_handle_t dummy_dev_acc_handle;	/* Device access handle */

	/* Stats Buffer */
	char	*stats_buffer;			/* The statistics buffer */
	struct fcoe_statistics_params prevstats; /* previously probed value */
	size_t	stats_size;			/* Actual size */
	ddi_dma_handle_t stats_dma_handle;	/* DMA handle for stats buf */
	ddi_dma_cookie_t stats_dma_cookie;	/* Stats Dma cookie */
	uint32_t	stats_cookie_cnt;	/* Number of cookies */
	ddi_acc_handle_t stats_dev_acc_handle;	/* Device access handle */
	uint16_t	session_ref_cnt;
	timeout_id_t	stats_pull_timeout;		/* Timeout pull stats */
	timeout_id_t	stats_report_timeout;	/* Timeout report stats */


	/* Interface Specific Information */
	uint8_t		mac_add[ETHERADDRL]; /* MAc address */
	uint8_t		if_state;		/* Online Offline etc */
	uint16_t	mtu_size;		/* mtu Size for RQ alloc */
	BnxeBinding	callbacks;		/* Callback to L2 driver */
	void *doorbellbar;			/* door bell bar base kaddr */
	ddi_acc_handle_t doorbell_acc;

	/* Firmware Initialization Info */
	kcondvar_t	fw_init_cond;		/* Conf Var for F/w init */
	kmutex_t	fw_init_mutex;		/* fw init Mutex */
	uint8_t		fw_init_state;		/* fw Init state */
#define	BNXEF_INIT_MSG_SENT		0x02
	uint32_t	fw_init_error;		/* Err ret for init/destroy */
	struct fcoe_kwqe_init1  fcoe_init1;
	struct fcoe_kwqe_init2  fcoe_init2;
	struct fcoe_kwqe_init3  fcoe_init3;

	/* Connection Offliad synchronization Variable */
	/* This condition variable will be used with hba mutex */
	kcondvar_t	conn_offload_cond;	/* Cond var for F/w */

	/* FIP  Handling Information */
	kmem_cache_t		*fcf_cache;	/* Slab of fcf structures */
	kcondvar_t		fcf_cv;		/* indicate selection of FCF */
	bnxef_fcf_list_t	fcf_list;	/* THis identifies the list */
	bnxef_fcf_info_t	*fcf_ident;	/* THis identifies the FCF */
	uint32_t		fcf_ref_cnt;	/* Reference count */
	uint16_t	 vlan_id;		/* Vlan Id used for FIP */
	uint8_t		 pri;			/* priority used in Vlan Tag */
	uint16_t	  max_fcoe_size;		/* Set to 0x86E size */

	uint32_t		fip_disc_state;	/* Discovery State */
#define	FIP_VLAN_DISC_STATE		0x01
#define	FIP_DISC_FCF_STATE		0x02
#define	FIP_LOGIN_READY_STATE		0x03
#define	FIP_VLINK_MAINT_STATE	0x04

	/* Indicates what frame has been sent */
	uint32_t		fip_disc_status;
#define	FIP_VLAN_REQUEST_SENT		0x0001
#define	FIP_VLAN_RESPONSE_RECEIVED	0x0002
#define	FIP_DISC_SOL_SENT		0x0004
#define	FIP_UNICAST_DISC_ADV_RECEIVED	0x0008
#define	FIP_MULTICAST_DISC_ADV_RECEIVED 0x0010

#define	FIP_TIMEOUT_ARMED		0x8000
#define	FIP_KEEP_ALIVE_TIMER_ARMED	0x4000

	/* The following is to enforce FIP DISC ADV reception */
	timeout_id_t		fip_keep_alive_tmr_hdl; /* timeout handle */
	timeout_id_t		fip_timeout; /* catch runaway condition */

	/* Support is based on the Property exported in driver.conf */
	uint8_t		flags_fpma_spma;	/* FPMA support orSPMA */
	uint8_t			ioctl_flags;	/* Used for ioctls */
#define	BNXEF_IOCTL_FLAG_OPEN		0x01
	BnxeFcoeInfo		bnxe_params;	/* Params about fcoe port */
	char	filename[128];
	int	 linenum;
	char	g_filename[128];
	int	 g_linenum;
	kcondvar_t stats_cv;
	kcondvar_t fp_up_cv;
	kmutex_t   global_event_log_mutex;
	bnxef_trace_events_t *global_event_logp;
	uint32_t	gev_log_idx;
	uint32_t	bnxef_max_trace_events;
} bnxef_driver_state_t;

typedef struct la_els_rrq {
	uint8_t		ls_code;
	uint8_t		mbz[3];
	uint8_t		reserved;
	uint8_t	 source_id[3];
	uint16_t	ox_id;
	uint16_t	rx_id;
	uint8_t		assoc_header[32];
} la_els_rrq_t;


#define	BNXEF_SEC2TICK(given_secs) (drv_usectohz((given_secs) * 1000000))
#define	BNXEF_GET_MTU_SIZE			 0x01
#define	BNXEF_GET_MAC_ADDR			0x02
#define	BNXEF_UNSET_CALLBACKS			0x03
#define	BNXEF_OPTIMAL_MTU_SIZE			2500
#define	BNXEF_PAGE_SIZE				(4*1024)
#define	BNXEF_PAGE_SHIFT			(12)
#define	BNXEF_PAGE_MASK				0xfff
#define	BNXEF_PAGE_ROUND_MASK			0xfffff000
#define	BNXEF_MAX_L2_RX_DESC			256
#define	BNXEF_UB_SIZE_LIMIT			2500
#define	BNXEF_UB_COUNT_LIMIT			(1024)
#define	BNXEF_MAX_TASK_ALLOC_RETRY		15
#define	BNXEF_5771X_DB_PAGE_SIZE		128
#define	BNXEF_RA_TOV				10

#ifdef BIG_ENDIAN
#define	CPU_TO_LE16(x) BSWAP_16(x)
#define	CPU_TO_LE32(x) BSWAP_32(x)
#define	LE16_TO_CPU(x) BSWAP_16(x)
#define	LE32_TO_CPU(x) BSWAP_32(x)
#define	CPU_TO_BE16(x) (x)
#define	CPU_TO_BE32(x) (x)
#define	BE16_TO_CPU(x) (x)
#define	BE32_TO_CPU(x) (x)
#else //  LITTLE_ENDIAN
#define	CPU_TO_LE16(x) (x)
#define	CPU_TO_LE32(x) (x)
#define	LE16_TO_CPU(x) (x)
#define	LE32_TO_CPU(x) (x)
#define	CPU_TO_BE16(x) BSWAP_16(x)
#define	CPU_TO_BE32(x) BSWAP_32(x)
#define	BE16_TO_CPU(x) BSWAP_16(x)
#define	BE32_TO_CPU(x) BSWAP_32(x)
#endif

/* Missing in the Leadville definition for this ELS */
#define	LA_ELS_FAN 				0x60
#define	LA_ELS_CSU				0x69
#define	LA_ELS_TEST				0x11
#define	LA_ELS_RRQ				0x12
#define	LA_ELS_REC				0x13
#define	LA_FCPLS_SRR				0x14

/* PRLI Indicating Target function bit  - Word 3 byte 3 */
#define	PRLI_TARGET_FUNCTION			0x10
#define	PRLI_INITIATOR_FUNCTION			0x20
#define	PRLI_CONF_COMPLETION_ALLOWED		0x80

/* PRLI Inidctating the RETRY and REC support - word 3 byte 2 */
#define	PRLI_RETRY_ALLOWED			0x01
#define	PRLI_REC_SUPPORT			0x04


#define	MUTEX_ENTER(a, b, c, d)   { \
		mutex_enter(b); \
		strcpy(a->filename, (char *)c); \
		a->linenum =  d; \
	}

#define	GLOBAL_mutex_enter(a, b, c, d)   { \
		mutex_enter(b); \
		strcpy(a->g_filename, (char *)c); \
		a->g_linenum =  d; \
	}

/* Prototypes of Exported functions */

int bnxef_task_cache_create(bnxef_driver_state_t *dsp);
int bnxef_task_cache_destroy(bnxef_driver_state_t *dsp);
int bnxef_task_free_list_init(bnxef_driver_state_t *dsp);
int bnxef_task_free_list_fini(bnxef_driver_state_t *dsp);
int bnxef_task_session_list_init(bnxef_driver_state_t *dsp,
			bnxef_session_t *sessionp);
int bnxef_task_session_list_fini(bnxef_session_t *sessionp);
int bnxef_task_list_add(bnxef_driver_state_t *dsp, bnxef_fcoe_task_t *taskp);

bnxef_fcoe_task_t *bnxef_task_list_get_first(bnxef_driver_state_t *dsp);
int
bnxef_task_session_list_add(bnxef_session_t *sessionp,
	bnxef_fcoe_task_t *taskp);
bnxef_fcoe_task_t *
bnxef_task_session_list_get_first(bnxef_session_t *sessionp);
void
bnxef_task_session_list_remove(bnxef_session_t *sessionp,
			bnxef_fcoe_task_t *taskp);
int
bnxef_fcf_cache_create(bnxef_driver_state_t *dsp);
int
bnxef_fcf_cache_destroy(bnxef_driver_state_t *dsp);
int
bnxef_fcf_list_init(bnxef_driver_state_t *dsp);
int
bnxef_fcf_list_fini(bnxef_driver_state_t *dsp);
bnxef_fcf_info_t *
bnxef_fcf_list_find_mac_addr(bnxef_driver_state_t *dsp, uint8_t *mac_addr);
int
bnxef_fcf_list_add(bnxef_driver_state_t *dsp, bnxef_fcf_info_t *fcfp);
int
bnxef_fcf_list_remove(bnxef_driver_state_t *dsp, bnxef_fcf_info_t *fcfp);

int
bnxef_port_cache_create(bnxef_driver_state_t *dsp);
int
bnxef_port_cache_destroy(bnxef_driver_state_t *dsp);
int
bnxef_port_list_init(bnxef_driver_state_t *dsp);
int
bnxef_port_list_fini(bnxef_driver_state_t *dsp);

bnxef_port_t *
bnxef_port_list_find_flags(bnxef_driver_state_t *dsp, uint32_t flags);
bnxef_port_t *bnxef_get_first_port(bnxef_driver_state_t *dsp);
bnxef_port_t *bnxef_port_list_find_id(bnxef_driver_state_t *dsp,
	uint8_t *macaddrp, uint16_t vlan_id, uint32_t fcid);
int bnxef_port_list_add(bnxef_driver_state_t *dsp, bnxef_port_t *portp);
int bnxef_port_list_remove(bnxef_driver_state_t *dsp, bnxef_port_t *portp);
int
bnxef_session_cache_destroy(bnxef_driver_state_t *dsp);
int
bnxef_tq_param_cache_destroy(bnxef_driver_state_t *dsp);
int bnxef_session_cache_create(bnxef_driver_state_t *dsp);
int bnxef_tq_param_cache_create(bnxef_driver_state_t *dsp);
void
bnxef_fw_resources_free(bnxef_driver_state_t *dsp, uint32_t stage);

int prv_uninit_pkt_descr(dev_info_t *dip);

boolean_t bnxef_l2_ctl(dev_info_t *dip, int cmd, void *data, int datalen);
boolean_t bnxef_l2_tx_completion(dev_info_t *dip, mblk_t *mblkp);
boolean_t bnxef_l2_rx(dev_info_t *dip, mblk_t *mblkp);
boolean_t bnxef_l2_cqe_rx(dev_info_t *dip, void *cqe[], int cnt);
int bnxef_send_fw_fcoe_uninit_msg(bnxef_driver_state_t *dsp);
int bnxef_send_fw_fcoe_init_msg(bnxef_driver_state_t *dsp);
int
bnxef_send_fw_fcoe_init_msg(bnxef_driver_state_t *dsp);
int bnxef_is_l2_interface_online(bnxef_driver_state_t *dsp);
void bnxef_fip_sm_start(void  *dsp);


#endif
