/*
 * The information contained in this file is confidential and proprietary to
 * Broadcom Corporation.  No part of this file may be reproduced or
 * distributed, in any form or by any means for any purpose, without the
 * express written permission of Broadcom Corporation.
 *
 * (c) COPYRIGHT 2010-2012 Broadcom Corporation, ALL RIGHTS RESERVED.
 */

#pragma ident "(C) 2010-2012 Broadcom Corporation $File$$Revision$$Change$$DateTime$"

#include <sys/mdb_modapi.h>
#include <sys/sysinfo.h>
#include <sys/conf.h>
#include <sys/ddi.h>
#include <sys/stat.h>
#include <sys/pci.h>
#include <sys/sunddi.h>
#include <sys/modctl.h>
#include <sys/cmn_err.h>
#include <sys/file.h>
#include <sys/cred.h>
#include <sys/byteorder.h>
#include <sys/atomic.h>
#include <sys/scsi/scsi.h>
#include <sys/modhash.h>
#include <sys/sunddi.h>
#include <sys/sunldi.h>
#include <sys/kstat.h>
#include <sys/kmem.h>
#include <sys/ethernet.h>
#include <sys/stream.h>
#include <sys/sunldi.h>

#include <sys/fibre-channel/fc_types.h>
#include <sys/fibre-channel/impl/fc_fcaif.h>
#include <bnxe_binding.h>
#include <bnxef_tq_param.h>
#include <bnxef_fcf_info.h>
#include <bnxef_hsi.h>
#include <bnxef_fcoe_task.h>
#include <bnxef_fcoe_session.h>
#include <bnxef_state.h>
#include <bnxef_fip.h>
#include <bnxef_fcoe.h>
#include <bnxef_debug.h>

char _depends_on[]= "misc/kmdbmod";

extern mdb_walker_t bnxef_walkers[];

/*
 * This dumps the Task PBL setup. This gets the address as the dsp address
 * No option need to be  provided to dump the task pbl.
 */
/*ARGSUSED1*/
static int
bnxef_task_pbl(uintptr_t addr, uint_t flags, int argc, const mdb_arg_t *argv)
{

	bnxef_driver_state_t	dsp;
	uint64_t	physicaladdress;
	uint64_t	*stateptr;
	int 		i;
	ssize_t		ret;
	uint64_t	item = 0x9999;
	uint_t  opt_n = FALSE;
	uint_t  opt_a = FALSE;

	if (mdb_getopts(argc, argv,
		'n', MDB_OPT_SETBITS, TRUE, &opt_n,
		'a', MDB_OPT_SETBITS, TRUE, &opt_a,
		'i', MDB_OPT_UINT64, &item, NULL) != argc) {
		return (DCMD_USAGE);
	}

	if (!(flags & DCMD_ADDRSPEC))
		return (DCMD_USAGE);

	if ((opt_n) && (opt_a) && (item != 0x9999)) 
		return (DCMD_USAGE);

	if ((opt_n) && (opt_a)) 
		return (DCMD_USAGE);
	
	if ((opt_n) && (item != 0x9999))
		return (DCMD_USAGE);

	if ((opt_a) && (item != 0x9999))
		return (DCMD_USAGE);


	if (mdb_vread(&dsp, sizeof (bnxef_driver_state_t), addr)
		== sizeof (dsp)) {
		if (dsp.magic == BNXEF_DRIVER_MAGIC) {
			/*LINTED: Alignment */
			stateptr = (uint64_t *)dsp.task_ctx_bd_table;
			if (opt_a) {
				mdb_printf("Task PBL List\n");
				mdb_printf("======================================\n");
				for (i = 0; i < dsp.num_task_dma_handles; i++) {
					ret = mdb_vread((void *) &physicaladdress, 8,
						(uintptr_t)stateptr);
					if (ret == -1)
						mdb_printf("Unable to read the PBL\n");
					mdb_printf(" Item  0%x --> 0x%p \n", i,
						physicaladdress);
					stateptr ++;
				}
			} else if (opt_n) {
				mdb_printf("Number of task PBL items = 0x%x\n",
					dsp.num_task_dma_handles);
			} else  {
				if (item > dsp.num_task_dma_handles) { 
					mdb_printf("Item requested is not available or within range allocated\n");
					return (DCMD_OK);
				}
				for (i= 0; i < dsp.num_task_dma_handles; i++) { 
					ret = mdb_vread((void *) &physicaladdress, 8,
						(uintptr_t)stateptr);
					if (ret == -1)
						mdb_printf("Unable to read the PBL\n");
					if (item == i) 
						mdb_printf(" Item  0x%x --> 0x%p \n", i,
							physicaladdress);
					stateptr ++;
				}
			}
			
		} else {
			mdb_printf(" Given address is not the Bnxef driver\
				 state\n");
		}
	}

	return (DCMD_OK);
}

static void
bnxef_task_pbl_help(void)
{

	mdb_printf("addr::     driver state address\n"); 
	mdb_printf("-a         Show all items in task pbl allocated\n");
	mdb_printf("-n         Display number of items allocated\n");
	mdb_printf("-i <item>  Display the requested item, <item> in hex.\n"); 
}


/*ARGSUSED1*/
static int
bnxef_hash_pbl(uintptr_t addr, uint_t flags, int argc, const mdb_arg_t *argv)
{
	bnxef_driver_state_t	dsp;
	uint64_t	physicaladdress;
	uint64_t	*stateptr;
	int 		i;
	ssize_t		ret;

	if (mdb_vread(&dsp, sizeof (bnxef_driver_state_t), addr)
			== sizeof (dsp)) {

		if (dsp.magic == BNXEF_DRIVER_MAGIC) {
			/* LINTED: Alignment */
			stateptr = (uint64_t *)dsp.hash_table_pbl;
			mdb_printf("Hash Table PBL List\n");
			mdb_printf("======================================\n");
			for (i = 0; i < dsp.hash_tbl_seg_count; i++) {
				ret = mdb_vread((void *)&physicaladdress, 8,
					(uintptr_t)stateptr);
				if (ret == -1)
					mdb_printf("Unable to read the PBL\n");
				mdb_printf(" Item  %d --> 0x%p \n", i,
					physicaladdress);
				stateptr ++;
			}
		} else {
			mdb_printf(" Given address is not the Bnxef driver\
				 state\n");
		}
	}

	return (DCMD_OK);
}


static int
bnxef_dump_hash_table(uintptr_t addr, uint_t flags, int argc,
	const mdb_arg_t *argv)
{

	uint64_t size;
	int ret;

	if (mdb_getopts(argc, argv,
		's', MDB_OPT_UINT64, &size, NULL) != argc) {
		return (DCMD_USAGE);
	}

	if (!(flags & DCMD_ADDRSPEC))
		return (DCMD_USAGE);


	ret = mdb_dumpptr(addr, size,
		MDB_DUMP_RELATIVE | MDB_DUMP_TRIM | MDB_DUMP_GROUP(1),
		(mdb_dumpptr_cb_t)mdb_pread, NULL);
	if (ret) {
		mdb_printf("Invalid Address Given\n");
		return (DCMD_ERR);
	}

	return (DCMD_OK);
}
static int
bnxef_driver_state(uintptr_t addr, uint_t flags, int argc,
	const mdb_arg_t *argv)
{
	uint_t	opt_f = FALSE;
	uint_t  opt_v = FALSE;
	uint64_t instance = 0;

	if (mdb_getopts(argc, argv,
		'v', MDB_OPT_SETBITS, TRUE, &opt_v,
		'f', MDB_OPT_SETBITS, TRUE, &opt_f,
		'i', MDB_OPT_UINT64, &instance, NULL) != argc) {
		return (DCMD_USAGE);
	}

	/* If -i is not given zero is assumed */
	if (!(flags & DCMD_ADDRSPEC))
		return (DCMD_USAGE);

	if (opt_f) {
		mdb_printf("Opt F is set");
		if (mdb_pwalk_dcmd("bnxef_task_free_list_walker", "bnxef_task",
			0, NULL, addr) == -1) {
			mdb_warn("Failed to walk: 'bnxef_task_free_list_walk'");
			return (DCMD_ERR);
		}
		return (DCMD_OK);
	}

	if (opt_v) {
		mdb_printf("Not Implemented Yet\n");

	}

	return (DCMD_OK);
}

/* ARGSUSED3 */
static int
bnxef_fcoe_task(uintptr_t addr, uint_t flags, int argc, const mdb_arg_t *argv)
{
	bnxef_fcoe_task_t task;

	if (argc != 0)
		return (DCMD_USAGE);

	if (flags & DCMD_ADDRSPEC) {
		if (mdb_vread(&task, sizeof (bnxef_fcoe_task_t), addr) !=
			sizeof (bnxef_fcoe_task_t)) {
			mdb_warn("Unable to read task structure");
			return (DCMD_ERR);
		}

		if (task.task_flags == BNXEF_FCOE_TASK_FREE) {
			mdb_printf(" Task 0x%x -> Free", task.x_id);
		} else {
			mdb_printf("Task Id 0x%x, flags 0x%x, pktp 0x%p,\
			fwtaskp 0x%p\n", task.x_id, task.task_flags,
				task.pktp, task.fwtaskp);
		}
		return (DCMD_OK);
	} else {
		return (DCMD_USAGE);
	}
}


static int
bnxef_dump_all_session_task_log (uintptr_t addr, uint_t flags, int argc, const mdb_arg_t *argv)
{
	int i,j, size;
	int num_instances;
	uint64_t *global_state;
	bnxef_driver_state_t *dsp;
	uint64_t *sessarray;
	mdb_arg_t myargv;

	mdb_readvar(&num_instances, "num_instances");

	global_state = mdb_alloc(num_instances * sizeof (uint64_t),
			UM_SLEEP);
	if (global_state != NULL) {
		size = mdb_readsym(global_state,
			BNXEF_MAX_INSTANCES * sizeof (uint64_t),
			"global_state");
		if (size != (BNXEF_MAX_INSTANCES * sizeof (uint64_t))) {
			mdb_printf("Unable to read State ptrs\n");
			return (DCMD_ERR);
		}
	}

	/* Read the Global State Structure into Local structure */
	dsp = mdb_alloc(sizeof (bnxef_driver_state_t),  UM_SLEEP);
	if (dsp == NULL) {
		mdb_printf("Unable to allocate State structure\n");
		return (DCMD_ERR);
	}

	for (i = 0; i < BNXEF_MAX_INSTANCES; i++) {
		if (global_state[i] != NULL) {
			mdb_printf("Attached Instance Number %d\n", i);
			/* Read the Driver State Structure in the memory allocated */
			if (mdb_vread(dsp,
				sizeof (bnxef_driver_state_t), global_state[i])
				!= sizeof (bnxef_driver_state_t)) {
				mdb_printf("Unable to read state information\n");
				return (DCMD_ERR);
			}

			sessarray = mdb_alloc((dsp->num_sessions * sizeof(uint64_t)), UM_SLEEP);
			mdb_printf("Session address pointer 0x%p", (uint64_t)dsp->sessp);
			if (dsp->sessp != NULL) {
				if (mdb_vread(sessarray, (dsp->num_sessions * sizeof(uint64_t)),
						(uint64_t)dsp->sessp) != (dsp->num_sessions *
							sizeof(uint64_t))) {
					mdb_printf("unable to read state_information");
					return (DCMD_ERR);
				}
		

				mdb_printf("Number of sessions 0x%x\n", dsp->num_sessions);
				for (j=0; j < dsp->num_sessions; j++) {
					mdb_printf("Got sessions 0x%p, %d\n",
						sessarray[j], j);
					if (mdb_call_dcmd("bnxef_completion_log",  sessarray[j],
						DCMD_ADDRSPEC, 1, &myargv) < 0) {
						mdb_printf("Call to DCMD failed\n");
						return (DCMD_ERR);
					}
				}
			}
		}
	}

	return (DCMD_OK);
				
}
static int
bnxef_task_completion_log(uintptr_t addr, uint_t flags, int argc, const mdb_arg_t *argv)
{
	bnxef_session_t session; 
	int i, j;
	char  type_string[128];
	bnxef_completion_log_t *logp;


	/* If -i is not given zero is assumed */
	if (!(flags & DCMD_ADDRSPEC))
		return (DCMD_USAGE);

	if (mdb_vread(&session, sizeof (bnxef_session_t), addr) !=
		sizeof (bnxef_session_t)) {
			mdb_warn("Unable to read session structure");
			return (DCMD_ERR);
	}

	if ((session.state & BNXEF_SESSION_OFFLOADED) == 0) 
		return(DCMD_OK);

	mdb_printf("Session: 0x%p\n", addr);
	mdb_printf("taskid\tpkt_state\tpkt_reason\tsession_state\tStart lbolt\tEnd Lbolt\tPktp\t task_flags\n");
	mdb_printf("======================================================================================\n");

	
	for (i=session.log_cur_idx; i< BNXEF_MAX_LOG_ENTRIES; i++) {
		logp = &session.c_log[i];
		mdb_printf("0x%x\t0x%x\t\t0x%x\t0x%x\t0x%p\t\t0x%p\t\t0x%p\t0x%x\n",
			logp->x_id, logp->pkt_state, logp->pkt_reason,
			logp->session_state, logp->start_lbolt, logp->end_lbolt,
			logp->pktp, logp->task_flags);
	}	

	if (session.log_cur_idx > 0) {
		for(i=0; i< session.log_cur_idx; i++) {
			logp = &session.c_log[i];
			mdb_printf("0x%x\t0x%x\t\t0x%x\t0x%x\t0x%p\t\t0x%p\t\t0x%p\t0x%x\n",
				logp->x_id, logp->pkt_state, logp->pkt_reason,
				logp->session_state, logp->start_lbolt, logp->end_lbolt,
				logp->pktp, logp->task_flags);
		}	
	}

	return (DCMD_OK);

}


static int
bnxef_session_life_cycle_log (uintptr_t addr, uint_t flags, int argc, const mdb_arg_t *argv)
{
	bnxef_port_t port;
	int i, j;
	char  type_string[128];


	/* If -i is not given zero is assumed */
	if (!(flags & DCMD_ADDRSPEC))
		return (DCMD_USAGE);

	if (mdb_vread(&port, sizeof (bnxef_port_t), addr) !=
		sizeof (bnxef_port_t)) {
			mdb_warn("Unable to read port structure");
			return (DCMD_ERR);
	}

	mdb_printf("IDX \t\t\t Session Ptr \t\t\t\t Type \t\t\t lbolt\n");
	mdb_printf("======================================================================================\n");
	for (i=port.slcpidx; i< BNXEF_SLCP_MAX_IDX; i++) {
		if (port.slcp[i].type == BNXEF_SESSION_ENABLE_TYPE) 
			strcpy(type_string, "Enable ");
		else if (port.slcp[i].type == BNXEF_SESSION_DISABLE_TYPE)
			strcpy(type_string, "Disable");
		else if (port.slcp[i].type == BNXEF_SESSION_OFFLOAD_TYPE)
			strcpy(type_string, "Offload");
		else if (port.slcp[i].type == BNXEF_SESSION_DESTROY_TYPE)
			strcpy(type_string, "Destroy");

		mdb_printf("%d \t\t\t 0x%p \t\t\t %s\t\t\t 0x%p\n",
			i, port.slcp[i].sessionp,
			type_string,
			port.slcp[i].lbolt64);
	}

	if (port.slcpidx > 0) {
		for (i=0; i<port.slcpidx; i++) {
			if (port.slcp[i].type == BNXEF_SESSION_ENABLE_TYPE) 
				strcpy(type_string, "Enable ");
			else if (port.slcp[i].type == BNXEF_SESSION_DISABLE_TYPE)
				strcpy(type_string, "Disable");
			else if (port.slcp[i].type == BNXEF_SESSION_OFFLOAD_TYPE)
				strcpy(type_string, "Offload");
			else if (port.slcp[i].type == BNXEF_SESSION_DESTROY_TYPE)
				strcpy(type_string, "Destroy");
			mdb_printf("%d \t\t\t 0x%p \t\t\t %s\t\t\t 0x%p\n",
			i, port.slcp[i].sessionp,
			type_string,
			port.slcp[i].lbolt64);
		}
	}
}
			


static int
bnxef_port_info(uintptr_t addr, uint_t flags, int argc, const mdb_arg_t *argv)
{
	bnxef_port_t port;
	uint_t	opt_v = FALSE;
	uint_t	opt_s = FALSE;
	uint_t  opt_u = FALSE;
	uint_t	opt_H = FALSE;
	uint64_t hash = 0xFFFF;
	uintptr_t sessionaddr;
	int i, j;


	if (mdb_getopts(argc, argv,
		'v', MDB_OPT_SETBITS, TRUE, &opt_v,
		'u', MDB_OPT_SETBITS, TRUE, &opt_u,
		'H', MDB_OPT_SETBITS, TRUE, &opt_H,
		'h', MDB_OPT_UINT64, &hash, NULL) != argc) {
		return (DCMD_USAGE);
	}

	/* If -i is not given zero is assumed */
	if (!(flags & DCMD_ADDRSPEC))
		return (DCMD_USAGE);

	if (mdb_vread(&port, sizeof (bnxef_port_t), addr) !=
		sizeof (bnxef_port_t)) {
			mdb_warn("Unable to read port structure");
			return (DCMD_ERR);
	}


	if (opt_H) {
		mdb_printf("\nHash #\tSession Address\t\tLport Addr\t\tRport Addr\t\tNUM I/O pending\tSession State\n");
		mdb_printf("======================================================================================================\n"); 
		for (i = 0; i < 256; i++) {
			sessionaddr = (uintptr_t)port.sessionhash[i].head;
			for (j=0; j< port.sessionhash[i].num_items; j++) {
				bnxef_session_t localsession;
				if (sessionaddr != NULL) {
					mdb_printf("0x%02X", i);

					if (mdb_pwalk_dcmd("bnxef_session_list_walker",
						"bnxef_session", 0, NULL, sessionaddr) == -1) {
						mdb_warn("Failed walk:bnxef_session_list_walk");
						return (DCMD_ERR);
					}
					if (mdb_vread(&localsession, sizeof(bnxef_session_t), sessionaddr) !=
						sizeof (bnxef_session_t)) {
						mdb_warn("Unable to read the session address\n");
						return (DCMD_ERR);
					}
					sessionaddr = (uintptr_t) localsession.next;
				}
			}
		}
		return (DCMD_OK);
	}

	if (hash != 0xFFFF) {
		if (hash > 255)
			return (DCMD_USAGE);
		mdb_printf("%d Connections Present in Hash Bucket %d\n",
				port.sessionhash[hash].num_items, hash);
		mdb_printf("============================================\n");
		sessionaddr = (uintptr_t)port.sessionhash[hash].head;
		if (mdb_pwalk_dcmd("bnxef_session_list_walker", "bnxef_session",
			0, NULL, sessionaddr) == -1) {
			mdb_warn("Failed to walk: 'bnxef_session_list_walk'");
			return (DCMD_ERR);
		}
		return (DCMD_OK);
	}

	if (opt_u) {
		mdb_printf("Not Implemented\n");
		return (DCMD_OK);
	}

	if (opt_s) {
		mdb_printf(" Not IMplemented\n");
		return (DCMD_OK);
	}

	if (opt_v) {
	    mdb_printf("Port Id\t VLAN ID\t MAC Address\t\tFlags\tFIP State\n");
		mdb_printf("============================================================\n");
		mdb_printf("0x%x\t 0x%x\t   %x:%x:%x:%x:%x:%x\t\t0x%x\t0x%x\n",
		port.id, port.vlan_id, port.mac_addr[5],
		port.mac_addr[4], port.mac_addr[3], port.mac_addr[2],
		port.mac_addr[1], port.mac_addr[0], port.flags, port.fip_state);
		return (DCMD_OK);
	}

	return (DCMD_OK);

}

static int
bnxef_dump_sq(uintptr_t addr, uint_t flags, int argc, const mdb_arg_t *argv)
{
	uint_t opt_t = FALSE;
	uint64_t prodidx;
	uint64_t x_id;
	uint8_t  *sqes;
	uint8_t sq_toggle = 0;
	uint8_t sq_toggle_neg;
	uint16_t size;
	uint16_t wqe;
	int i;

	const mdb_arg_t *margv = argv;


	if (mdb_getopts(argc, argv,
		's', MDB_OPT_UINT64, &size,
		'p', MDB_OPT_UINT64, &prodidx,
		't', MDB_OPT_SETBITS, TRUE, &opt_t,
		NULL)  != argc) {
		return (DCMD_USAGE);
	}

	if (!(flags & DCMD_ADDRSPEC))
		return (DCMD_USAGE);

	sqes = mdb_alloc(size, UM_SLEEP);
	if (sqes == NULL) {
		mdb_printf("Allocation failed");
		return (DCMD_ERR);
	}

	if (opt_t == TRUE) {
		sq_toggle = 1;
		sq_toggle_neg = 0;
	}
	mdb_printf("Size 0x%x, toggle 0x%x considx 0x%x\n", size, sq_toggle,
		prodidx);
	if (mdb_vread(sqes, size, addr) != size) {
		mdb_printf("Unable read the CQ Area\n");
		return (DCMD_ERR);
	}

	mdb_printf("X_ID Submitted to firmware\n");
	mdb_printf("=====================================================\n");

	for (i = prodidx; i < size/sizeof (uint16_t); i++) {
		/* LINTED: Alignment */
		wqe = *((uint16_t *)(sqes + i * sizeof (uint16_t)));
		if ((wqe & FCOE_SQE_TOGGLE_BIT) ==
			(sq_toggle_neg  << FCOE_SQE_TOGGLE_BIT_SHIFT)) {
				x_id =  wqe & FCOE_PEND_WQ_CQE_TASK_ID;
			mdb_printf("%04x ", x_id);
		}
		if ((i % 16) == 0) mdb_printf("\n");
	}

	mdb_printf("\nRestarting from Beginning\n");
	for (i = 0; i < prodidx; i++) {
		/* LINTED: Alignment */
		wqe = *((uint16_t *)(sqes + i * sizeof (uint16_t)));
		if ((wqe & FCOE_SQE_TOGGLE_BIT) ==
			(sq_toggle  << FCOE_SQE_TOGGLE_BIT_SHIFT)) {
			x_id =  wqe & FCOE_PEND_WQ_CQE_TASK_ID;
			mdb_printf("%04x ", x_id);
		}
		if ((i % 16) == 0) mdb_printf("\n");
	}
	return (DCMD_OK);

}


static int
bnxef_dump_cq(uintptr_t addr, uint_t flags, int argc, const mdb_arg_t *argv)
{
	uint_t opt_t = FALSE;
	uint64_t considx;
	uint64_t x_id;
	uint8_t  *cqes;
	uint8_t cq_toggle = 0;
	uint8_t cq_toggle_neg;
	uint16_t size;
	uint16_t wqe;
	int i;

	if (mdb_getopts(argc, argv,
		's', MDB_OPT_UINT64, &size,
		'c', MDB_OPT_UINT64, &considx,
		't', MDB_OPT_SETBITS, TRUE, &opt_t,
		NULL)  != argc) {
		return (DCMD_USAGE);
	}

	if (!(flags & DCMD_ADDRSPEC))
		return (DCMD_USAGE);

	cqes = mdb_alloc(size, UM_SLEEP);
	if (cqes == NULL) {
		mdb_printf("Allocation failed");
		return (DCMD_ERR);
	}

	if (opt_t == TRUE) {
		cq_toggle = 1;
		cq_toggle_neg = 0;
	}
	mdb_printf("Size 0x%x, toggle 0x%x considx 0x%x\n", size, cq_toggle,
		considx);
	if (mdb_vread(cqes, size, addr) != size) {
		mdb_printf("Unable read the CQ Area\n");
		return (DCMD_ERR);
	}

	mdb_printf("X_ID completed\n");
	mdb_printf("=====================================================\n");

	for (i = considx; i < size/sizeof (uint16_t); i++) {
		/* LINTED: Alignment */
		wqe = *((uint16_t *)(cqes + i * sizeof (uint16_t)));
		if (wqe  & FCOE_CQE_CQE_TYPE) {
			mdb_printf("Unsolicited Event 0x%x\n", wqe);
			continue;
		}
		if ((wqe & FCOE_CQE_TOGGLE_BIT) ==
			(cq_toggle_neg  << FCOE_CQE_TOGGLE_BIT_SHIFT)) {
				x_id =  wqe & FCOE_PEND_WQ_CQE_TASK_ID;
			mdb_printf("%04x ", x_id);
		}
		if ((i % 16) == 0) mdb_printf("\n");
	}

	mdb_printf("\nRestarting from Beginning\n");
	for (i = 0; i < considx; i++) {
		/* LINTED: Alignment */
		wqe = *((uint16_t *)(cqes + i * sizeof (uint16_t)));
		if (wqe  & FCOE_CQE_CQE_TYPE) {
			mdb_printf("Unsolicited Event 0x%x\n", wqe);
			continue;
		}
		if ((wqe & FCOE_CQE_TOGGLE_BIT) ==
			(cq_toggle  << FCOE_CQE_TOGGLE_BIT_SHIFT)) {
			x_id =  wqe & FCOE_PEND_WQ_CQE_TASK_ID;
			mdb_printf("%04x ", x_id);
		}
		if ((i % 16) == 0) mdb_printf("\n");
	}
	return (DCMD_OK);
}

static int
bnxef_t2_hash_tbl(uintptr_t addr, uint_t flags, int argc,
	const mdb_arg_t *argv)
{

	struct fcoe_t2_hash_table_entry t2entry;
	uint_t opt_a = FALSE;
	int i;

	if (mdb_getopts(argc, argv,
		'a', MDB_OPT_SETBITS, TRUE, &opt_a, NULL) != argc) {
		return (DCMD_USAGE);
	}

	if (!(flags & DCMD_ADDRSPEC))
		return (DCMD_USAGE);

	for (i = 0; i < 2048; i++) {
		if (mdb_vread(&t2entry,
			sizeof (struct fcoe_t2_hash_table_entry), addr) !=
			sizeof (struct fcoe_t2_hash_table_entry)) {
			return (DCMD_ERR);
		}
		mdb_printf("t2 Entry %d:-> 0x%x%x\n", i, t2entry.next.hi,
				t2entry.next.lo);
		addr += sizeof (struct fcoe_t2_hash_table_entry);
	}

	return (DCMD_OK);
}


static int
bnxef_session_info(uintptr_t addr, uint_t flags, int argc,
	const mdb_arg_t *argv)
{
	bnxef_session_t session;
	bnxef_port_t	lport;
	bnxef_port_t	rport;
	uint_t	opt_v = FALSE;
	uint_t  opt_T = FALSE;
	uint32_t dma_size;
	struct regpair pbl;
	uintptr_t taskaddr;
	int i;

	if (mdb_getopts(argc, argv,
		'T', MDB_OPT_SETBITS, TRUE, &opt_T,
		'v', MDB_OPT_SETBITS, TRUE, &opt_v, NULL) != argc) {
		return (DCMD_USAGE);
	}

	/* If -i is not given zero is assumed */
	if (!(flags & DCMD_ADDRSPEC))
		return (DCMD_USAGE);

	if (mdb_vread(&session, sizeof (bnxef_session_t), addr) !=
		sizeof (bnxef_session_t)) {
			mdb_warn("Unable to read port structure");
			return (DCMD_ERR);
	}
	if (opt_T) {
		taskaddr = (uintptr_t)session.sessiontasklist.head;
		if (taskaddr == NULL) {
			mdb_printf("Nothing Pending\n");
			return (DCMD_ERR);
		}

		if (mdb_pwalk_dcmd("bnxef_session_task_list_walker",
				"bnxef_task", 0, NULL, taskaddr) == -1) {
			mdb_warn("Failed to walk: 'bnxef_session_list_walk'");
			return (DCMD_ERR);
		}
		return (DCMD_OK);
	}



	if (opt_v) {
		if (mdb_vread(&lport, sizeof (bnxef_port_t),
			(uintptr_t)session.lport) != sizeof (bnxef_port_t)) {
			mdb_warn("Unable to read the Port Structure\n");
			return (DCMD_ERR);
		}
		if (mdb_vread(&rport, sizeof (bnxef_port_t),
			(uintptr_t)session.rport) != sizeof (bnxef_port_t)) {
			mdb_warn("Unable to read the Remote port Structure\n");
			return (DCMD_ERR);
		}
		mdb_printf("Session: Lport ID 0x%x ---> Rport ID 0x%x\n",
			lport.id, rport.id);
		mdb_printf("RDF size of Lport is 0x%x, and of Rport is 0x%x\n",
			lport.plogi_info.class_3.rcv_size,
			rport.plogi_info.class_3.rcv_size);
		mdb_printf("SQ information:\n");
		mdb_printf("\t  SQEs configured %d\n", session.max_sqes);
		mdb_printf("\t  SQ address 0x%p\n", session.sq_addr);
		mdb_printf("\t  SQ length 0x%x\n", session.sq_length);
		mdb_printf("\t  SQ prod Index 0x%x\n", session.sq_prod_idx);
		mdb_printf("\t  SQ toggle 0x%x\n", session.sq_toggle);
		mdb_printf("\t  SQ Cookie Count 0x%x\n\n",
				session.sq_cookie_cnt);

		mdb_printf("CQ Information:\n");
		mdb_printf("\t CQEs configured %d\n", session.max_cqes);
		mdb_printf("\t CQ address 0x%p\n", session.cq_addr);
		mdb_printf("\t CQ length 0x%x\n", session.cq_length);
		mdb_printf("\t CQ Consumer Index 0x%x\n", session.cq_cons_idx);
		mdb_printf("\t CQ Toggle 0x%x\n", session.cq_toggle);
		mdb_printf("\t CQ Cookie Count 0x%x\n\n",
				session.cq_cookie_cnt);

		mdb_printf("RQ Information:\n");
		mdb_printf("\t RQEs configured %d\n", session.max_rqes);
		mdb_printf("\t RQ address 0x%p\n", session.rq_addr);
		mdb_printf("\t RQ Lenth 0x%x\n", session.rq_length);
		mdb_printf("\t RQ Producer Index 0x%x\n", session.rq_prod_idx);
		mdb_printf("\t RQ Consumer Index 0x%x\n", session.rq_cons_idx);
		mdb_printf("\t RQ Cookie Count 0x%x\n\n",
			session.rq_cookie_cnt);

		mdb_printf("RQ Page Buffer List Information:\n");
		mdb_printf("\t RQ Page Buffer length 0x%x\n",
			session.rq_pbl_length);
		mdb_printf("\t RQ Page Buffer Starting address 0x%p\n",
			session.rq_pbl_addr);
		mdb_printf("\t RQ PBL List:\n");
		dma_size = session.rq_cookie.dmac_size;
		i = 0;
		while (dma_size) {
			pbl.hi = (session.rq_cookie.dmac_laddress +
					i * BNXEF_PAGE_SIZE) >> 32;
			pbl.lo = (session.rq_cookie.dmac_laddress +
				i * BNXEF_PAGE_SIZE) & 0xFFFFFFFF;
			mdb_printf("\t\t Entry %d : 0x%x%x\n",
					i, pbl.hi, pbl.lo);
			if (dma_size  < BNXEF_PAGE_SIZE) break;
			dma_size -= BNXEF_PAGE_SIZE;
			i++;
		}
		mdb_printf("XFERQ Information:\n");
		mdb_printf("\t XFERQ Address 0x%p\n", session.xferq_addr);
		mdb_printf("\t XFERQ length   0x%x\n", session.xferq_length);
		mdb_printf("\t XFERQ Alloc size 0x%x\n", session.xferq_sz);
		mdb_printf("\t XFERQ cookie Count 0x%x\n",
				session.xferq_cookie_cnt);

		mdb_printf("CONFQ Information:\n");
		mdb_printf("\t CONFQ Address 0x%p\n", session.confq_addr);
		mdb_printf("\t CONFQ length 0x%x\n", session.confq_length);
		mdb_printf("\t CONFQ Actual Size alloacted 0x%x\n",
			session.confq_sz);
		mdb_printf("\t CONFQ Cookie count 0x%x\n",
			session.confq_cookie_cnt);

		mdb_printf("CONFQ PBL Information:\n");
		mdb_printf("\t CONFQ PBL Address 0x%p\n",
			session.confq_pbl_addr);
		mdb_printf("\t CONFQ PBL Length 0x%x\n",
			session.confq_pbl_length);
		mdb_printf("\t CONFQ PBL Actual size allocated 0x%x\n",
			session.confq_pbl_sz);
		mdb_printf("\t CONFQ PBL List:\n");
		dma_size = session.confq_cookie.dmac_size;
		i = 0;
		while (dma_size) {
			pbl.hi = (session.confq_cookie.dmac_laddress +
					i * BNXEF_PAGE_SIZE) >> 32;
			pbl.lo = (session.confq_cookie.dmac_laddress +
					i * BNXEF_PAGE_SIZE) & 0xFFFFFFFF;
			mdb_printf("\t\t Entry %d : 0x%x%x\n",
					i, pbl.hi, pbl.lo);
			if (dma_size  < BNXEF_PAGE_SIZE) break;
			dma_size -= BNXEF_PAGE_SIZE;
			i++;
		}
		mdb_printf("\n");

		mdb_printf("LCQ information:\n");
		mdb_printf("\t LCQ Address 0x%p\n", session.lcq_addr);
		mdb_printf("\t LCQ Length 0x%x\n", session.lcq_length);
		mdb_printf("\t LCQ actual size allocated 0x%x\n",
				session.lcq_sz);
		mdb_printf("\t LCQ Cookie count 0x%x\n\n",
			session.lcq_cookie_cnt);

		mdb_printf("Connection DB information:\n");
		mdb_printf("\t Connection DB address 0x%p\n",
			session.conn_db_addr);
		mdb_printf("\t Connection DB address length 0x%x\n",
			session.conn_db_length);
		mdb_printf("\t connection DB Cookie Count 0x%x\n",
			session.conn_db_cookie_cnt);
	} else {
		mdb_printf("\t0x%p\t0x%p\t0x%p\t0x%x\t\t0x%x\n",
			addr, session.lport, session.rport,
			session.sessiontasklist.num_items,session.state);
	}

	return (DCMD_OK);
}


void
bnxef_dump_physical_port_info(bnxef_driver_state_t *dsp)
{
	mdb_printf("\nPhysical Port Information - Instance %d\n", dsp->instance);
	mdb_printf("=========================================\n");

	mdb_printf("The physical port is %s\n", (dsp->physical_port_state? "Online":"offline"));
	mdb_printf("The devinfo pointer 0x%p\n", dsp->dip);
	mdb_printf("Mac Address %x:%x:%x:%x:%x:%x\n",
		dsp->mac_add[0], dsp->mac_add[1],
		dsp->mac_add[2], dsp->mac_add[3],
		dsp->mac_add[4], dsp->mac_add[5]);
	mdb_printf("Mtu size %d\n", dsp->mtu_size);

}


void
bnxef_dump_fcf_info(bnxef_driver_state_t *dsp)
{
	int i;
	bnxef_fcf_info_t fcf;
	mdb_arg_t argv;
	uintptr_t addr;


	mdb_printf("\nFCF Information - Instance %d\n", dsp->instance);
	mdb_printf("===================================\n");

	addr = (uintptr_t) dsp->fcf_list.head;
	for (i = 0; i < dsp->fcf_list.num_items; i++) {
		mdb_printf("FCF No. %d - %s \n", 
				i, ((uintptr_t) dsp->fcf_ident == addr)? "Selected": "Not Selected");
		if (mdb_call_dcmd("bnxef_fcf_info",  addr,
			 DCMD_ADDRSPEC, 1, &argv) < 0) {
			mdb_printf("Call to DCMD failed\n");
			return;
		}
		if (mdb_vread(&fcf, sizeof (bnxef_fcf_info_t), addr) !=
			sizeof (bnxef_fcf_info_t)) {
			mdb_printf("Unable to read the FCF info\n"); 
			return;
		}
		addr = (uintptr_t) fcf.next;
	}
}

void
bnxef_dump_config_info(bnxef_driver_state_t *dsp)
{

	mdb_printf("\nConfiguration Parameters- Instance %d\n", dsp->instance);
	mdb_printf("=======================================\n");
	mdb_printf("Max Exchanges supported - %d\n", dsp->max_exchanges);
	mdb_printf("Max Sessions Supported - %d\n", dsp->max_sessions);
	mdb_printf("Max Fc Ports Supported - %d\n", dsp->max_fc_ports);

	

}

void
bnxef_dump_free_task_list_info(bnxef_driver_state_t *dsp)
{

	mdb_printf("\nFree Task List Information - Instance %d\n", dsp->instance);
	mdb_printf("============================================\n");

	mdb_printf(" Number of Free Task available %d\n", dsp->bnxef_task_free_list.num_items);

}

void
bnxef_dump_task_res_info(bnxef_driver_state_t *dsp)
{

	int i;
	bnxef_fcoe_task_t task; 
	uintptr_t addr;
	uintptr_t *taskarray;	

	mdb_printf("\nTask Information - Instance %d\n", dsp->instance);
	mdb_printf("\nX_ID\t Task Vaddr\t\tTaskContext Ptr\t\tTask State\tSession Ptr\n");
	mdb_printf("==========================================================================================\n");

	addr = (uintptr_t) dsp->taskarray;
	for (i = 0; i < dsp->max_exchanges; i++) {
		taskarray = mdb_alloc(sizeof (uintptr_t), UM_SLEEP);
		if (taskarray == NULL) {
			mdb_printf("Unable to allocate space\n");
			return;
		}	

		if (mdb_vread(taskarray, sizeof (uintptr_t), addr) !=
			sizeof (uintptr_t)) {
			mdb_printf("Unable to read mdb taskarray\n");
			return;
		}
		if (mdb_vread(&task, sizeof (bnxef_fcoe_task_t), (uintptr_t) (*taskarray))  !=
			sizeof (bnxef_fcoe_task_t)) { 
			mdb_printf("Unable to read Task Address\n");
			return;
		}
		mdb_printf("0x%04X\t 0x%p \t 0x%p \t 0x%05X\t 0x%p\n", 
			task.x_id, addr, task.fwtaskp, task.task_flags, task.sessionp);

		addr += sizeof (uintptr_t);
		mdb_free(taskarray, sizeof (uintptr_t));
	}

}
		
		
void
bnxef_dump_global_fw_res_info(bnxef_driver_state_t *dsp)
{

	mdb_printf("\nGlobal firmware resource information - Instance %d\n", dsp->instance);
	mdb_printf("====================================================\n");

	mdb_printf("Hash Table Pointer 0x%p, Hash Table Length 0x%x\n",
			dsp->hash_tbl, dsp->hash_tbl_len);
	mdb_printf("Hash Table Buffer Ptr 0x%p, Hash Table PBL length 0x%x\n",
			dsp->hash_table_pbl, dsp->hash_table_pbl_len);
	mdb_printf("T2 hash table Ptr 0x%p, T2 hash table length 0x%x\n",
			dsp->t2_hash_tbl_ptr, dsp->t2_hash_tbl_ptr_len);
	mdb_printf("Dummy Buffer ptr 0x%p, Dummy buffer Size 0x%x\n",
			dsp->dummy_buffer, dsp->dummy_size);
}
		
void
bnxef_dump_fw_res_info(bnxef_driver_state_t *dsp)
{

	bnxef_dump_global_fw_res_info(dsp);

	/* THis will dump the task address, XID, FW ctx and task state */
	bnxef_dump_task_res_info(dsp);


}

void
bnxef_dump_port_session_info(bnxef_port_t *portp)
{

	int i, j;
	uintptr_t sessionaddr;
	mdb_arg_t argv[5];
	char options[128];
	char options1[128];
	char options2[128];
	uintptr_t sqaddr,cqaddr;
	int argc;

	/* Get each one of the session from the hash chain */
	for (i = 0; i < 256; i++) {
		sessionaddr = (uintptr_t)portp->sessionhash[i].head;
		for (j=0; j< portp->sessionhash[i].num_items; j++) {
			bnxef_session_t localsession;
			if (sessionaddr != NULL) {
				mdb_printf("\nComplete Information about Session address: 0x%p\n", sessionaddr );
				mdb_printf("==================================================================\n");
				argv[0].a_type = MDB_TYPE_STRING;
				strcpy(options, "-v");
				argv[0].a_un.a_str = options;
				if (mdb_pwalk_dcmd("bnxef_session_list_walker",
					"bnxef_session", 1, &argv[0], sessionaddr) == -1) {
					mdb_warn("Failed walk:bnxef_session_list_walk");
					return;
				}
				mdb_printf("List of tasks pending on this session: 0x%p\n", sessionaddr);
				mdb_printf("============================= ==============================\n");

				argv[0].a_type = MDB_TYPE_STRING;
				strcpy(options, "-T");
				argv[0].a_un.a_str = options;
				if (mdb_pwalk_dcmd("bnxef_session_list_walker",
					"bnxef_session", 1, 
					&argv[0], sessionaddr) == -1) {
					mdb_warn("Failed walk:\
						bnxef_session_list_walk");
					return;
				}

				if (mdb_vread(&localsession,
					sizeof (bnxef_session_t),
					 sessionaddr) !=
						sizeof (bnxef_session_t)) {
						mdb_warn("Unable to read\
							the session address\n");
						return;
				}

				sqaddr = (uintptr_t)localsession.sq_addr;
				if (sqaddr != NULL) {
					argv[0].a_type = MDB_TYPE_STRING;
					sprintf(options, "-s 0x%x",
						localsession.sq_length);
					sprintf(options1, "-p 0x%x",
						localsession.sq_prod_idx);
					snprintf(options2, 4, "%s",
						(localsession.sq_toggle)?"-t":"");
					argv[0].a_un.a_str = options;
					options[2] = '\0';
					argv[1].a_type = MDB_TYPE_STRING;
					argv[1].a_un.a_str = (options + 3);
					argv[2].a_type = MDB_TYPE_STRING;
					argv[2].a_un.a_str = options1;
					argv[3].a_type = MDB_TYPE_STRING;
					options1[2] = '\0';
					argv[3].a_un.a_str = options1 + 3;
					argv[4].a_type = MDB_TYPE_STRING;
					if (localsession.sq_toggle) {
						argv[4].a_un.a_str = options2;
						argc = 5;
					} else argc = 4;

					if (mdb_call_dcmd("bnxef_sqdump",
						sqaddr, DCMD_ADDRSPEC, argc,
						argv) < 0) {
						mdb_warn("Unable to\
							execute the DCMD\n");
						return;
					}
				}

				cqaddr = (uintptr_t)localsession.cq_addr;
				if (cqaddr != NULL) {
					argv[0].a_type = MDB_TYPE_STRING;
					sprintf(options, "-s 0x%x",
						localsession.cq_length);
					sprintf(options1, "-c 0x%x",
						localsession.cq_cons_idx);
					snprintf(options2, 4, "%s",
					(localsession.cq_toggle)?"-t":"");
					argv[0].a_un.a_str = options;
					options[2] = '\0';
					argv[1].a_type = MDB_TYPE_STRING;
					argv[1].a_un.a_str = (options + 3);
					argv[2].a_type = MDB_TYPE_STRING;
					argv[2].a_un.a_str = options1;
					argv[3].a_type = MDB_TYPE_STRING;
					options1[2] = '\0';
					argv[3].a_un.a_str = options1 + 3;
					argv[4].a_type = MDB_TYPE_STRING;
					if (localsession.cq_toggle) {
						argv[4].a_un.a_str = options2;
						argc = 5;
					} else argc = 4;

					if (mdb_call_dcmd("bnxef_cqdump",
						cqaddr, DCMD_ADDRSPEC,
						argc, argv) < 0) {
						mdb_warn("Unable to \
							execute the DCMD\n");
						return;
					}
				}
				sessionaddr = (uintptr_t)localsession.next;
			}
		}
	}


}


void
bnxef_dump_nport_info(bnxef_driver_state_t *dsp)
{

	bnxef_port_t port;
	uintptr_t addr;
	int i;
	mdb_arg_t argv;
	char  options[128];

	mdb_printf("\nPort Information - Instance %d\n", dsp->instance);
	mdb_printf("==================================\n");

	mdb_printf("Number of Instantiated Ports 0x%x\n\n", dsp->num_fc_ports);

	addr = (uintptr_t)dsp->port_list.head;
	for (i = 0; i < dsp->port_list.num_items; i++) {
		if (mdb_vread(&port, sizeof (bnxef_port_t), addr) !=
			sizeof (bnxef_port_t)) {
			mdb_printf("Unable to read port information\n");
			return;
		}

		argv.a_type = MDB_TYPE_STRING;
		strcpy(options, "-v");
		argv.a_un.a_str = options;


		if (mdb_call_dcmd("bnxef_port",  addr,
			DCMD_ADDRSPEC, 1, &argv) < 0) {
			mdb_printf("Call to DCMD failed\n");
			return;
		}

		argv.a_type = MDB_TYPE_STRING;
		strcpy(options, "-H");
		argv.a_un.a_str = options;
		if (mdb_call_dcmd("bnxef_port", addr,
			DCMD_ADDRSPEC, 1, &argv) < 0) {
			mdb_printf("Call to DCMD failed");
			return;
		}

		bnxef_dump_port_session_info(&port);

		addr = (uintptr_t)port.next;
	}
}

/* ARGSUSED2 */
static int
bnxef_fcf_info(uintptr_t addr, uint_t flags, int argc,
	const mdb_arg_t *argv)
{

	bnxef_fcf_info_t fcf;

	/* If -i is not given zero is assumed */
	if (!(flags & DCMD_ADDRSPEC))
		return (DCMD_USAGE);

	if (mdb_vread(&fcf, sizeof (bnxef_fcf_info_t), (uint64_t)addr)
		!= sizeof (bnxef_fcf_info_t)) {
		mdb_printf("Unable to read the FCF info\n");
		return (DCMD_ERR);
	}

	mdb_printf("Mac Address %x:%x:%x:%x:%x:%x\n",
			fcf.mac_addr[0], fcf.mac_addr[1],
			fcf.mac_addr[2], fcf.mac_addr[3],
			fcf.mac_addr[4], fcf.mac_addr[5]);
	mdb_printf("Vlan Id 0x%x, priority 0x%x, vf_id  0x%x,\
		 fka_adv_period 0x%x, fc_map 0x%x\n",
		fcf.vlan_id, fcf.priority, fcf.vf_id,
		fcf.fka_adv_period, fcf.fc_map);
	mdb_printf("d Bit 0x%x, available for login %s,\
		 max_fcoe_size_verified 0x%x\n",
		fcf.d_bit_info,
		fcf.available_for_login? "Yes": "No",
		fcf.max_fcoe_size_verified);

	return (DCMD_OK);
}

static int
bnxef_count_active_tasks(uintptr_t addr, uint_t flags, int argc, const mdb_arg_t *argv)
{
	bnxef_driver_state_t *dsp;
	uint64_t *taskarray;
	bnxef_fcoe_task_t task;
	int count = 0, i;
	int cleanup_count= 0;
	
	
	/* If -i is not given zero is assumed */
	if (!(flags & DCMD_ADDRSPEC))
		return (DCMD_USAGE);

	/* Read the Global State Structure into Local structure */
	dsp = mdb_alloc(sizeof (bnxef_driver_state_t),  UM_SLEEP);
	if (dsp == NULL) {
		mdb_printf("Unable to allocate State structure\n");
		return (DCMD_ERR);
	}


	if (mdb_vread(dsp, sizeof (bnxef_driver_state_t), addr) !=
		sizeof (bnxef_driver_state_t)) {
			mdb_warn("Unable to read session structure");
			return (DCMD_ERR);
	}

	taskarray = mdb_alloc((dsp->max_exchanges * sizeof(uint64_t)), UM_SLEEP);
	if (mdb_vread(taskarray, (dsp->max_exchanges * sizeof(uint64_t)),
			(uint64_t)dsp->taskarray) != (dsp->max_exchanges *
			sizeof(uint64_t))) {
			mdb_printf("unable to read state_information");
			return (DCMD_ERR);
	}
	for (i=0; i< dsp->max_exchanges; i++) {
		if (mdb_vread(&task, sizeof(bnxef_fcoe_task_t),
				(uint64_t)taskarray[i]) !=
				sizeof(bnxef_fcoe_task_t)) {
			mdb_printf("unable to read task structure");
			return (DCMD_ERR);
		}
		if (task.task_flags & BNXEF_FCOE_TASK_ALLOCATED) {
			count++;
			mdb_printf("Task id -> 0x%x, flags 0x%x\n",
				task.x_id, task.task_flags);
			if (task.task_flags & BNXEF_FCOE_TASK_CLEANUP)
				cleanup_count++;
		}
	}
		
	mdb_printf("Total task allocated 0x%x\n", count);
	mdb_printf("Total Cleanup task allocated 0x%x\n", cleanup_count);
}

int
bnxef_mdb_check_cookie(uintptr_t cookieaddr, uint64_t g_addr)
{

	ddi_dma_cookie_t cookie;
	uint64_t start_phy_addr;
	uint64_t end_phy_addr;


	if (mdb_vread(&cookie, sizeof(ddi_dma_cookie_t), 	
		cookieaddr) != sizeof(ddi_dma_cookie_t)) {
		mdb_warn("unable to read cmd cookie\n");
		return (-1);
	}
	start_phy_addr = cookie._dmu._dmac_ll;
	end_phy_addr = cookie._dmu._dmac_ll +
				cookie.dmac_size -1;
	mdb_printf("0x%llx\t0x%llx\t", start_phy_addr, end_phy_addr); 
	if ((g_addr >= start_phy_addr) && 
		(g_addr <= end_phy_addr))  return 1;


	return 0;
}
					
void
bnxef_print_event_info (bnxef_trace_events_t *eventp)
{

	switch (eventp->event_id) {
	case BNXEF_SESSION_ENABLE_REQ:
		mdb_printf("0x%llx:\t Session Enable request sent on 0x%llx\n",
			eventp->lbolt, eventp->reserved1);
		break;
	case BNXEF_SESSION_ENABLE_COMP:
		mdb_printf("0x%llx:\t Session enable completion on 0x%llx\n",
		eventp->lbolt, eventp->reserved1);
		break;

	case BNXEF_SESSION_DISABLE_REQ:
		mdb_printf("0x%llx:\t Session disable request on 0x%llx\n",
		eventp->lbolt, eventp->reserved1);
		break;
	case BNXEF_SESSION_DISABLE_COMP:
		mdb_printf("0x%llx:\t Session disable completion on 0x%llx\n",
		eventp->lbolt, eventp->reserved1);
		break;
	
	case BNXEF_SESSION_OFFLOAD_REQ:
		mdb_printf("0x%llx:\t Session offload request on 0x%llx\n",
		eventp->lbolt, eventp->reserved1);
		break;

	case BNXEF_SESSION_OFFLOAD_COMP:
		mdb_printf("0x%llx:\t Session offload completion on 0x%llx\n",
		eventp->lbolt, eventp->reserved1);
		break;

	case BNXEF_SESSION_DESTROY_COMP:
		mdb_printf("0x%llx:\t Session destroy completion on 0x%llx\n",
		eventp->lbolt, eventp->reserved1);
		break;
	case BNXEF_SESSION_DESTROY_REQ:
		mdb_printf("0x%llx:\t Session destroy request on 0x%llx\n",
		eventp->lbolt, eventp->reserved1);
		break;

	case BNXEF_SCSI_TASK_INIT:
		mdb_printf("0x%llx:\t Task Initiated for task id 0x%x, on session 0x%llx\n", eventp->lbolt, eventp->task_id,
			eventp->reserved1);
		break;
	case BNXEF_SCSI_TASK_COMPLETE:
		mdb_printf("0x%llx:\t Task completed for task id 0x%x,on session 0x%llx\n", eventp->lbolt, eventp->task_id,
			eventp->reserved1);
		break;
	case BNXEF_SCSI_TASK_CLEANUP_COMP:
		mdb_printf("0x%llx:\t Task Cleanup Complete for task id 0x%x on session 0x%llx\n",
			eventp->lbolt, eventp->task_id, eventp->reserved1);
		break;
	case BNXEF_SCSI_TASK_SENSE_COMPLETED:
		mdb_printf("0x%llx:\t Task completed for task id 0x%x,on session 0x%llx, with SCSI status 0x%x\n",
			eventp->lbolt, eventp->task_id,
			eventp->reserved1, eventp->reserved3);
		break;
	case BNXEF_SCSI_TASK_TIMEOUT:
		mdb_printf("0x%llx:\t Task timeout for task id 0x%x,on session 0x%llx\n",
			eventp->lbolt, eventp->task_id, eventp->reserved1);
		break;
	case BNXEF_SCSI_TASK_ABTS_INITIATED:
		mdb_printf("0x%llx:\t ABTS initiated on task id 0x%x, using task id 0x%x, on session 0x%llx\n",
			eventp->lbolt, eventp->task_id,
			eventp->secondary_task_id, eventp->reserved1);
		break;
	case BNXEF_SCSI_TASK_BAACC_RECVD:
		mdb_printf("0x%llx:\t BA_ACC received on task id 0x%x, using task id 0x%x, on session 0x%llx\n",
			eventp->lbolt, eventp->task_id,
			eventp->secondary_task_id, eventp->reserved1);
		break;
	case BNXEF_SCSI_TASK_BARJT_RECVD:
		mdb_printf("0x%llx:\t BA_RJT received on task id 0x%x, using task id 0x%x, on session 0x%llx\n",
			eventp->lbolt, eventp->task_id,
			eventp->secondary_task_id, eventp->reserved1);
		break;
	case BNXEF_SCSI_TASK_ABTS_TIMEOUT:	
		mdb_printf("0x%llx:\t ABTS timeout on task id 0x%x, using task id 0x%x, on session 0x%llx\n",
			eventp->lbolt, eventp->task_id,
			eventp->secondary_task_id, eventp->reserved1);
		break;
	case BNXEF_SCSI_TASK_RRQ_INITIATED:
		mdb_printf("0x%llx:\t RRQ initiated on task id 0x%x, using task id 0x%x, on session 0x%llx\n",
			eventp->lbolt, eventp->task_id, eventp->secondary_task_id, eventp->reserved1);
		break;
	case BNXEF_SCSI_TASK_RRQ_COMPLETED:	
		mdb_printf("0x%llx:\t RRQ Completed on task id 0x%x, using task id 0x%x, on session 0x%llx\n",
			eventp->lbolt, eventp->task_id, eventp->secondary_task_id, eventp->reserved1);
		break;
	case BNXEF_SCSI_TASK_RRQ_LS_RJT:
		mdb_printf("0x%llx:\t RRQ LS_RJT on task id 0x%x, using task id 0x%x, on session 0x%llx\n",
			eventp->lbolt, eventp->task_id, eventp->secondary_task_id, eventp->reserved1);
		break;
	case BNXEF_SCSI_TASK_RRQ_TIMEOUT:
		mdb_printf("0x%llx:\t RRQ timeout on task id 0x%x, using task id 0x%x, on session 0x%llx\n",
			eventp->lbolt, eventp->task_id, eventp->secondary_task_id, eventp->reserved1);
		break;
	case BNXEF_SCSI_TASK_ERRORS:
		mdb_printf("0x%llx:\t Task error on task id 0x%x, on session 0x%llx, error 0x%x \n",
			eventp->lbolt, eventp->task_id, eventp->reserved1, eventp->reserved2);
		break;
			
	default:
		mdb_printf("0x%llx:\t Event ID 0x%x \n",
			eventp->lbolt, eventp->event_id);
	}
}
		
		
		
static int
bnxef_dump_trace_events(uintptr_t addr, uint_t flags, int argc,
	const mdb_arg_t *argv)
{

	bnxef_driver_state_t state;
	bnxef_trace_events_t event;
	uint64_t  event_addr;
	int i;

	if (!(flags & DCMD_ADDRSPEC))
		return (DCMD_USAGE);


	if (mdb_vread(&state, sizeof (bnxef_driver_state_t), addr) !=
		sizeof (bnxef_driver_state_t)) {
			mdb_warn("Unable to read state structure");
			return (DCMD_ERR);
	}

	for(i=state.gev_log_idx; i < state.bnxef_max_trace_events; i++) {
		event_addr = (uint64_t)state.global_event_logp + i *
				sizeof(bnxef_trace_events_t);
		if (mdb_vread(&event, sizeof(bnxef_trace_events_t), event_addr)
			!= sizeof(bnxef_trace_events_t)) {
			mdb_warn("Unable to read the event structure");
			return (DCMD_ERR);
		}
		if (event.event_id != 0) 
			bnxef_print_event_info(&event);	
	}


	for (i=0; i < state.gev_log_idx; i++) {
		event_addr = (uint64_t)state.global_event_logp + i *
				sizeof(bnxef_trace_events_t);
		if (mdb_vread(&event, sizeof(bnxef_trace_events_t), event_addr) 
			!= sizeof(bnxef_trace_events_t)) {
			mdb_warn("Unable to read the event structure");
			return (DCMD_ERR);
		}
		if (event.event_id != 0) 
			bnxef_print_event_info(&event);
	}

	return (DCMD_OK);
}




static int
bnxef_find_phy_address(uintptr_t addr, uint_t flags, int argc, 
	const mdb_arg_t *argv)
{
	bnxef_driver_state_t state;
	bnxef_fcoe_task_t    task;
	fc_packet_t	pkt;
	uint_t	opt_a = FALSE;
	int i, j, ret;
	uintptr_t taskaddr;
	uintptr_t pktaddr;
	uintptr_t cookieaddr;
	char *opt_addr_str = NULL;
	u_longlong_t g_addr, start_phy_addr, end_phy_addr; 

	if (mdb_getopts(argc, argv,
		'a', MDB_OPT_STR, &opt_addr_str,
		 NULL) != argc) {
		return (DCMD_USAGE);
	}

	g_addr = mdb_strtoull(opt_addr_str);

	mdb_printf("The Given Address 0x%llx\n", g_addr);

	if (!(flags & DCMD_ADDRSPEC))
		return (DCMD_USAGE);


	if (mdb_vread(&state, sizeof (bnxef_driver_state_t), addr) !=
		sizeof (bnxef_driver_state_t)) {
			mdb_warn("Unable to read state structure");
			return (DCMD_ERR);
	}

	mdb_printf("Number of Exchanges configured 0x%x\n", 
		state.max_exchanges);
	mdb_printf("Task Array start pointer 0x%p\n",
		state.taskarray);

	for (i=0; i< state.max_exchanges; i++) {
		addr = (uintptr_t)state.taskarray + i*sizeof(uintptr_t);
		if (mdb_vread(&taskaddr, sizeof(uintptr_t),
				addr) != sizeof(uintptr_t)) {
			mdb_warn("Unable to read taskaddr");
			return (DCMD_ERR);
		}
//		mdb_printf("Task id 0x%x Task Address 0x%p\n", i, taskaddr);

		if (mdb_vread(&task, sizeof(bnxef_fcoe_task_t),
				taskaddr) != sizeof(bnxef_fcoe_task_t)) {
			mdb_warn("unable to read task structure");
			return (DCMD_ERR);
		}

		
		if (task.task_flags &  BNXEF_FCOE_TASK_ALLOCATED) {
			mdb_printf("\n0x%x\t 0x%x\t", i, task.task_flags);
			if (task.bd_cookie_cnt > 0) {
				start_phy_addr = task.bd_cookie._dmu._dmac_ll;
				end_phy_addr = task.bd_cookie._dmu._dmac_ll + 
						task.bd_cookie.dmac_size -1;
		//		mdb_printf("BD_ADDR: Start phy Addr 0x%llx, end Phy addr 0x%llx\n",
		//			start_phy_addr, end_phy_addr);
				mdb_printf("0x%llx\t0x%llx\t", start_phy_addr,
						end_phy_addr); 
				if ((g_addr >= start_phy_addr) && 
					(g_addr <= end_phy_addr)) {
					mdb_printf("Task 0x%x, bd_addr_start 0x%x, bd_addr_end 0x%llx\n",
						 i, start_phy_addr, end_phy_addr);
				}
				if ((g_addr + 4096 -1 >= start_phy_addr) &&
				 	(g_addr + 4096 -1 <= end_phy_addr)) {
					mdb_printf("Task 0x%x, bd_addr_start 0x%x, bd_addr_end 0x%llx\n",
						i, start_phy_addr, end_phy_addr);
				}
			}
				
			pktaddr = (uintptr_t) task.pktp;
			
			if (pktaddr != 0) {
				if (mdb_vread(&pkt, sizeof(fc_packet_t),
						pktaddr) != sizeof(fc_packet_t)) {
					mdb_warn("Unable to read pkt structure");
					return (DCMD_ERR);
				}

				/* check the cmd_len */

				if (pkt.pkt_cmdlen > 0) {
					if (pkt.pkt_cmd_cookie != 0) {
					for (j=0; j< pkt.pkt_cmd_cookie_cnt; j++) {
						cookieaddr = (uintptr_t)
						pkt.pkt_cmd_cookie + j *
							sizeof(ddi_dma_cookie_t);
						ret = bnxef_mdb_check_cookie(cookieaddr,
							g_addr);
						if (ret < 0) {
						mdb_printf("error in reading pkt_cmd_cookie Task 0x%x, Cookie 0x%x\n", i, j);
						} else if (ret == 1) {
							mdb_printf("Task 0x%x -> Pkt_cmd_cookie\n", i);
						}
					}
					}
				}

				if (pkt.pkt_rsplen > 0) {
					if (pkt.pkt_resp_cookie != 0) {
					for (j=0; j< pkt.pkt_resp_cookie_cnt; j++) {
						cookieaddr = (uintptr_t)
						pkt.pkt_resp_cookie + j *
							sizeof(ddi_dma_cookie_t);
						ret = bnxef_mdb_check_cookie(cookieaddr,
							g_addr);
						if (ret < 0) {
							mdb_printf("error in reading pkt_resp_cookie 0x%x Cookie 0x%x\n", i, j);
						} else if (ret == 1) {
							mdb_printf("Task 0x%x-> pkt_rsp_cookie\n", i);
						}
					}
					}

				}

				if (pkt.pkt_datalen > 0) {
					for (j=0; j< pkt.pkt_data_cookie_cnt; j++) {
						cookieaddr = (uintptr_t)
						pkt.pkt_data_cookie + j *
							sizeof(ddi_dma_cookie_t);
						ret = bnxef_mdb_check_cookie(cookieaddr,
							g_addr);
						if (ret < 0) {
							mdb_printf("error in reading pkt_data__cookie 0x%x Cookie 0x%x\n", i, j);
						} else if (ret == 1) {
							mdb_printf("Task 0x%x-> pkt_data_cookie 0%x\n", i, j);
						}
					}
				}
			}
		} 
	} 
}

static int
bnxef_dump_dma_cookies(uintptr_t addr, uint_t flags, int argc,
	const mdb_arg_t *argv)
{
	int i,j,s,r, size;
	int num_instances;
	uint64_t *global_state;
	bnxef_driver_state_t state;
	bnxef_driver_state_t *dsp;
	bnxef_port_t port, *portp;
	bnxef_session_t session, *sessionp;
	uint64_t sessionaddr;
	int count=0;

	/* If -i is not given zero is assumed */
	if (!(flags & DCMD_ADDRSPEC))
		return (DCMD_USAGE);


	mdb_printf("Printing Information about First Instance\n");
	if (addr != NULL) {
		mdb_printf("Attached Instance Number %d\n", i);
		/* Read the Driver State Structure in the memory allocated */
		if (mdb_vread(&state,
			sizeof (bnxef_driver_state_t), addr)
			!= sizeof (bnxef_driver_state_t)) {
			mdb_printf("Unable to read state information\n");
			return (DCMD_ERR);
		}
		dsp = &state;

		mdb_printf("Global DMA cookies:\n");
		mdb_printf("===================\n");
		mdb_printf("\n Handle Name\t\tStart Addr\t\t Size\n");
		mdb_printf("==============================================================\n");
		for (j=0; j< dsp->num_task_dma_handles; j++) {
			mdb_printf("%s%03d\t\t0x%p\t\t0x%x\n",
				"Task#", j, dsp->task_dma_cookie[j]._dmu._dmac_ll, 
				dsp->task_dma_cookie[j].dmac_size);
		}

		mdb_printf("TaskCtxBD\t\t0x%p\t\t0x%x\n",
			dsp->task_ctx_bd_cookie._dmu._dmac_ll,
			dsp->task_ctx_bd_cookie.dmac_size);
		mdb_printf("HashTbl\t\t\t0x%p\t\t0x%x\n",
			dsp->hash_tbl_dma_cookie._dmu._dmac_ll,
			dsp->hash_tbl_dma_cookie.dmac_size);
		mdb_printf("HashTblPbl\t\t0x%p\t\t0x%x\n",
			dsp->hash_table_pbl_cookie._dmu._dmac_ll,
			dsp->hash_table_pbl_cookie.dmac_size);
		mdb_printf("T2HashTbl\t\t0x%p\t\t0x%x\n",
			dsp->t2_hash_table_dma_cookie._dmu._dmac_ll,
			dsp->t2_hash_table_dma_cookie.dmac_size);
		mdb_printf("T2HashTblPbl\t\t0x%p\t\t0x%x\n",
			dsp->t2_hash_table_dma_ptr_cookie._dmu._dmac_ll,
			dsp->t2_hash_table_dma_ptr_cookie.dmac_size);
		mdb_printf("DummyBuf\t\t0x%p\t\t0x%x\n",
			dsp->dummy_dma_cookie._dmu._dmac_ll,
			dsp->dummy_dma_cookie.dmac_size);
		mdb_printf("StatsBuf\t\t0x%p\t\t0x%x\n",
			dsp->stats_dma_cookie._dmu._dmac_ll,
			dsp->stats_dma_cookie.dmac_size);

		if (mdb_vread(&port,
			sizeof (bnxef_port_t), (uint64_t)dsp->port_list.head)
			!= sizeof (bnxef_port_t)) {
			mdb_printf("Unable to read port information\n");
			return (DCMD_ERR);
		}
		portp = &port;


		count=0;
		for (s = 0; s < 256; s++) {
			sessionaddr = (uintptr_t)portp->sessionhash[s].head;
			for (r=0; r< portp->sessionhash[s].num_items; r++) {
				if (sessionaddr != NULL) {
					if (mdb_vread(&session,
						sizeof (bnxef_session_t),
						sessionaddr) !=
						sizeof (bnxef_session_t)) {
						mdb_warn("Unable to read\
							the session address\n");
						return (DCMD_ERR);
					}
					sessionp = &session;
					if (sessionp->state & BNXEF_SESSION_OFFLOADED) { 
						count++;
						mdb_printf("\nSession #:%d\n", count);
						mdb_printf("------------\n");
						mdb_printf("SQCookie\t\t0x%p\t\t0x%x\n",
							sessionp->sq_cookie._dmu._dmac_ll,
							sessionp->sq_cookie.dmac_size);
						mdb_printf("CQCookie\t\t0x%p\t\t0x%x\n",
							sessionp->cq_cookie._dmu._dmac_ll,
							sessionp->cq_cookie.dmac_size);
						mdb_printf("RQCookie\t\t0x%p\t\t0x%x\n",
							sessionp->rq_cookie._dmu._dmac_ll,
							sessionp->rq_cookie.dmac_size);
						mdb_printf("RQpblCookie\t\t0x%p\t\t0x%x\n",
							sessionp->rq_pbl_dma_cookie._dmu._dmac_ll,
							sessionp->rq_pbl_dma_cookie.dmac_size);
						mdb_printf("XferQCookie\t\t0x%p\t\t0x%x\n",
							sessionp->xferq_cookie._dmu._dmac_ll,
							sessionp->xferq_cookie.dmac_size);
						mdb_printf("ConfqCookie\t\t0x%p\t\t0x%x\n",
							sessionp->confq_cookie._dmu._dmac_ll,
							sessionp->confq_cookie.dmac_size);
						mdb_printf("ConfqpblCookie\t\t0x%p\t\t0x%x\n",
							sessionp->confq_pbl_cookie._dmu._dmac_ll,
							sessionp->confq_pbl_cookie.dmac_size);
						mdb_printf("lcqCookie\t\t0x%p\t\t0x%x\n",
							sessionp->lcq_cookie._dmu._dmac_ll,
							sessionp->lcq_cookie.dmac_size);
						mdb_printf("ConndbCookie\t\t0x%p\t\t0x%x\n",
							sessionp->conn_db_cookie._dmu._dmac_ll,
							sessionp->conn_db_cookie.dmac_size);
					}
					sessionaddr = (uintptr_t)sessionp->next;
				}
			}
		}
	}

	return (DCMD_OK);

}


/* ARGSUSED0 */
static int
bnxef_dump_all(uintptr_t addr, uint_t flags, int argc,
	const mdb_arg_t *argv)
{

	int i, size;
	int num_instances;
	uint64_t *global_state;
	bnxef_driver_state_t *dsp;

	mdb_readvar(&num_instances, "num_instances");

	global_state = mdb_alloc(num_instances * sizeof (uint64_t),
			UM_SLEEP);
	if (global_state != NULL) {
		size = mdb_readsym(global_state,
			BNXEF_MAX_INSTANCES * sizeof (uint64_t),
			"global_state");
		if (size != (BNXEF_MAX_INSTANCES * sizeof (uint64_t))) {
			mdb_printf("Unable to read State ptrs\n");
			return (DCMD_ERR);
		}
	}

	for (i = 0; i < BNXEF_MAX_INSTANCES; i++) {
		/* Read the Global State Structure into Local structure */
		dsp = mdb_alloc(sizeof (bnxef_driver_state_t),  UM_SLEEP);
		if (dsp == NULL) {
			mdb_printf("Unable to allocate State structure\n");
			return (DCMD_ERR);
		}
		

		

		if (global_state[i] != NULL) {
			mdb_printf("Attached Instance Number %d\n", i);
			/* Read the Driver State Structure in the memory allocated */
			if (mdb_vread(dsp,
				sizeof (bnxef_driver_state_t), global_state[i])
				!= sizeof (bnxef_driver_state_t)) {
				mdb_printf("Unable to read state information\n");
				return (DCMD_ERR);
			}

			/* This will dump the physical port information */
			bnxef_dump_physical_port_info(dsp);

			/* This will dump the fcf information */
			bnxef_dump_fcf_info(dsp);

			/* Dump Configuration Information */
			bnxef_dump_config_info(dsp);

			/* Firmware resource information */
			bnxef_dump_fw_res_info(dsp);

			/* VN/N_PORT Information */
			bnxef_dump_nport_info(dsp);

			/* Dump the Free task list information */
			bnxef_dump_free_task_list_info(dsp);
		}

		/* Free the resource allocated */
		mdb_free(dsp, sizeof (bnxef_driver_state_t));
	}

	return (DCMD_OK);
}



static const mdb_dcmd_t bnxef_dcmds[] = {
    {
        "bnxef_task_pbl",
        "? [-a |-n | -i <item> ] ",
        "Prints task PBL information",
        bnxef_task_pbl,
	bnxef_task_pbl_help
    },
    {
        "bnxef_hash_pbl",
        "[args ...]",
        "Prints Hash PBL information",
        bnxef_hash_pbl
    },
    {
        "bnxef_fcf_info",
        "[args ...]",
        "Prints FCF information",
        bnxef_fcf_info
    },
    {
        "bnxef_driver_state",
        "[args ...]",
        "Print Task list Information",
        bnxef_driver_state
    },
    {
        "bnxef_task",
        "",
        "Prints Task information",
        bnxef_fcoe_task
    },
    {
        "bnxef_port",
        "[-v -H -h <bucket #> -s -u]",
        "Print Port information",
        bnxef_port_info
    },
    {
        "bnxef_session",
        "[-v]",
        "Print Session Information",
        bnxef_session_info
    },
    {
        "dumpphyaddr",
        "[-s <size>]",
        "Dumps size bytes starting from physcial address",
        bnxef_dump_hash_table
    },
    {
        "bnxef_t2_hash_tbl",
        "[-a -i <session number>]",
        "Print the T2 hash table content",
        bnxef_t2_hash_tbl
    },
    {
        "bnxef_cqdump",
        "[-s <size in bytes> -t -c <cons idx>]",
        "Print the Current completions from firmware",
        bnxef_dump_cq
    },
    {
        "bnxef_sqdump",
        "[-s <sizein bytes> -p <prod idx> -t]",
        "Print the submission to firmware",
        bnxef_dump_sq
    },
    {
        "bnxef_dump_all",
        "[No options]",
        "Prints all information needed for debugging",
        bnxef_dump_all
    },
    {
	"bnxef_sess_life_cycle",
	"[No Options]",
	"Print log of session life cycle",
	bnxef_session_life_cycle_log
     },
    {
	"bnxef_completion_log",
	"[No Options]",
	"Print all completed task info",
	bnxef_task_completion_log
     },
     {
	"bnxef_dump_all_task_log",
	"[no options]",
	"print all completed task info for all sessions",
	bnxef_dump_all_session_task_log
      },
	
      {
	"bnxef_dma_cookies",
	"[no options]",
	"print all global DMA cookie and Session specific cookies",
	bnxef_dump_dma_cookies
	},
      {
        "bnxef_find_phy_addr",
         "[-a <Physical address> ]",
	 "print context under which this Phy address is used",
	bnxef_find_phy_address
        },

      {
	"bnxef_dump_trace_events",
	"[no options]",
	"prints the global trace events for this instance",
	bnxef_dump_trace_events
       },
	
    {
        NULL
    }
};

static mdb_modinfo_t bnxef_mdb =  {
    MDB_API_VERSION,
    bnxef_dcmds,
    bnxef_walkers
};

mdb_modinfo_t * _mdb_init(void)
{
    return &bnxef_mdb;
}

void _mdb_fini(void)
{
    return;
}

